/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.datastore.DataStoreCache;
import javax.jdo.listener.InstanceLifecycleListener;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.jmx.comm.CommunicatorAttributes;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.mapper.rdb.lib.ConnectionSpecJDBC;
import org.objectweb.perseus.cache.api.CacheAttributeController;
import org.objectweb.perseus.concurrency.pessimistic.PessimisticConcurrencyManagerAC;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoVersion;
import org.objectweb.speedo.api.TransactionListener;
import org.objectweb.speedo.jmx.JMXConfigurator;
import org.objectweb.speedo.lib.FractalHelper;
import org.objectweb.speedo.mapper.api.JormFactoryAttributes;
import org.objectweb.speedo.mapper.api.MapperAttributes;
import org.objectweb.speedo.mapper.rdb.JDBCMapperAttributes;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.query.api.QueryManagerAttribute;
import org.objectweb.speedo.sequence.api.SequenceManager;
import org.objectweb.speedo.workingset.lib.SpeedoTransaction;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

/**
 * This class is a client helper which permits to create a new speedo
 * PersistentManagerFactory. Due to the JDO constraint the persistence manager
 * factory implementation must have a public empty contructor. But the use of
 * the fractal component (www.objectweb.org/fractal) needs to use a
 * fractal implementation to initialize components. Then this class is an
 * implementation of the PersistentManagerFactory interface which delegates all
 * calls on a persistent manager factory delegate. The constructor of this class
 * make the components creation and their initialization via Julia (fractal
 * implementation).
 *
 * @author S.Chassande-Barrioz
 */
public class Speedo implements PersistenceManagerFactory, ProxyManagerFactory {

    public final static String CONNECTION_SPEC_FACTORY =
        "org.objectweb.speedo.connectionSpecFactory";

    private final static String LOGGER_NAME = SpeedoProperties.LOGGER_NAME + ".init";
    private final static String SPEEDO_TEMPLATE = "org.objectweb.speedo.Speedo";
    private final static String SPEEDO_JMX_TEMPLATE = "org.objectweb.speedo.SpeedoJMX";
    private final static String OPTIMISTIC_CONCURRENCY_TEMPLATE =
		"org.objectweb.perseus.persistence.concurrency.POptimisticConcurrencyManager";
    private final static String PESSIMISTIC_CONCURRENCY_TEMPLATE =
		"org.objectweb.perseus.persistence.concurrency.PPessimisticConcurrencyManager";
    private final static String DBDELEGATE_CONCURRENCY_TEMPLATE =
		"org.objectweb.perseus.persistence.concurrency.PDbDelegateConcurrencyManager";
    public final static String PM_POOL_PATH = "proxy-manager-pool";
    public final static String MEMORY_CACHE_PATH = "tpm.cache-manager";
    public final static String DEPENDENCY_GRAPH_PATH = "tpm.dependency-graph";
    public final static String QUERY_CACHE_PATH = "compiled-query-cache";
    public final static String QUERY_MANAGER = "query-manager";
    public final static String CONNECTION_POOL_PATH = "mapper.pool";
    public final static String MONOLOG_FACTORY_PATH = "monolog-factory";
    public final static String PRIMITIVE_MAPPER_PATH = "mapper.mapper";
    public final static String JORM_FACTORY_PATH = "mapper.jorm-factory";
    public final static String PMF_PATH = "proxy-manager-factory";
    public final static String JMX_AGENT_PATH = "agent";
    public final static String HTML_JMX_AGENT_PATH = JMX_AGENT_PATH + ".html";


    private static Map INSTANCES = new HashMap();
    private final static String FRACTAL_PROVIDER = "fractal.provider";
    private final static String DEFAULT_FRACTAL_PROVIDER
            = "org.objectweb.fractal.julia.Julia";

    private final static String JULIA_LOADER = "julia.loader";
    private final static String DEFAULT_JULIA_LOADER
            = "org.objectweb.fractal.julia.loader.DynamicLoader";

    private final static String JULIA_CONFIG = "julia.config";
    private final static String DEFAULT_JULIA_CONFIG = "julia.cfg";
    
    private static Factory factory = null;
    private static Component speedoTemplate = null;

    // method called by the JDOHelper.getPersistenceManagerFactory method
    public static PersistenceManagerFactory getPersistenceManagerFactory(
            Properties props) throws Throwable {
        return getPersistenceManagerFactory((Map) props);
    }
    // method called by the JDOHelper.getPersistenceManagerFactory method
    public static PersistenceManagerFactory getPersistenceManagerFactory(
            Map m) throws Throwable {
        Object instance;
        synchronized (INSTANCES) {
            instance = INSTANCES.get(m);
            if (instance == null) {
                instance = new Speedo(m);
                INSTANCES.put(m, instance);
            }
        }
        return (PersistenceManagerFactory) instance;
    }

    public static void clearInstances() {
        INSTANCES.clear();
    }

    /**
     * The persistence manager factory delegate
     */
    private PersistenceManagerFactory delegate = null;
	ProxyManagerFactory pmfComponent = null;
	private Logger logger = null;
	private Component speedo = null;
	boolean isPropertiesInitialized = false;

	public Speedo() throws Throwable {
		getSpeedoComponent(null);
		isPropertiesInitialized = false;
	}
    /**
     * It creates and initializes a real SpeedoProxyManagerFactory with
     * Julia (Fractal implementation).
     */
    public Speedo(Map props) throws Throwable {
		try {
			init(getSpeedoComponent(props), props);
			isPropertiesInitialized = true;
		} catch(Throwable e) {
			System.err.println("Error during the instanciation of the Speedo " +
					"persistence manager factory:");
			e.printStackTrace(System.err);
			throw e;
		}
	}

	public ProxyManagerFactory getPMFComponent() throws Exception {
		return pmfComponent;
	}

    private Factory getADLFactory () throws Exception {
        if (factory == null) {
            factory = FactoryFactory.getFactory(FactoryFactory.FRACTAL_BACKEND);
        }
        return factory;
    }
    
    private static Object getProperty(Map m, String propName) {
        return m.get(propName);
    }

    private static String getProperty(Map m, String propName, String defaultValue) {
        Object res = m.get(propName);
        if (res instanceof String) {
            return (String) res;
        } else  {
            return defaultValue;
        }
    }
	private Component getSpeedoComponent(Map props) throws Throwable {
		
		if (speedo == null) { 
			//use Julia as Fractal implementation
			System.setProperty(FRACTAL_PROVIDER, DEFAULT_FRACTAL_PROVIDER);
			System.setProperty(JULIA_LOADER, DEFAULT_JULIA_LOADER);
			System.setProperty(JULIA_CONFIG, DEFAULT_JULIA_CONFIG);

			//Allocate a Logger through the LoggerFactory component

			Monolog.initialize();

			LoggerFactory loggerFactory = Monolog.monologFactory;
			logger = loggerFactory.getLogger(LOGGER_NAME);
			String templateName = SPEEDO_TEMPLATE;
			if (props != null 
			        && isTrue(getProperty(props, SpeedoProperties.JMX, "false"))) {
			    templateName = SPEEDO_JMX_TEMPLATE;
			}
			//instanciate the Speedo component
			try {
				if (speedoTemplate == null) {
					speedoTemplate = (Component)getADLFactory().newComponent(
					        templateName, 
					        Collections.singletonMap("template", "true"));
				}
				//speedo = (Component)getADLFactory().newComponent(SPEEDO_TEMPLATE, null);
				speedo = Fractal.getFactory(speedoTemplate).newFcInstance();
			} catch (Throwable e) {
				System.err.println("Error during the template loading or the " +
						"component instanciation '" + SPEEDO_TEMPLATE + "': ");
				throw e;
			}
			Fractal.getNameController(speedo).setFcName("org.objectweb.speedo");
			delegate = (PersistenceManagerFactory) speedo.getFcInterface("persistence-manager-factory");	
		}
		return speedo;
	}
	
	public void init(Component speedo, Map pmfProps) throws Throwable {
	    logger.log(BasicLevel.INFO, "Speedo " + SpeedoVersion.SPEEDO_VERSION 
	            + " is starting ...");

		//Manage properties of the PersistenceManagerFactory
		//--------------------------------------------------
		Map props = new HashMap();
		props.putAll(pmfProps);
		String strval;

		props.remove(SpeedoProperties.JDO_PERSISTENCE_MANAGER_FACTORY_CLASS);

		strval = getProperty(props, SpeedoProperties.IMRICATED_PM_ALLOWED, "false");
		props.remove(SpeedoProperties.IMRICATED_PM_ALLOWED);
		boolean imbricatedPM = Boolean.valueOf(strval).booleanValue();
		if (imbricatedPM) {
		    logger.log(BasicLevel.INFO, "PM reused on imbricated PMF.getPM()");
		}
		
		//Debug Mode
		strval = getProperty(props, SpeedoProperties.DEBUG, null);
		props.remove(SpeedoProperties.DEBUG);
		if (strval != null) {
			Debug.ON = Boolean.valueOf(strval).booleanValue();
			logger.log(BasicLevel.INFO, SpeedoProperties.DEBUG + ": " + Debug.ON);
		}

		boolean useConnectionFactory = configureMapper(speedo, props);
		configurePool(getSubComponent(speedo, PM_POOL_PATH),
		        "PersistenceManager pool: ",
				SpeedoProperties.PM_POOL_MIN,
				SpeedoProperties.PM_POOL_MAX,
				SpeedoProperties.PM_POOL_TTL,
				SpeedoProperties.PM_POOL_INACTIVETTL,
				SpeedoProperties.PM_POOL_TIMEOUT,
				props);
	    configureCache(getSubComponent(speedo, MEMORY_CACHE_PATH), 
	            "Persistent objects cache: ",
	            SpeedoProperties.CACHE_SIZE,
	            SpeedoProperties.CACHE_AUTO_CLEAN_SIZE,
	            SpeedoProperties.CACHE_AUTO_CLEAN_THRESHOLD,
	            SpeedoProperties.CACHE_REPLCAEMENT,
	            props);
	    configureCache(getSubComponent(speedo, QUERY_CACHE_PATH), 
	            "Compiled query cache: ",
	            SpeedoProperties.COMPILED_QUERY_CACHE_SIZE,
	            SpeedoProperties.COMPILED_QUERY_CACHE_AUTO_CLEAN_SIZE,
	            SpeedoProperties.COMPILED_QUERY_CACHE_AUTO_CLEAN_THRESHOLD,
	            SpeedoProperties.COMPILED_QUERY_CACHE_POLICY,
	            props);

		//deprecated properties
		final String TM_NAME_old = "org.objectweb.perseus.connector.ra.jdo.TMName";
		strval = getProperty(props, TM_NAME_old, "");
		if (strval.length()>0) {
			logger.log(BasicLevel.WARN, "Property " + TM_NAME_old
					+ " is deprecated, you must use " + SpeedoProperties.TM_NAME);
			props.remove(TM_NAME_old);
			pmfProps.remove(TM_NAME_old);
			if (getProperty(props, SpeedoProperties.TM_NAME) == null) {
                props.put(SpeedoProperties.TM_NAME, strval);
				pmfProps.put(SpeedoProperties.TM_NAME, strval);
			}
		}

		// Transaction managed by an application server
		strval = getProperty(props, SpeedoProperties.MANAGED, "");
		if (strval.length()>0) {
			boolean b = isTrue(strval);
			if (!useConnectionFactory && b) {
				String msg = "In a managed environnement, a connection factory is required";
				logger.log(BasicLevel.ERROR, msg);
				throw new JDOUserException(msg);
			}
			logger.log(BasicLevel.INFO, SpeedoProperties.MANAGED + ": Speedo is used "
					+ (b
					? "in a managed environnement"
					: "in stand alone"));

			//TransactionManager JNDI name
			strval = getProperty(props, SpeedoProperties.TM_NAME, "");
			props.remove(SpeedoProperties.TM_NAME);
			logger.log(BasicLevel.INFO, SpeedoProperties.TM_NAME + ": "
					+ (strval.length()>0
					? strval
					:  ": no value specified, use a well known application server"));
		} else {
			props.remove(SpeedoProperties.TM_NAME);
		}

		//Listener of Transaction life cycle (statistic)
		strval = getProperty(props, SpeedoProperties.TX_LISTENER, "");
		if (strval.length()>0) {
			try {
				SpeedoTransaction.txListener = (TransactionListener)
						Class.forName(strval).newInstance();
				if (SpeedoTransaction.txListener instanceof Loggable) {
					((Loggable) SpeedoTransaction.txListener)
							.setLoggerFactory(Monolog.monologFactory);
				}
				logger.log(BasicLevel.INFO, "Transaction listener: " + strval);
			} catch (Exception e) {
				logger.log(BasicLevel.WARN,
						"Impossible to instanciate the TransactionListener specified '"
						+ strval + "': ", e);
			}
		}

		configurePrefetching(speedo, props);
		configureConcurrencyManager(speedo, props);
		boolean jmxOn = configureJMX(speedo, props);
		
		//managed other properties
        for (Iterator it = props.entrySet().iterator(); it.hasNext();) {
            Map.Entry me = (Map.Entry) it.next();
            String key = (String) me.getKey();
            String value = (String) me.getValue();
			boolean knownProperty = true;
            if (SpeedoProperties.JDO_OPTION_RETAIN_VALUES.equals(key)) {
                delegate.setRetainValues(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_RESTORE_VALUES.equals(key)) {
                delegate.setRestoreValues(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_IGNORE_CACHE.equals(key)) {
                delegate.setIgnoreCache(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_READ.equals(key)) {
                delegate.setNontransactionalRead(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_WRITE.equals(key)) {
                delegate.setNontransactionalWrite(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_MULTITREADED.equals(key)) {
                delegate.setMultithreaded(Boolean.valueOf(value).booleanValue());

            } else {
				knownProperty = false;
            }
			if (knownProperty) {
				logger.log(BasicLevel.INFO, key + ": " + value);
				it.remove();
			}
        }

        //Warn the unknwon properties
		for (Iterator it = props.entrySet().iterator(); it.hasNext();) {
			Map.Entry me = (Map.Entry) it.next();
			String key = (String) me.getKey();
			if (key.startsWith(SpeedoProperties.CACHE_CLASS_POLICY+"(")
			        || key.startsWith(SpeedoProperties.USER_CACHE_CLASS_POLICY+"(")
			        || key.startsWith(SpeedoProperties.PREFETCH_ON_GENCLASS+"(")
			        || key.startsWith(SpeedoProperties.PREFETCH_ON_QUERY+"(")
			        || key.startsWith(SpeedoProperties.PREFETCH_ON_EXTENT+"(")
			        ) {
			    continue;
			}
			String value = (String) me.getValue();
			logger.log(BasicLevel.WARN, "The (" + key + ", " + value
				+ ") property is not managed");
		}

		delegate.getProperties().putAll(pmfProps);
        Fractal.getLifeCycleController(speedo).startFc();
		pmfComponent = (ProxyManagerFactory) getSubComponent(
				((Interface) delegate).getFcItfOwner(), PMF_PATH)
				.getFcInterface("proxy-manager-factory");
		if (jmxOn) {
		    new JMXConfigurator(speedo, logger).init();
		}
    }
    private Component getSubComponent(Component parent,
            String path) throws Exception {
        return FractalHelper.getSubComponent(parent, path, logger);
    }

    private void configurePool(Component pool,
            				   String poolLabel,
							   String minProp,
							   String maxProp,
							   String ttlProp,
							   String inactivettlProp,
							   String timeoutProp,
							   Map props) throws Exception {
		String strval = null;
		int poolMinSize = -2;
		int poolMaxSize = -2;
		long poolTTL = -2;
		long poolInactiveTTL = -2;
		long poolTimeout = -2;
		boolean conf = false;
		StringBuffer sb = new StringBuffer(poolLabel);
		final String sep = ", ";
		if (minProp != null) {
			strval = getProperty(props, minProp, "");
			if (strval.length() > 0) {
				props.remove(minProp);
				poolMinSize = Integer.parseInt(strval);
				if (conf) {
				    sb.append(sep);
				}
				sb.append("min=" + strval);
				conf = true;
			}
		}
		if (maxProp != null) {
			strval = getProperty(props, maxProp, "");
			if (strval.length() > 0) {
				props.remove(maxProp);
				if (strval.equalsIgnoreCase("nolimit")) {
					poolMaxSize = -1;
				} else {
					poolMaxSize = Integer.parseInt(strval);
				}
				if (conf) {
				    sb.append(sep);
				}
				sb.append("max=" + strval);
				conf = true;
			}
		}
		if (ttlProp != null) {
			strval = getProperty(props, ttlProp, "");
			if (strval.length() > 0) {
				props.remove(ttlProp);
				try {
					poolTTL = 1000 * Long.parseLong(strval);
				} catch (NumberFormatException e) {
					poolTTL = -1;
					strval = "NOTTL";
				}
				if (conf) {
				    sb.append(sep);
				}
				sb.append("ttl=" + strval);
				conf = true;
			}
		}
		if (inactivettlProp != null) {
			strval = getProperty(props, inactivettlProp, "");
			if (strval.length() > 0) {
				props.remove(inactivettlProp);
				try {
					poolInactiveTTL = 1000 * Long.parseLong(strval);
				} catch (NumberFormatException e) {
				    poolInactiveTTL = -1;
					strval = "NOTTL";
				}
				if (conf) {
				    sb.append(sep);
				}
				sb.append("inactivettlProp=" + strval);
				conf = true;
			}
		}
		if (timeoutProp != null) {
			strval = getProperty(props, timeoutProp, "");
			if (strval.length() > 0) {
				props.remove(timeoutProp);
				try {
					poolTimeout = Long.parseLong(strval);
				} catch (NumberFormatException e) {
					poolTimeout = -1;
					strval = "WAIT";
				}
				if (conf) {
				    sb.append(sep);
				}
				sb.append("timeout=" + strval);
				conf = true;
			}
		}
		if (conf) {
			logger.log(BasicLevel.INFO, sb.toString());
		      PoolAttributes poolAttr;
		      try {
		        poolAttr = (PoolAttributes)
							Fractal.getAttributeController(pool);
		      } catch (Exception e) {
		        Component[] children = Fractal.getContentController(speedo).getFcSubComponents();
		        String[] strs= new String[children.length];
		        for (int i = 0; i < strs.length; ++i) {
		          strs[i] = Fractal.getNameController(children[i]).getFcName();
		        }
		        throw new RuntimeException("" + Arrays.asList(strs));
		      }
			if (poolMaxSize > -2) {
				poolAttr.setMaxSize(poolMaxSize);
			}
			if (poolMinSize > -2) {
				poolAttr.setMinSize(poolMinSize);
			}
			if (poolTTL > -2) {
				poolAttr.setTTL(poolTTL);
			}
			if (poolInactiveTTL > -2) {
				poolAttr.setInactiveTTL(poolInactiveTTL);
			}
			if (poolTimeout > -2) {
				poolAttr.setTimeout(poolTimeout);
			}
		}
	}

	/**
	 * Configures the memory cache.
	 * @param speedo is the speedo composite component
	 * @param props is the speedo initialisation properties
	 */
	private void configureCache(Component cm, 
	        String cacheLabel,
	        String sizeProp,
	        String autoCleanSizeProp,
	        String autoCleanThresholdProp,
	        String policyProp,
	        Map props)
			throws Exception {
		int cacheSize = -2;
		String autoCleanSize = null;
		String autoCleanThreshold = null;
		String policy = null;
		//configure the cache size
		StringBuffer sb = new StringBuffer(cacheLabel);
		final String sep = ", ";
		boolean conf = false;
		String strval = getProperty(props, sizeProp, "");
		if (strval.length()>0) {
			props.remove(sizeProp);
			if (strval.equalsIgnoreCase("nolimit")) {
				cacheSize = CacheAttributeController.NO_LIMIT;
			} else {
				cacheSize = Integer.parseInt(strval);
			}
			if (conf) {
			    sb.append(sep);
			}
			sb.append("max size=" + strval);
			conf = true;
		}

		//configure the cache auto clean size
		strval = getProperty(props, autoCleanSizeProp, "");
		if (strval.length() > 0) {
			props.remove(autoCleanSizeProp);
			autoCleanSize = strval;
			if (conf) {
			    sb.append(sep);
			}
			sb.append("auto replacement size=" + strval);
			conf = true;
		}

		//configure the cache auto clean size
		strval = getProperty(props, autoCleanThresholdProp, "");
		if (strval.length() > 0) {
			props.remove(autoCleanThresholdProp);
			autoCleanThreshold = strval;
			if (conf) {
			    sb.append(sep);
			}
			sb.append("auto replacement threshold=" + strval);
			conf = true;
		}

		//configure the replacement policy of the cache (Replacement manager)
		strval = getProperty(props, policyProp, "");
		if (strval.length() > 0) {
			props.remove(policyProp);
			policy = strval;
			if (conf) {
			    sb.append(sep);
			}
			sb.append("replacement policy=" + strval);
			conf = true;
		}
		if (!conf) {
		    return;
		}
	    logger.log(BasicLevel.INFO, sb.toString());
		Component cmC = getSubComponent(cm, "cache-manager");
		CacheAttributeController cacheAttr = (CacheAttributeController)
			Fractal.getAttributeController(cmC);
		if (cacheSize > -2) {
			cacheAttr.setMaxObjects(cacheSize);
		}
		if (autoCleanSize != null) {
			cacheAttr.setAutoCleanSize(autoCleanSize);
		}
		if (autoCleanThreshold != null) {
			cacheAttr.setAutoCleanThreshold(autoCleanThreshold);
		}
		if (policy != null) {
			String tempName = null;
			if (SpeedoProperties.CACHE_REPLCAEMENT_LRU.equalsIgnoreCase(policy)) {
			} else if (SpeedoProperties.CACHE_REPLCAEMENT_MRU.equalsIgnoreCase(policy)) {
				tempName = "org.objectweb.perseus.cache.replacement.lib.MRUReplacementManager";
			} else if (SpeedoProperties.CACHE_REPLCAEMENT_FIFO.equalsIgnoreCase(policy)) {
				tempName = "org.objectweb.perseus.cache.replacement.lib.FIFOReplacementManager";
			} else {
				logger.log(BasicLevel.ERROR, new SpeedoException(
						"Unmanaged cache replacement policy: " + policy));
			}
			if (tempName != null) {
				//unbind the rm from the composite
				Fractal.getBindingController(cm).unbindFc("replacement-manager");
				//unbind the rm from the primitive cm
				Fractal.getBindingController(cmC).unbindFc("replacement-manager");
				//unbind the primitive cm from the old rm
				Fractal.getBindingController(
						getSubComponent(cm, "replacement-manager"))
						.unbindFc("unbind-manager");

				//remove the old rm from the composite
				Fractal.getContentController(cm).removeFcSubComponent(
						getSubComponent(cm, "replacement-manager"));

				//instanciate the new component
				Component new_rm = (Component)getADLFactory().newComponent(tempName, null);
				Fractal.getNameController(new_rm).setFcName("replacement-manager");

				//Add the new rm in the composite
				Fractal.getContentController(cm).addFcSubComponent(new_rm);

				Object rm = new_rm.getFcInterface("replacement-manager");
				//bind the rm to the composite
				Fractal.getBindingController(cm).bindFc("replacement-manager", rm);
				//bind the rm to the primitive cm
				Fractal.getBindingController(cmC).bindFc("replacement-manager", rm);
				//bind the primitive cm to the new rm
				Fractal.getBindingController(new_rm).bindFc("unbind-manager",
						cmC.getFcInterface("unbind-manager"));
			}
		}
	}


	private void configureConcurrencyManager(Component speedo,
											 Map props)
			throws Exception {
		String tempName;
		String concurrencyMode;
		StringBuffer sb = new StringBuffer("Transaction/Concurrency management: ");
        String strval = getProperty(props, SpeedoProperties.JDO_OPTION_OPTIMISTIC, "false");
        boolean optimistic = isTrue(strval);
        sb.append("mode=").append((optimistic ? "Optimistic" : "Pessimitic"));
	    strval = getProperty(props, SpeedoProperties.TRANSACTION_LOCKING,
	            SpeedoProperties.SPEEDO_TRANSACTION_LOCKING);
	    props.remove(SpeedoProperties.TRANSACTION_LOCKING);
		//fetch the composite TPM
		Component composite_tpm = getSubComponent(speedo, "tpm");
		Component cm = getSubComponent(composite_tpm, "concurrency-manager");
	    if (strval.equals(SpeedoProperties.DB_TRANSACTION_LOCKING)) {
	        tempName = DBDELEGATE_CONCURRENCY_TEMPLATE;
	        sb.append(", delegated to database");
	    } else {
	        sb.append(", managed by Speedo");
	        if (!optimistic) {
	            //Configure the pessimistic concurrency manager
		        
	    	    //policy= mutex | rw-fifo | rw-reader
	    	    strval = getProperty(props, 
		                SpeedoProperties.TRANSACTION_LOCKING_PESSIMISTIC_POLICY, 
		                SpeedoProperties.TRANSACTION_LOCKING_PESSIMISTIC_POLICY_RW_FIFO);
		        props.remove(SpeedoProperties.TRANSACTION_LOCKING_PESSIMISTIC_POLICY);
		        PessimisticConcurrencyManagerAC pcmAC = null;
		        if (!strval.equals(SpeedoProperties.TRANSACTION_LOCKING_PESSIMISTIC_POLICY_RW_FIFO)) {
	    	        pcmAC = (PessimisticConcurrencyManagerAC) 
	    	            Fractal.getAttributeController(cm);
	    	        pcmAC.setPolicy(strval);
	    	        sb.append(", locking policy=").append(strval);
	    	    }

		        //locking level = instance | field
		        strval = getProperty(props, 
		                SpeedoProperties.TRANSACTION_LOCKING_LEVEL_ENABLETHIN, 
		                "false");
		        props.remove(SpeedoProperties.TRANSACTION_LOCKING_LEVEL_ENABLETHIN);
		        boolean b = isTrue(strval);
		        if (b) {
			        logger.log(BasicLevel.INFO, "Speedo pessimistic thin locking level: enabled");
	    	        pcmAC = (PessimisticConcurrencyManagerAC) 
	    	            Fractal.getAttributeController(cm);
	    	        pcmAC.setThinkLockAllowed(true);
	    	        sb.append(", With thin locking level");
	    	    }
			    logger.log(BasicLevel.INFO, sb.toString());
	            return;
	        }
			tempName = OPTIMISTIC_CONCURRENCY_TEMPLATE;
	    }
	    logger.log(BasicLevel.INFO, sb.toString());
        props.remove(SpeedoProperties.JDO_OPTION_OPTIMISTIC);

        ContentController tpmCC = Fractal.getContentController(composite_tpm);
		//fetch the primitive concurrency manager
		BindingController tpmBC = Fractal.getBindingController(
				getSubComponent(composite_tpm, "transactional-persistence-manager"));
		tpmBC.unbindFc("concurrency-manager");

		//Remove the old concurrency manager
		BindingController cmBC = Fractal.getBindingController(cm);
		cmBC.unbindFc("state-manager");
		cmBC.unbindFc("storage-manager");
		cmBC.unbindFc("dependency-graph");
		tpmCC.removeFcSubComponent(cm);


		//instanciate the new ConcurrencyManager
		cm = (Component)getADLFactory().newComponent(tempName, null);
		//Add the new concurrency manager into the composite
		tpmCC.addFcSubComponent(cm);
		//bind the server interface of the new concurrency manager
		cmBC = Fractal.getBindingController(cm);
		cmBC.bindFc("state-manager",
				tpmCC.getFcInternalInterface("state-manager"));
		cmBC.bindFc("storage-manager",
				tpmCC.getFcInternalInterface("storage-manager"));
		cmBC.bindFc("dependency-graph",
				getSubComponent(composite_tpm, "dependency-graph").getFcInterface("dependency-graph"));

		//bind the client interface of the new ConcurrencyManager
		tpmBC.bindFc("concurrency-manager", cm.getFcInterface("concurrency-manager"));
	}

	private void configurePrefetching(Component speedo,
									  Map props) throws Throwable {
		//Data prefetching on queries
		boolean prefetch = true;
		StringBuffer sb = new StringBuffer("Data prefetching: ");
		String strval = getProperty(props, SpeedoProperties.PREFETCH, "");
		if (strval.length()>0) {
			prefetch = !isFalse(strval);
			props.remove(SpeedoProperties.PREFETCH);
		}
		Component qm = getSubComponent(speedo, QUERY_MANAGER);
		QueryManagerAttribute qma = ((QueryManagerAttribute) Fractal.getAttributeController(qm));
		if (!prefetch) {
			qma.setPrefetchActivatedOnQuery(false);
			qma.setPrefetchActivatedOnExtent(false);
			logger.log(BasicLevel.INFO, sb.append("NONE").toString());
		    return;
		}
		boolean prefetchOnQuery = qma.getPrefetchActivatedOnQuery();
		boolean prefetchOnExtent = qma.getPrefetchActivatedOnExtent();
		boolean prefetchOnGenClass = true;
		String sep = ", ";

		boolean oneOf = false;
	    strval = getProperty(props, SpeedoProperties.PREFETCH_ON_QUERY, "");
		if (strval.length()>0) {
			prefetchOnQuery = !isFalse(strval);
			props.remove(SpeedoProperties.PREFETCH_ON_QUERY);
		}
		if (prefetchOnQuery) {
			if (oneOf) {
			    sb.append(sep);
			}
		    sb.append("QUERY");
		}
		oneOf |= prefetchOnQuery;
		qma.setPrefetchActivatedOnQuery(prefetchOnQuery);

		strval = getProperty(props, SpeedoProperties.PREFETCH_ON_EXTENT, "");
		if (strval.length()>0) {
			prefetchOnExtent = !isFalse(strval);
			props.remove(SpeedoProperties.PREFETCH_ON_EXTENT);
		}
		if (prefetchOnExtent) {
			if (oneOf) {
			    sb.append(sep);
			}
		    sb.append("EXTENT");
		}
		oneOf |= prefetchOnExtent;
		qma.setPrefetchActivatedOnExtent(prefetchOnExtent);

		strval = getProperty(props, SpeedoProperties.PREFETCH_ON_GENCLASS, "");
		if (strval.length()>0) {
			prefetchOnGenClass= !isFalse(strval);
			props.remove(SpeedoProperties.PREFETCH_ON_GENCLASS);
		}
		if (prefetchOnGenClass) {
			if (oneOf) {
			    sb.append(sep);
			}
		    sb.append("COLLECTION");
		}
		oneOf |= prefetchOnGenClass;
		
		logger.log(BasicLevel.INFO, sb.toString());
	}

	/**
	 * Configures the mapper component
	 * @param speedo is the Speedo component
	 * @param props is the properties of Speedo
	 * @return true if a connection factory is used, false if the JDBC driver is
	 * used directly.
	 * @throws Throwable if an error or an exception occurs
	 */
	private boolean configureMapper(Component speedo,
								 Map props) throws Throwable {
		boolean useConnectionFactory;
		String mapperName = getProperty(props, SpeedoProperties.MAPPER_NAME, "rdb.automatic");
		props.remove(SpeedoProperties.MAPPER_NAME);

		Component component = getSubComponent(speedo, PRIMITIVE_MAPPER_PATH);
		String strval = null;
		//Choose the mapper in according to the connection factory (JCA or JDBC)
		String cfName = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY_NAME, null);
		props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY_NAME);
		useConnectionFactory = cfName != null && cfName.length() > 0;
		MapperAttributes ma = (MapperAttributes) Fractal.getAttributeController(component);
		if (useConnectionFactory) {
			Object cf = null;
			//There is a connection factory
			try {
				InitialContext ic = new InitialContext();
				cf = ic.lookup(cfName);
			} catch (NamingException e) {
				throw new SpeedoException(
						"Problem to get the connection factory in JNDI ("
						+ cfName + ")", e);
			}
			if (cf == null) {
				throw new SpeedoException(
						"No connection factory registered in JNDI with the name "
						+ cfName);
			} else if (cf instanceof ConnectionSpecJDBC) {
				// Use the JDBC Mapper ==> nothing to do
			} else if (cf instanceof javax.sql.DataSource) {
				// Use the JDBC Mapper ==> nothing to do
			} else if (cf instanceof javax.resource.cci.ConnectionFactory) {
				System.out.println("The connection is JCA compliant: " + cf);
				throw new SpeedoException("JCA datasource no yet supported");
			} else {
				throw new SpeedoException("The connection factory registered "
						+ "in JNDI is not supported by Speedo " + cfName + " => "
						+ cf);
			}
			boolean ignoring = removeProps(props, new String[]{
			        SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old,
			        SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old2,
			        SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME,
			        SpeedoProperties.JDO_OPTION_CONNECTION_URL,
			        SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME,
			        SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD});
			if (ignoring) {
			    logger.log(BasicLevel.WARN, 
			            "Ignoring direct JDBC access information (driver, url, user, password)");
			}
			ignoring = removeProps(props, new String[]{
			        SpeedoProperties.CONNECTION_POOL_MIN,
			        SpeedoProperties.CONNECTION_POOL_MAX,
			        SpeedoProperties.CONNECTION_POOL_TTL,
			        SpeedoProperties.CONNECTION_POOL_TIMEOUT});
			if (ignoring) {
			    logger.log(BasicLevel.WARN, 
			            "Ignoring connection pool configuration(min, max, ttl, timeout)");
			}

			//Assign the connection factory to the mapper
			logger.log(BasicLevel.INFO, "Data source access through a connection factory found in JNDI:"
					+ "\n\tjndi name: " + cfName
					+ "\n\tconnection factory found: " + cf
					+ "\n\tmapper name: " + mapperName);
			PMapper mapper = (PMapper) component.getFcInterface("mapper");
			mapper.setMapperName(mapperName);
			mapper.setConnectionFactory(cf);

		} else {
			JDBCMapperAttributes jdbcma = (JDBCMapperAttributes) ma;

			//deprecated properties
			strval = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old, "");
			if (strval.length()>0) {
				logger.log(BasicLevel.WARN, "Property " + SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old
						+ " is deprecated, you must use " + SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME);
				props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old);
				
				if (getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME) == null) {
                    props.put(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME, strval);
				}
			}
			strval = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old2, "");
			if (strval.length()>0) {
				logger.log(BasicLevel.WARN, "Property " + SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old2
						+ " is deprecated, you must use " + SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME);
				props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old);
				
				if (getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME) == null) {
                    props.put(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME, strval);
				}
			}
			String jdbcDriverCN = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME);
			String jdbcUser = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME);
			String jdbcUrl = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_URL, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_URL);
			String jdbcPass = getProperty(props, SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD);
			logger.log(BasicLevel.INFO, "Data source access through a jdbc driver:"
					+ "\n\tdriver= " + jdbcDriverCN
					+ "\n\turl= " + jdbcUrl
					+ "\n\tuser= " + jdbcUser
					+ "\n\tmapper name: " + mapperName);
			jdbcma.setMapperName(mapperName);
			jdbcma.setDriverClassName(jdbcDriverCN);
			jdbcma.setURL(jdbcUrl);
			jdbcma.setUserName(jdbcUser);
			jdbcma.setPassword(jdbcPass);
			jdbcma.setPoolConnection(true);

			//configure the pool of connection
			configurePool(getSubComponent(speedo, CONNECTION_POOL_PATH),
			        "Connection pool: ",
					SpeedoProperties.CONNECTION_POOL_MIN,
					SpeedoProperties.CONNECTION_POOL_MAX,
					SpeedoProperties.CONNECTION_POOL_TTL,
					SpeedoProperties.CONNECTION_POOL_INACTIVETTL,
					SpeedoProperties.CONNECTION_POOL_TIMEOUT,
					props);
		}
		strval = (String) props.remove(SpeedoProperties.CONNECTION_CHECK);
		if (strval != null) {
		    boolean v = isTrue(strval);
		    ma.setCheckConnectivityAtStartup(v);
		    if (!v) {
				logger.log(BasicLevel.INFO, "No connection checking at start up !");
		    }
		}		

		//initialize the JormFactory with the mappingStructureRule attribute
		byte mappingStructureRule = JormFactoryAttributes.CREATE_IF_REQUIRED;
		strval = getProperty(props, SpeedoProperties.MAPPING_STRUCTURE, "");
		if (strval.length()>0) {
			props.remove(SpeedoProperties.MAPPING_STRUCTURE);
			mappingStructureRule = getMappingStructure(strval);
			if (mappingStructureRule == -1) {
			    mappingStructureRule = JormFactoryAttributes.CREATE_IF_REQUIRED;
				logger.log(BasicLevel.WARN, "Unexpected value for the property '"
						+ SpeedoProperties.MAPPING_STRUCTURE
						+ "', found: '" + strval + "', expected ["
						+ SpeedoProperties.MAPPING_STRUCTURE_DN
						+ " | " + SpeedoProperties.MAPPING_STRUCTURE_DD
						+ " | " + SpeedoProperties.MAPPING_STRUCTURE_CIR
						+ " | " + SpeedoProperties.MAPPING_STRUCTURE_FC
						+ "]");
			}
		}
		component = getSubComponent(speedo, JORM_FACTORY_PATH);
		JormFactoryAttributes jfa = (JormFactoryAttributes)
                    Fractal.getAttributeController(component);
		jfa.setMappingStructureRule(mappingStructureRule);
		logger.log(BasicLevel.INFO, "Data Structure: "
				+ getMappingStructureString(mappingStructureRule));

		Properties jormFactoryProperties = new Properties();
        for (Iterator it = props.entrySet().iterator(); it.hasNext();) {
            Map.Entry me = (Map.Entry) it.next();
            String key = (String) me.getKey();
            String value = (String) me.getValue();
			boolean knownProperty = true;
        	if (key.indexOf('(') != -1) {
        	    jormFactoryProperties.put(key, value);
            } else {
				knownProperty = false;
            }
			if (knownProperty) {
				logger.log(BasicLevel.INFO, key + ": " + value);
				it.remove();
			}
        }
        jfa.setSpeedoProperties(jormFactoryProperties);
		
		return useConnectionFactory;
	}

	private boolean configureJMX(Component speedo,
			 Map props) throws Throwable {
	    String val = (String) props.remove(SpeedoProperties.JMX);
	    if (val == null || !isTrue(val)) {
	        return false;
	    }
	    //JMX must me configured
	    if (Monolog.monologFactory.getLogger(SpeedoProperties.JMX)
	            .isLoggable(BasicLevel.DEBUG)) {
	        //active MX4J logging
			System.setProperty("mx4j.log.priority", "debug");
	    }
	    val = (String) props.remove(SpeedoProperties.JMX_HTTP_PORT);
	    int port = 0;
	    if (val != null) {
	        try {
                port = Integer.parseInt(val);
            } catch (NumberFormatException e) {
                logger.log(BasicLevel.WARN, "Bad HTTP port for the administration console: " + val);
                port = -1;
            }
	    }
		Component component = getSubComponent(speedo, HTML_JMX_AGENT_PATH);
	    if (port < 0) {
	        //unbind the html sub component
	        Fractal.getBindingController(component).unbindFc("adminAtt");
	        // remove the html sub component from the agent component.
			Fractal.getContentController(
			        getSubComponent(speedo, JMX_AGENT_PATH))
			        	.removeFcSubComponent(component);
			logger.log(BasicLevel.INFO, "JMX activated");
	    } else {
	        if (port == 0) {
		        //TODO: find a port
	            port = 8000;
	        }
	        //Assign the port to the component
			CommunicatorAttributes att = (CommunicatorAttributes) 
				Fractal.getAttributeController(component);
	        att.setPort(port);
			logger.log(BasicLevel.INFO, "JMX and HTTP console on port '" +  port + "' activated");
	    }
	    return true;
	}
	
	private boolean removeProps(Map props, String[] wanted) {
	    boolean found = false;
	    for (int i = 0; i < wanted.length; i++) {
            found |= props.remove(wanted[i]) != null;
        }
	    return found;
	}
	
    private boolean isFalse(String val) {
        return "false".equalsIgnoreCase(val)
                || "off".equalsIgnoreCase(val)
                || "no".equalsIgnoreCase(val)
                || "non".equalsIgnoreCase(val)
                || "n".equalsIgnoreCase(val);
    }

    private boolean isTrue(String val) {
        return "true".equalsIgnoreCase(val)
                || "on".equalsIgnoreCase(val)
                || "yes".equalsIgnoreCase(val)
                || "oui".equalsIgnoreCase(val)
                || "y".equalsIgnoreCase(val)
                || "o".equalsIgnoreCase(val);
    }

	public static String getMappingStructureString(byte b) {
		switch(b) {
		case JormFactoryAttributes.DO_NOTHING:
			return SpeedoProperties.MAPPING_STRUCTURE_DN;
		case JormFactoryAttributes.DELETE_DATA:
			return SpeedoProperties.MAPPING_STRUCTURE_DD;
		case JormFactoryAttributes.FORCE_CREATE:
			return SpeedoProperties.MAPPING_STRUCTURE_FC;
		case JormFactoryAttributes.CREATE_IF_REQUIRED:
			return SpeedoProperties.MAPPING_STRUCTURE_CIR;
		default:
			return "UNKNOWN";
		}
	}
	
	public static byte getMappingStructure(String strval) {
		if (SpeedoProperties.MAPPING_STRUCTURE_DN.equals(strval)) {
			return JormFactoryAttributes.DO_NOTHING;
		} else if (SpeedoProperties.MAPPING_STRUCTURE_DD.equals(strval)) {
			return JormFactoryAttributes.DELETE_DATA;
		} else if (SpeedoProperties.MAPPING_STRUCTURE_CIR.equals(strval)) {
			return JormFactoryAttributes.CREATE_IF_REQUIRED;
		} else if (SpeedoProperties.MAPPING_STRUCTURE_FC.equals(strval)) {
			return JormFactoryAttributes.FORCE_CREATE;
		} else {
		    return -1;
		}
	    
	}

	public Object getConcurrencyManagerComponent() throws Exception {
		return getSubComponent(speedo, "tpm.concurrency-manager");
	}
	
	public DependencyGraph getDependencyGraph() throws Exception {
	    Component cmC = (Component) getConcurrencyManagerComponent();
	    Object dg = Fractal.getBindingController(cmC).lookupFc("dependency-graph");
	    return (DependencyGraph) dg;
	}
	
	public Collection getMemoryCacheEntries() throws Exception {
	    Component cm = getSubComponent(speedo, MEMORY_CACHE_PATH);
		Component cmC = getSubComponent(cm, "cache-manager");
		CacheAttributeController cacheAttr = (CacheAttributeController)
			Fractal.getAttributeController(cmC);
		return cacheAttr.getCurrentEntryIdentifiers();
	}
	
    // IMPLEMENTATION OF THE PersistenceManagerFactory INTERFACE //
    //-----------------------------------------------------------//

    public PersistenceManager getPersistenceManager() {
    	if (!isPropertiesInitialized) {
    		try {
    		    Properties p = getProperties();
    			init(getSpeedoComponent(p), p);
    		} catch(Throwable e) {
    			System.err.println("Error during the initilization of the Speedo " +
    			"persistence manager factory:");
    			e.printStackTrace(System.err);
    			return null;
    		}    		
    	}
    	
        return delegate.getPersistenceManager();
    }

    public PersistenceManager getPersistenceManager(String s, String s1) {
    	if (!isPropertiesInitialized) {
    		try {
    		    Properties p = getProperties();
    			init(getSpeedoComponent(p), p);
    		} catch(Throwable e) {
    			System.err.println("Error during the initilization of the Speedo " +
    			"persistence manager factory:");
    			e.printStackTrace(System.err);
    			return null;
    		}    		
    	}
    	return delegate.getPersistenceManager(s, s1);
    }

    public void setConnectionUserName(String s) {
        delegate.setConnectionUserName(s);
    }

    public String getConnectionUserName() {
        return delegate.getConnectionUserName();
    }

    public void setConnectionPassword(String s) {
        delegate.setConnectionPassword(s);
    }

    public void setConnectionURL(String s) {
        delegate.setConnectionURL(s);
    }

    public String getConnectionURL() {
        return delegate.getConnectionURL();
    }

    public void setConnectionDriverName(String s) {
        delegate.setConnectionDriverName(s);
    }

    public String getConnectionDriverName() {
        return delegate.getConnectionDriverName();
    }

    public void setConnectionFactoryName(String s) {
        delegate.setConnectionFactoryName(s);
    }

    public String getConnectionFactoryName() {
        return delegate.getConnectionFactoryName();
    }

    public void setConnectionFactory(Object o) {
        delegate.setConnectionFactory(o);
    }

    public Object getConnectionFactory() {
        return delegate.getConnectionFactory();
    }

    public void setConnectionFactory2Name(String s) {
        delegate.setConnectionFactory2Name(s);
    }

    public String getConnectionFactory2Name() {
        return delegate.getConnectionFactory2Name();
    }

    public void setConnectionFactory2(Object o) {
        delegate.setConnectionFactory2(o);
    }

    public Object getConnectionFactory2() {
        return delegate.getConnectionFactory2();
    }

    public void setMultithreaded(boolean b) {
        delegate.setMultithreaded(b);
    }

    public boolean getMultithreaded() {
        return delegate.getMultithreaded();
    }

    public void setOptimistic(boolean b) {
        delegate.setOptimistic(b);
    }

    public boolean getOptimistic() {
        return delegate.getOptimistic();
    }

    public void setRetainValues(boolean b) {
        delegate.setRetainValues(b);
    }

    public boolean getRetainValues() {
        return delegate.getRetainValues();
    }

    public void setRestoreValues(boolean b) {
        delegate.setRestoreValues(b);
    }

    public boolean getRestoreValues() {
        return delegate.getRestoreValues();
    }

    public void setNontransactionalRead(boolean b) {
        delegate.setNontransactionalRead(b);
    }

    public boolean getNontransactionalRead() {
        return delegate.getNontransactionalRead();
    }

    public void setNontransactionalWrite(boolean b) {
        delegate.setNontransactionalWrite(b);
    }

    public boolean getNontransactionalWrite() {
        return delegate.getNontransactionalWrite();
    }

    public void setIgnoreCache(boolean b) {
        delegate.setIgnoreCache(b);
    }

    public boolean getIgnoreCache() {
        return delegate.getIgnoreCache();
    }

    public Properties getProperties() {
        return delegate.getProperties();
    }

    public Collection supportedOptions() {
        return delegate.supportedOptions();
    }

    public PersistenceManagerFactory getDelegate() {
        return delegate;
    }

	public ProxyManager lookup() {
		return pmfComponent.lookup();
	}

	public SequenceManager getSequenceManager() {
		return pmfComponent.getSequenceManager();
	}
	
	public void setSequenceManager(SequenceManager sequenceManager) {
		pmfComponent.setSequenceManager(sequenceManager);
	}
	
	public void unbindPM() {
		pmfComponent.unbindPM();
	}

	public void bindPM2Thread(ProxyManager pm) {
		pmfComponent.bindPM2Thread(pm);
	}

	public void proxyManagerClosed(ProxyManager pr) {
		pmfComponent.proxyManagerClosed(pr);
	}

	public void close() {
		pmfComponent.close();
	}
	
    public DataStoreCache getDataStoreCache() {
        return pmfComponent.getDataStoreCache();
    }

    public String getMapping() {
        return pmfComponent.getMapping();
    }

    public boolean isClosed() {
        return pmfComponent.isClosed();
    }

    public void setMapping(String arg0) {
        pmfComponent.setMapping(arg0);
    }

    public void evict(Object arg0) {
        pmfComponent.evict(arg0);
    }

    public void evictAll() {
        pmfComponent.evictAll();
    }
    
    public void evictAll(Class arg0, boolean arg1) {
        pmfComponent.evictAll(arg0, arg1);
    }

    public void evictAll(Collection arg0) {
        pmfComponent.evictAll(arg0);
    }

    public void evictAll(Object[] arg0) {
        pmfComponent.evictAll(arg0);
    }
    public void pin(Object arg0) {
        pmfComponent.pin(arg0);
    }

    public void pinAll(Class arg0, boolean arg1) {
        pmfComponent.pinAll(arg0, arg1);
    }

    public void pinAll(Collection arg0) {
        pmfComponent.pinAll(arg0);
    }

    public void pinAll(Object[] arg0) {
        pmfComponent.pinAll(arg0);
    }

    public void unpin(Object arg0) {
        pmfComponent.unpin(arg0);
    }

    public void unpinAll(Class arg0, boolean arg1) {
        pmfComponent.unpinAll(arg0, arg1);
    }

    public void unpinAll(Collection arg0) {
        pmfComponent.unpinAll(arg0);
    }

    public void unpinAll(Object[] arg0) {
        pmfComponent.unpinAll(arg0);
    }
    
    public void addInstanceLifecycleListener(
            InstanceLifecycleListener arg0,
            Class[] arg1) {
        pmfComponent.addInstanceLifecycleListener(arg0, arg1);
    }
    public void removeInstanceLifecycleListener(InstanceLifecycleListener arg0) {
        pmfComponent.removeInstanceLifecycleListener(arg0);
    }
}
