/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.generation.enhancer;

import org.objectweb.jorm.lib.PBindingImpl;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoPackage;
import org.objectweb.speedo.metadata.SpeedoXMLDescriptor;
import org.objectweb.speedo.metadata.SpeedoIdentity;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.attrs.Attributes;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.util.List;
import java.util.TreeSet;
import java.util.Set;

/**
 * Enhances a set of Java classes and renames it with <code>Delegate</code> as
 * suffix. <p>The modifications to be done are described by the JDO
 * Implementation's rules.</p>
 *
 * @author	S.Chassande-Barrioz
 */
public class PersistentClassEnhancer extends EnhancerComponent {

    public final static String LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".generation.enhancer";

    /**
     * Initializes this PersistentClassEnhancer
     */
    public boolean init() {
        logger = scp.loggerFactory.getLogger(LOGGER_NAME);
        //TODO: manage classes localized in a jar file
        isSrcJar = false;
	    return !scp.getXmldescriptor().isEmpty();
    }

    /**
     * Loads all binary classes described by the Object Model and applies
     * revelant modification to each of them.
     *
     * @exception org.objectweb.speedo.generation.enhancer.SpeedoEnhancerException if something goes wrong
     */
    public void process() throws SpeedoException {
        if (scp.getXmldescriptor().isEmpty())
            return;

        ArrayList except = new ArrayList();
        Collection xmls = scp.getXmldescriptor().values();
		Logger log = null;
        for (Iterator itDesc = xmls.iterator(); itDesc.hasNext();) {
            SpeedoXMLDescriptor desc = (SpeedoXMLDescriptor) itDesc.next();
            for (Iterator itPack = desc.jdoPackage.values().iterator(); itPack.hasNext();) {
                SpeedoPackage sp = (SpeedoPackage) itPack.next();
                for (Iterator itclass = sp.jdoClass.values().iterator(); itclass.hasNext();) {
                    SpeedoClass jdoclass = (SpeedoClass) itclass.next();

                    String name = sp.name + '.' + jdoclass.name;
                    String proxyName = NamingRules.proxyName(name);

                    ClassWriter cw = new ClassWriter(true);

                    RedundencyRemover rr = new RedundencyRemover(cw, logger);

                    List itfs = new ArrayList();

                    // merges the Binding class
	                Set oldClassNames = new TreeSet();
	                oldClassNames.add(name.replace('.', '/'));
	                ClassReader cr;
	                MergedClassModifier mmodifier;
	                ClassRenamer crenamer;
	                if (jdoclass.superClassName == null) {
	                    String bindingName = PBindingImpl.class.getName();
						log = scp.loggerFactory.getLogger(LOGGER_NAME + ".merge.binding." + name);
						log.log(BasicLevel.DEBUG, "Merging the Binding of class: " + name);
	                    cr = loadJavaClass(false, bindingName, scp.output, true);
                        mmodifier = new MergedClassModifier(rr, itfs, log);
                        oldClassNames.add(bindingName.replace('.', '/'));
                        crenamer = new ClassRenamer(mmodifier,
                                oldClassNames, name.replace('.', '/'), log);
                        cr.accept(crenamer, Attributes.getDefaultAttributes(), false);
	                }
                    // merges the Proxy class
					log = scp.loggerFactory.getLogger(LOGGER_NAME + ".merge.proxy." + name);
					log.log(BasicLevel.DEBUG, "Merging the Proxy of class: " + name);
                    String superClass = jdoclass.superClassName;
                    cr = loadJavaClass(false, proxyName, scp.output, true);
                    if (superClass == null || superClass.length() == 0) {
                        mmodifier = new MergedClassModifier(rr, itfs, log);
                    } else {
                        mmodifier = new MergedClassModifier(rr, itfs,
                                superClass.replace('.', '/'), log);
                    }
                    crenamer = new ClassRenamer(mmodifier, oldClassNames,
                            name.replace('.', '/'), logger);
					cr.accept(crenamer, Attributes.getDefaultAttributes(), false);

                    // modifies the original code
					log = scp.loggerFactory.getLogger(LOGGER_NAME + ".fieldaccess." + name);
					log.log(BasicLevel.DEBUG, "Modifying the field access of class: " + name);
                    cr = loadJavaClass(false, name, scp.output, false);
                    ClassAccessorModifier caccessmodifier =
                            new ClassAccessorModifier(rr, xmls, log);
                    ClassInterfaceAdder citfadder = new ClassInterfaceAdder(
                            caccessmodifier, itfs, log);
					JDOImplRegistrationAdder reg = new JDOImplRegistrationAdder(citfadder, jdoclass, logger);
                    cr.accept(reg, Attributes.getDefaultAttributes(), false);
                    writeJavaClass(name, cw, scp.output);

                    if (jdoclass.identityType == SpeedoIdentity.USER_ID
                        && jdoclass.objectidClass != null
                        && (jdoclass.superClassName == null //I have not a parent 
                                ||  !jdoclass.objectidClass.equals(jdoclass.getSuper().objectidClass))
                                ) {
						log = scp.loggerFactory.getLogger(LOGGER_NAME + ".userid." + name);
                        cw = new ClassWriter(true);
                        String cn = jdoclass.objectidClass;
                        cr = loadJavaClass(false, cn, scp.output, true);
                        
                        cr.accept(new UserIdEnhancer(cw, jdoclass, log),
                                Attributes.getDefaultAttributes(), false);
                        writeJavaClass(cn, cw, scp.output);

                        cw = new ClassWriter(false);
                        cn = NamingRules.pnameName(jdoclass.objectidClass);
                        cr = loadJavaClass(false, cn, scp.output, true);
                        cr.accept(new PNameEnhancer(cw, jdoclass, log), 
                                Attributes.getDefaultAttributes(), false);
                        writeJavaClass(cn, cw, scp.output);
                    }
                }
            }
        }

        // Displays thrown exceptions
        if (!except.isEmpty() && logger.isLoggable(BasicLevel.WARN)) {
            for (Iterator it = except.iterator(); it.hasNext();) {
                Exception e = (Exception) it.next();
                logger.log(BasicLevel.WARN, e.getLocalizedMessage());
            }
        }
    }
}
