/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.metadata;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.jdo.JDOUserException;

/**
 * Describes a persistence capable class.
 * @author	S.Chassande-Barrioz
 */
public class SpeedoClass extends SpeedoElement {
    /**
     * Class name.
     */
    public String name;

    /**
     * Attribute identity-type.
     */
    public byte identityType = SpeedoIdentity.CONTAINER_ID;

	/**
	 * Description of the version.
	 */
	public SpeedoVersion version;
	
	/**
	 * The datastore-id.
	 */
	public SpeedoDatastoreId datastoreId;
	
    /**
     * Attribute objectid-class.
     */
    public String objectidClass;

    /**
     * Attribute requires-extent.
     */
    public boolean requiresExtent = true;

	/**
	 * Attribute detachable.
	 */
	public boolean isDetachable = false;
		
    /**
     * Super class name.
     */
    public String superClassName = "";

    /**
     * Description of persistent capable fields of this class. The HashMap key
     * is the field name.
     */
    public Map jdoField = new HashMap();
    
    /**
     * Description of fetchgroups defined for this class. The HashMap key
     * is the field name.
     */
    public Map jdoFetchGroup = new HashMap();

    /**
     * Descriptor of this class code.
     */
    public ClassCode classCode = new ClassCode();

    /**
     * Description of the package which contains this class.
     */
    public SpeedoPackage jdoPackage;

    /**
     * Indicates if the class is abstract.
     */
    public boolean isAbstract = false;

    /**
     * Indicates if the class implements InstanceCallbacks
     */
    public boolean isInstanceCallbacks = false;

    /**
     * Indicates if the class implements InstanceCallbacks
     */
    public boolean isSerializable = false;

    /**
     * Class signature.
     */
    public String signature;

    /**
     * Version UID for serialization.
     */
    public long VersionUID = 0;

    /**
     * Indicates whether something goes wrong during the JDO Instances generation.
     */
    public boolean failed = false;

    /**
     * Class jorm corresponding to this jdoclass
     */
    public Class jormclass;

    /** 
     * The name of the datastore sequence
     */
    public String datastoreSequence;
    
    /**
     * Number of fields of this class (including inherited fields).
     */
    private int fieldsCount = -1;

    private int pkFieldsCount = -1;

	public boolean classAlreadyEnhanced = false;
	public boolean requireEnhancement = true;
	
	public Map name2query = new HashMap();

	public final static byte NO_NO_ARG_CONSTRUCTOR = 1;
	public final static byte NON_PUBLIC_NO_ARG_CONSTRUCTOR = 2;
	public final static byte PUBLIC_NO_ARG_CONSTRUCTOR = 3;

	public byte noArgConstructorStatus = NO_NO_ARG_CONSTRUCTOR;
	
	public final static byte SUBCLASS_TABLE = 1;
	public final static byte NEW_TABLE = 2;
	public final static byte SUPERCLASS_TABLE = 3;
	
	public byte inheritanceStrategy = 0;

    public String getFQName() {
        if (jdoPackage.name == null || jdoPackage.name.length() == 0) {
            return name;
        } else {
            return jdoPackage.name + "." + name;
        }
    }

    /**
     * Transforms a SpeedoClass into a String.
     * @return the Sting corresponding to the SpeedoClass.
     */
    public String toString() {
        String s = ("\n class name : " + name +
                ", identityType : " + identityType +
                ", objectidClass : " + objectidClass +
                ", requiresExtent : " + requiresExtent +
                ",version : " + version +
                ", persistenceCapableSuperClass : " + superClassName);
        s += ", \t fields:[";
        Iterator it = jdoField.values().iterator();
        while (it.hasNext()) {
            s = s + "\t" + it.next().toString();
        }
        s += "], \t fetchGroups:[";
		Iterator it2 = jdoFetchGroup.values().iterator();
			while (it2.hasNext()) {
				s += "\t" + it2.next().toString();
		}
		s += "]";
		return s;
    }

    /**
     * Adds a SpeedoField to the class.
     * Precondition: this field doesn't exist in the jdoFields HashMap.
     * @param field field to add.
     */
    public void add(Object field) {
        SpeedoField f = (SpeedoField) field;
        f.jdoClass = this;
        f.number = fieldsCount++;
        jdoField.put(f.name, field);
    }

    /**
     * Adds a SpeedoField to the class.
     * @param field field to add.
     * @param failsOnError if an error must be thrown or creates a warning.
     * @param logger logger for writting warn message if necessary.
     * @exception SpeedoException if the field was already defined into the class.
     */
    public void add(Object field, boolean failsOnError, Logger logger) throws SpeedoException {
        SpeedoField f = (SpeedoField) field;
        if (jdoField.containsKey(f.name)) {
            if (failsOnError)
                throw new SpeedoException("The field " + f.name + " of class "
                        + name + " is defined twice.");
            else
                logger.log(BasicLevel.ERROR, "The field " + f.name + " of class "
                        + name + " is defined twice.");
        } else {
            logger.log(BasicLevel.DEBUG, "Add the field '" + f.name
                    + "' in class '" + name + "'.");
            f.jdoClass = this;
            jdoField.put(f.name, field);
        }
    }

	/**
	 * Set the versioning strategy of the class.
	 * @param version : the version to set
	 */
	public void addVersion(Object version) {
		SpeedoVersion v = (SpeedoVersion) version;
		v.jdoClass = this;
		this.version = v;
	}
	
	public void addDatastoreId(Object datastoreId) {
		SpeedoDatastoreId d = (SpeedoDatastoreId) datastoreId;
		d.jdoClass = this;
		this.datastoreId = d;
		//add the datasore sequence: can be null
		this.datastoreSequence = d.sequence;
	}
	
	/**
	 * Adds a SpeedoFetchGroup to the class.
	 * Precondition: this fetchgroup doesn't exist in the jdoFtechGroups HashMap.
	 * @param fetchGroup the fetchgroup to add.
	 */
	public void addFetchGroup(Object fetchGroup) {
		SpeedoFetchGroup fg = (SpeedoFetchGroup) fetchGroup;
		jdoFetchGroup.put(fg.name, fetchGroup);
	}

	/**
	 * Adds a SpeedoFetchgroup to the class.
	 * @param fetchGroup the fetchgroup to add.
	 * @param failsOnError if an error must be thrown or creates a warning.
	 * @param logger logger for writting warn message if necessary.
	 * @exception SpeedoException if the field was already defined into the class.
	 */
	public void addFetchGroup(Object fetchGroup, boolean failsOnError, Logger logger) throws SpeedoException {
		SpeedoFetchGroup fg = (SpeedoFetchGroup) fetchGroup;
		if (jdoFetchGroup.containsKey(fg.name)) {
			if (failsOnError)
				throw new SpeedoException("The fetchgroup " + fg.name + " of class "
						+ name + " is defined twice.");
			else
				logger.log(BasicLevel.ERROR, "The fetchgroup " + fg.name + " of class "
						+ name + " is defined twice.");
		} else {
			logger.log(BasicLevel.DEBUG, "Add the fetchgroup '" + fg.name
					+ "' in class '" + name + "'.");
			jdoFetchGroup.put(fg.name, fetchGroup);
		}
	}
	
	
    /**
     * Computes the field numbers for the fields of this class.
     *
     * @return the number of fields of this class (including inherited fields).
     */
    public int computeFieldNumbers() {
        if (fieldsCount != -1) {
            return fieldsCount;
        }
        if (superClassName.length() == 0) {
            fieldsCount = 0;
        } else {
            SpeedoClass jdoSuperClass = jdoPackage.jdoXMLDescriptor.smi
                .getSpeedoClass(superClassName, jdoPackage);
            if (jdoSuperClass == null) {
                throw new JDOUserException(
                        "The persistence-capable-superclass field has a bad value: '" 
                        + superClassName 
                        + "'. See the description of the class '" + getFQName() 
                        + "' from the file '" + getJDOFileName() + "'.");
            }
            fieldsCount = jdoSuperClass.computeFieldNumbers();
        }
        Iterator i = jdoField.values().iterator();
        while (i.hasNext()) {
            SpeedoField jf = (SpeedoField) i.next();
            jf.number = fieldsCount++;
        }
        return fieldsCount;
    }

    public String getJormFileName() {
        return getFQName().replace('.', File.separatorChar) + ".pd";
    }

    public boolean generateObjectId() {
        return getPkFieldCount() > 1;
    }

    public int getPkFieldCount() {
        if (pkFieldsCount == -1) {
            pkFieldsCount = 0;
            if (identityType == SpeedoIdentity.USER_ID &&
                    (objectidClass == null || objectidClass.length() == 0)) {
                Iterator it = jdoField.values().iterator();
                while(it.hasNext()) {
                    if (((SpeedoField) it.next()).primaryKey) {
                        pkFieldsCount++;
                    }
                }
            }
        }
        return pkFieldsCount;
    }
	public SpeedoClass getSpeedoClassFromContext(String className) {
		return jdoPackage.jdoXMLDescriptor.smi.getSpeedoClass(className, jdoPackage);
	}

	public String getJDOFileName() {
		return jdoPackage.jdoXMLDescriptor.xmlFile;
	}

	public SpeedoClass getSuper() {
		if (superClassName == null) {
			return null;
		} else {
			return getSpeedoClassFromContext(superClassName);
		}
	}

	public SpeedoClass getAncestor() {
		SpeedoClass tmp = getSuper();
		SpeedoClass ancestor = null;
		while(tmp != null && ancestor != tmp) {
			ancestor = tmp;
			tmp = tmp.getSuper();
		}
		return ancestor;
	}
}
