/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.naming.api;

import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.generation.jorm.JormMIMappingBuilder;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.mapper.api.JormFactory;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.CommonClassMapping;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.Logger;

import java.util.Collection;
import java.util.Map;
import java.util.Properties;

/**
 * It defines a manager of a type of naming. This interface covers needs at
 * compile time and at run time. In particular a naming manager is able to
 * fill the Speedo meta information with the particularity of the a naming.
 *
 * @author S.Chassande-Barrioz
 */
public interface NamingManager {

    String SEP = "/";

    /**
     * Indicates if the naming manager is able to manage the identifier of a
     * persistent class.
     * @param sc The speedo meta object representing the persistent class
     */
    boolean canManage(SpeedoClass sc);

    /**
     * indicates if the naming manager is able to build a PBinder instance for
     * this hints parameter specified
     * @param hints is the helper of the PBinder building
     * @param classLoader is the classloader to use for instanciate the PBinder
     */
	boolean canProvidePBinder(Object hints, ClassLoader classLoader);

	/**
	 * indicates if the naming manager is able to build a PNC instance for
	 * this hints parameter specified
	 * @param hints is the helper of the PNC building
	 * @param classLoader is the classloader to use for instanciate the PNC
	 */
    boolean canProvidePNamingContext(Object hints, ClassLoader classLoader);

	/**
	 * Build a PBinder instance for this hints parameter specified
	 * @param hints is the helper of the PBinder building. This hints must be
	 * the result of the getJormConfig call.
	 * @param classLoader is the classloader to use for instanciate the PBinder
	 * @param mappingStructureRule indicates the rule to apply concerning the
	 * data strucuture. Indeed a PBinder building can require to read the data
	 * support.
	 * @param cn2binder is the map of the existing PBinder instance
	 * (key = class name, value = PBinder instance)
	 * @param cn2pnc is the map of the existing PNC instance
	 * (key = class name, value = PNC instance)
	 */
    PBinder getPBinder(String className,
                       String hints,
                       ClassLoader classLoader,
                       byte mappingStructureRule,
                       Map cn2binder,
                       Map cn2pnc) throws PException;

	/**
	 * Build a PNamingContext instance for this hints parameter specified
	 * @param hints is the helper of the PBinder building. This hints must be
	 * the result of the getJormConfig call.
	 * @param classLoader is the classloader to use for instanciate the PBinder
	 * @param mappingStructureRule indicates the rule to apply concerning the
	 * data strucuture. Indeed a PBinder building can require to read the data
	 * support.
	 * @param cn2binder is the map of the existing PBinder instance
	 * (key = class name, value = PBinder instance)
	 * @param cn2pnc is the map of the existing PNC instance
	 * (key = class name, value = PNC instance)
	 */
    PNamingContext getPNamingContext(String className,
                                     String hints,
                                     ClassLoader classLoader,
                                     byte mappingStructureRule, Map cn2binder,
                                     Map cn2pnc, Manager miManager,
									 PClassMapping pcm) throws PException;

    /**
     * Decodes an object identifier into a PName.
     * @param pnc is the PNameCoder instance to use for the decoding the object
     * identifier. The pnc parameter can be null. In this case the oid must
     * @param oid is the object identifier
     * @param clazz is the java.lang.Class of the persistent class identified by
     * the oid
     * @param jf is the JormFactory instance. It permits to
     * @return the object identifier decode as PName
     * @throws PException
     */
	PName decode(PNameCoder pnc,
	             Object oid,
	             Class clazz,
	             JormFactory jf) throws PException;

    /**
	 * Encode a PName into an object
	 * @param pn is the pname to encode
	 */
	Object encode(PName pn) throws PException;

    /**
	 * indicates if the use of PNamingContext is support by this mapping. If
	 * true then getPNamingContext method can be use, otherwise the getPBinder
	 * must be use instead. The case of PNamingContext are not supported, means
	 * the PBinder are used as PNameCoder by PBinding as reference coder.
	 */
	boolean supportPNamingcontext();

    /**
     * Fill the name def of an identifier or a reference
     * @param nd is the name def to fill
     * @param isIdentifier indicates if the name represents an identifier (true)
     * or a reference (false).
     * @param isInGenClass indicates if the name is defined in a generic class
     * (true) of in a class (false).
     * @param ref meta object which the name def must be
     * defined. This value is used only in the case of isIdentifier == false and
     * isInGenClass == false.
     * @param tsc is the speedo meta object representing the referenced class.
     * This value is used only in the case of isIdentifier == false.
     * @param mo is the jorm meta object hosting the name def and on which the
     * eventual hidden field will be created.
     * @param hcm if the mapping structure hosting the mapping of the reference.
     * This value must be ClassMapping or GenClassMapping instance.
     * @param mb is the mapping builder permitting the creation of the mapping
     * part.
     * @param createdMOs is a result paramter. This collection must be fill with
     * the created Jorm Meta objects representing  a class or a composite name.
     * Here only the new used composite name will be added.
     * @throws SpeedoException if the speedo meta information is not completly
     * defined.
     */
    void fillNameDef(MIBuilderHelper mibh,
                     Manager manager,
                     NameDef nd,
                     SpeedoClass tsc,
                     SpeedoClass ssc,
                     MetaObject mo,
                     Reference ref,
                     CommonClassMapping hcm,
                     JormMIMappingBuilder mb,
                     boolean isIdentifier,
                     boolean isInGenClass,
                     boolean createField,
                     Collection createdMOs) throws SpeedoException, PException;

    /**
	 * computes the naming configuration information for a class or a reference.
	 * This information is store as a String. The string value will be passed
	 * to the getPBinder and getPNamingContext methods as hints. This method is
	 * used at generation time.
	 * <table>
	 * <tr><td>naming of a/td><td>MOClass/td></tr>
	 * <tr><td>class/td><td>Class/td></tr>
	 * <tr><td>reference to a Class/td><td>ClassRef/td></tr>
	 * <tr><td>reference to a GenClass/td><td>GenClassRef/td></tr>
	 * <tr><td>a GenClass identifier/td><td>GenClassRef/td></tr>
	 * </table>
	 *
	 * @param targetClass is the Speedo meta object representing the class which
	 * is designates by the
	 * @param result is the Properties which must be filled
	 */
	void getJormNamingConfig(NameDef nd,
                         SpeedoClass targetClass,
                         MetaObject sourceMO,
						 String key,
						 Properties result) throws SpeedoException;

    /**
	 * compute a string value which permit to help the building of the class
	 * PName (export)
	 * @param sc is the Speedo class which the hints must be specified
	 * @param nd is the name def of the class
	 */
	String getPNameHints(SpeedoClass sc, NameDef nd);

    /**
	 * compute a string value which permit to help the building of the genclass
	 * PName (export)
	 * @param sc is the Speedo class which the hints must be specified
	 * @param nd is the name def of the gen class
	 */
	String getGCPNameHints(SpeedoClass sc, NameDef nd);

	void setPMapper(PMapper mapper) throws PException;

    void setLogger(Logger logger);
    
    void setCache(CacheManager cache);
    
    void setPmf(ProxyManagerFactory pmf);

    /**
	 * Retrieves the array of field used in the identifier of a persistent class
	 * and which values are not calculated by the application. The application
	 * field must be set at makePersistent time.
	 *
	 * @param sc is the speedo meta object of the persistent class
	 */
	NamingField[] getNamingfields(SpeedoClass sc) throws PException;

    /**
	 * Defines the field used in an identifier and which must be set at
	 * makePersistent time.
	 */
	class NamingField {
        private String userFieldName = null;
        private Class userFieldType = null;
        private String namingFieldName = null;
        private PType namingFieldType = null;

        /**
		 * @param userFieldName is the field name in the persistent class
		 * @param userFieldType is the field type (java type) in the persistent
		 * class
		 * @param namingFieldName is the field name in the identifier
		 * @param namingFieldType is the field type (Jorm type) in the identifier
		 */
		public NamingField(String userFieldName,
                           Class userFieldType,
                           String namingFieldName,
                           PType namingFieldType) {
            this.userFieldName = userFieldName;
            this.userFieldType = userFieldType;
            this.namingFieldName = namingFieldName;
            this.namingFieldType = namingFieldType;
        }
        /**
		 * @return the field name in the persistent class
		 */
		public String getUserFieldName() { return userFieldName; }
        /**
		 * @return the field type (java type) in the persistent
		 */
		public Class getUserFieldType() { return userFieldType; }
        /**
		 * @return the field name in the identifier
		 */
		public String getNamingFieldName() { return namingFieldName; }
        /**
		 * @return the field type (Jorm type) in the identifier
		 */
		public PType getNamingFieldType() { return namingFieldType; }
    }
}