/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.naming.lib;

import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.speedo.naming.api.NamingManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.mapper.api.JormFactory;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Arrays;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class NamingManagerFactory {

    private NamingManager[] namingManagers;

    private PMapper mapper;
    private CacheManager cache;
    private ProxyManagerFactory pmf;
    private Logger logger;


    public NamingManagerFactory() {
        namingManagers = new NamingManager[]{
			new UserIdSingleNamingManager(),
			new PolymorphIdNamingManager(),
			new RdbSequenceNamingManager(),
			new UserIdCompositeNamingManager(),
			new LongIdNamingManager(),
			new OLongIdNamingManager()
		};
    }

    public PMapper getMapper() {
        return mapper;
    }

    public void setMapper(PMapper mapper) {
        this.mapper = mapper;
        for(int i=0; i<namingManagers.length; i++) {
			try {
				namingManagers[i].setPMapper(mapper);
			} catch (PException e) {
				if (logger != null) {
					logger.log(BasicLevel.WARN, "The naming manager "
						+ namingManagers[i] + " does not accept the mapper: ", e);
				}
			}
		}
    }

    public CacheManager getCache() {
        return cache;
    }

    public void setCache(CacheManager cache) {
        this.cache = cache;
        for(int i=0; i<namingManagers.length; i++) {
			namingManagers[i].setCache(cache);
		}
    }
    
	public ProxyManagerFactory getPmf() {
		return pmf;
	}
	
	public void setPmf(ProxyManagerFactory pmf) {
		this.pmf = pmf;
		for(int i=0; i<namingManagers.length; i++) {
			namingManagers[i].setPmf(pmf);
		}
	}
	
    public Logger getLogger() {
        return logger;
    }

    public void setLogger(Logger logger) {
        this.logger = logger;
        for(int i=0; i<namingManagers.length; i++) {
            namingManagers[i].setLogger(logger);
        }
    }

    public synchronized void bindNamingManager(NamingManager nm) {
		NamingManager[] neo = new NamingManager[namingManagers.length + 1];
		System.arraycopy(namingManagers, 0, neo, 0, namingManagers.length);
		neo[namingManagers.length] = nm;
		namingManagers = neo;
    }

    public synchronized boolean unbindNamingManager(NamingManager nm) {
        int idx = Arrays.binarySearch(namingManagers, nm);
		if (idx >= 0) {
			//put the last in place of the removed
			namingManagers[idx] = namingManagers[namingManagers.length - 1];
			//create a array without the lastest element
			NamingManager[] neo = new NamingManager[namingManagers.length - 1];
			System.arraycopy(namingManagers, 0, neo, 0, namingManagers.length - 1);
			namingManagers = neo;
			return true;
		} else {
			return false;
		}
    }

    public NamingManager getNamingManager(SpeedoClass sc) throws SpeedoException {
        for(int i=0; i<namingManagers.length; i++) {
            if (namingManagers[i].canManage(sc)) {
                return namingManagers[i];
            }
        }
        throw new SpeedoException("No identity manager found for the class '"
                + sc.getFQName() + "', " + sc.identityType + ", " +
                sc.getExtension(SpeedoProperties.VENDOR_NAME, SpeedoProperties.ID));
    }

    public NamingManager getNamingManager(String hints, ClassLoader classloader) throws PException {
        for(int i=0; i<namingManagers.length; i++) {
            if (namingManagers[i].canProvidePBinder(hints, classloader)) {
                return namingManagers[i];
            }
        }
        throw new PException("No identity manager found for the hints '"
                + hints + "'.");
    }

    public PName decode(PNameCoder pnc, Object oid, Class clazz, JormFactory jf) throws PException {
        for(int i=0; i<namingManagers.length; i++) {
            PName pn = namingManagers[i].decode(pnc, oid, clazz, jf);
            if (pn != null) {
                return pn;
            }
        }
        throw new PException("No identity manager able to decode:"
                + "\n\tpnc=" + pnc
				+ "\n\toid=" + oid
				+ "\n\tclazz=" + clazz);
    }

    public Object encode(PName pn) throws PException {
        for(int i=0; i<namingManagers.length; i++) {
            Object o = namingManagers[i].encode(pn);
            if (o != null) {
                return o;
            }
        }
        throw new PException("No identity manager able to encode ("
                + pn + ")");
    }
}
