/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo.pm.api;

import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.concurrency.lib.Semaphore;

import javax.jdo.datastore.JDOConnection;
import javax.jdo.PersistenceManager;
import javax.transaction.Synchronization;

import java.util.Collection;
import java.util.Map;

/**
 * This interfaces defined a manager of persistent instance. This concept
 * is based on the javax.jdo.PersistenceManager interface.
 * A ProxyManager is also a javax.transaction.Synchronization for the JTA
 * integration. This permits to support the container transaction demarcation.
 * On the beforeCompletion method the JDO transaction is prepared.
 * A proxy manager used a TransactionalPersistenceManager for managing the
 * concurrency, the loading and the caching aspects.
 *
 * @see org.objectweb.speedo.pm.api.ProxyManagerFactory
 * @see org.objectweb.speedo.pm.api.ProxyManagerSwitch
 * @see org.objectweb.perseus.persistence.api.TransactionalPersistenceManager
 * 
 * @author S.Chassande-Barrioz
 */
public interface ProxyManager
		extends PersistenceManager, Synchronization, JDOConnection {

    /**
	 * Retrieves the TransactionalPersistenceManager used by this ProxyManager.
	 */
	TransactionalPersistenceManager getTransactionalPersistenceManager();

    /**
     * Opens the persistent manager. This operation is the opposite of the
	 * javax.jdo.PersistenceManager.close() method. It prepares a ProxyManager
	 * to be used. During the preparation, the optimistic and multithread modes
	 * are initialized.
     * @param connectionSpec is the information to access to the data store
     * (user, password, ...)
     */
    void open(Object connectionSpec);

	/**
	 * @return the connection information to access the data store
	 */
    Object getConnectionSpec();

    /**
     * Signal to the persistence maneger that it is used. A persistence managed
	 * can be used by several thread. In this case each thread have done a
	 * PersistenceManagerFactory.getPersistentceManager() to obtain a proxy
	 * manager instance. The threads will do a close() operation, but only the
	 * last has to be taken in account. This method permits to knwon how many
	 * users uses the current proxy manager.
     */
    void addUse();

	/**
	 * is the same method than PersistenceManager.getObjectById(Object, boolean)
	 * but no checking is done, because this is an internal call of Speedo.
	 * @param oid is an object representing an persistent object.
	 */
    Object speedoGetObjectById(Object oid, boolean validate);

    /**
	 * Make persistent a SpeedoProxy. This method does the same thing than the
	 * PersistenceManager.makePersistent(Object) method except the call to the
	 * bind to the ProxyManager to the current thread.
	 *
	 * @param sp is the instance to make persistent.
	 * @param byAttach is a boolean value indicating if the instance to make
	 * persistent becomes persistent because the instance is linked/attached to
	 * a persistent object (true), or because the user calls explicitly
	 * pm.makePersistent(Obj) (false).
	 */
	void speedoMakePersistent(SpeedoProxy sp, boolean byAttach);

	/**
	 * Delete persistent a SpeedoProxy. This method does the same thing than the
	 * PersistenceManager.deletePersistent(Object) method except the call to the
	 * bind to the ProxyManager to the current thread.
	 *
	 * @param o is the instance to make persistent.
	 */
    void speedoDeletePersistent(Object o);

	/**
	 * Retrieves the semaphore object permiting the multithreading mode.
	 */
    Semaphore getSemaphore();
    	
	/**
	 * Same as detachCopy(Object detached), but uses a map to avoid cycles when detaching objects referencing each other
	 * @param map the map contains element of type: [SpeedoProxy, clone of SpeedoProxy]
	 */
	Object speedoDetachCopy(SpeedoProxy sp, Map map, Collection fgHints);
	
	Object speedoAttachCopy(Object detached, boolean makeTransactional, Map map);
	
	void speedoRefresh(SpeedoProxy sp, Map map, Collection fgHints);
	
	void speedoRetrieve(SpeedoProxy sp, Map map, Collection fgHints);
}
