/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo.pm.lib;

import org.objectweb.speedo.pm.api.ProxyManagerSwitch;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Collections;

/**
 * This class is an implementation of the ProxyManagerSwitch based on the use
 * of a ThreadLocal field. This field contains an instance of ProxyManager of an
 * ArrayList of ProxyManager. A SpeedoProxyManagerSwitch is a fractal
 * component which have no bindings.
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoProxyManagerSwitch implements ProxyManagerSwitch {

    protected ThreadLocal pms = new ThreadLocal();

    public final static String BIND_ERROR_MSG = "Impossible to associated to ProxyManager linked to the same PersistenceManagerFactory in the same context (thread)";

    // IMPLEMENTATION OF THE ProxyManagerSwitch INTERFACE //
    //----------------------------------------------------//

    /**
     * @param pmf is persistent manager factory which manages the returned
     * proxy manager.
     * @return the ProxyManager managed by the given persistence manager factory
     * and bound to current the context, or the null value if there is no
     * ProxyManager.
     */
    public ProxyManager lookup(ProxyManagerFactory pmf) {
        Object o = pms.get();
        if (o == null) {
            return null;
        } else if (o instanceof ProxyManager) {
            if (((ProxyManager) o).getPersistenceManagerFactory() == pmf)
                return (ProxyManager) o;
        } else {
            for (Iterator it = ((List) o).iterator(); it.hasNext();) {
                ProxyManager pm = (ProxyManager) it.next();
                if (pm.getPersistenceManagerFactory() == pmf)
                    return pm;
            }
        }
        return null;
    }

    /**
     * It assignes a ProxyManager to the current context.
     * @param pm is the ProxyManager
     */
    public void bind(ProxyManager pm) {
        Object o = pms.get();
        if (o == null) {
            pms.set(pm);
        } else if (o instanceof ProxyManager) {
            // There is one ProxyManager, then check if the old and the new
            // proxy manager are manager by the same PMF.
            ProxyManager pm1 = (ProxyManager) o;
            if (pm1.getPersistenceManagerFactory()
                    == pm.getPersistenceManagerFactory()) {
                pms.set(pm);
            } else {
                // Put the old and the new in a List
                ArrayList al = new ArrayList(3);
                al.add(pm1);
                al.add(pm);
                pms.set(al);
            }
        } else {
            // The element is a list of ProxyManager instances, then check if
            // there is a proxy manager linked to the same PersistenceManagerFactory
            List l = (List) o;
            for (Iterator it = l.iterator(); it.hasNext();) {
                if (((ProxyManager) it.next()).getPersistenceManagerFactory()
                        == pm.getPersistenceManagerFactory()) {
                    it.remove();
                }
            }
            //No element with the same pmf, then add the proxy manager
            ((List) o).add(pm);
        }
    }

    /**
     * It clears the list of ProxyManager for the current context.
     */
    public void clear() {
        pms.set(null);
    }

    /**
     * It clears a ProxyManager for the current context.
     */
    public boolean unbind(ProxyManager pm) {
        Object o = pms.get();
        if (o == null) {
			return false;
		} else if (o instanceof ProxyManager) {
            if (o == pm) {
                pms.set(null);
                return true;
            }
        } else if (o instanceof Collection) {
            return ((Collection) o).remove(pm);
        }
        return false;
    }

	public boolean unbind(ProxyManagerFactory pmf) {
		Object o = pms.get();
		if (o == null) {
			return false;
		} else if (o instanceof ProxyManager) {
			if (((ProxyManager) o).getPersistenceManagerFactory() == pmf) {
				pms.set(null);
				return true;
			}
		} else { // The element is a list of ProxyManager instances
            for (Iterator it = ((List) o).iterator(); it.hasNext();) {
                if (((ProxyManager) it.next()).getPersistenceManagerFactory()
                        == pmf) {
                    it.remove();
					return true;
                }
            }
        }
		return false;
	}

    /**
     * @return all ProxyManager instances bound with the current context
     */
    public Collection entries() {
        Object o = pms.get();
        if (o instanceof ProxyManager)
            return Collections.singletonList(o);
        else
            return (Collection) o;
    }
}
