/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.query.lib;

import org.objectweb.speedo.query.api.QueryDefinition;

import javax.jdo.Extent;

import java.util.Iterator;
import java.util.List;
import java.util.Collection;

/**
 * is a basic implementation of the QueryDefinition interface.
 * 
 * @author S.Chassande-Barrioz
 */
public class QueryDefinitionImpl implements QueryDefinition {
	/**
	 * IgnoreCache option.
	 * The ignoreCache option setting specifies whether the query should execute
	 * entirely in the back end, instead of in the cache.  If this flag is set
	 * to true, an implementation might be able to optimize the query
	 * execution by ignoring changed values in the cache.  For optimistic
	 * transactions, this can dramatically improve query response times.
	 */
	protected boolean ignoreCache = true;

	/**
	 * Candidate classes or extent for the query.
	 */
	protected Collection candidateInstances = null;
	protected Extent extentClass = null;

	/**
	 * Class of candadate classes.
	 */
	protected Class candidateClass = null;

	/**
	 * Query filter.
	 */
	protected String filter = "true";

	/**
	 * Query parameters and variables declaration
	 */
	protected String parameters = null;
	protected String variables = null;

	/**
	 * the import statements is transformed into a simple vector
	 */
	protected List importStatements = null;
	protected List order = null;

	transient protected long indexFirst = 0;

	transient protected long indexLast = Long.MAX_VALUE;

	protected boolean unique;

	protected String result;

	protected Class resultClass;
	
	protected boolean withPrefetch;
	
	boolean includeSubClasses = true;
	
	protected String grouping;

	public QueryDefinitionImpl() {
	}

	public QueryDefinitionImpl(QueryDefinition qd) {
		defineWith(qd);
	}

	public void defineWith(QueryDefinition qd) {
	    if (qd == null) {
		    this.candidateClass = null;
		    this.candidateInstances = null;
		    this.extentClass = null;
		    this.filter = null;
		    this.order = null;
		    this.variables = null;
		    this.parameters = null;
		    this.importStatements = null;
		    this.ignoreCache = true;
			this.indexFirst = 0;
			this.indexLast = Long.MAX_VALUE;
			this.withPrefetch = true;
			this.grouping = null;
			this.result = null;
			this.resultClass = null;
			this.includeSubClasses = true;
			this.unique = false;
	    } else {
		    this.candidateClass = qd.getCandidateClass();
		    this.candidateInstances = qd.getCollection();
		    this.extentClass = qd.getExtent();
		    this.filter = qd.getFilter();
		    this.order = qd.getOrder();
		    this.variables = qd.getVariables();
		    this.parameters = qd.getParameters();
		    this.importStatements = qd.getImportStatements();
		    this.ignoreCache = qd.isIgnoreCache();
			this.indexFirst = qd.getIndexFirst();
			this.indexLast = qd.getIndexLast();
			this.withPrefetch = qd.withPrefetch();
			this.grouping = qd.getGrouping();
			this.result = qd.getResult();
			this.resultClass = qd.getResultClass();
			this.includeSubClasses = qd.getIncludeSubClasses();
			this.unique = qd.getUnique();
	    }
	}


	public Class getCandidateClass() {
		return candidateClass;
	}

	public boolean isCollection() {
		return candidateInstances != null;
	}

	public Collection getCollection() {
		return candidateInstances;
	}

	public String getParameters() {
		return parameters;
	}

	public String getVariables() {
		return variables;
	}

	public String getFilter() {
		return filter;
	}

	public Extent getExtent() {
		return extentClass;
	}

	public List getImportStatements() {
		return importStatements;
	}

	public boolean isIgnoreCache() {
		return ignoreCache;
	}

	public List getOrder() {
		return order;
	}

	public long getIndexFirst() {
		return indexFirst;
	}

	public long getIndexLast() {
		return indexLast;
	}

	public boolean getUnique() {
		return unique;
	}

	public String getResult() {
		return result;
	}

	public Class getResultClass() {
		return resultClass;
	}

	public boolean getIncludeSubClasses() {
	    return includeSubClasses;
	}
		
	public String getGrouping() {
		return grouping;
	}
	public boolean withPrefetch() {
		return withPrefetch;
	}
	public void withPrefetch(boolean withPrefetch) {
		this.withPrefetch = withPrefetch;
	}
	public String toString() {
	    return qdToString(true);
	}
	public String qdToString(final boolean oneLine) {
		final StringBuffer sb = new StringBuffer();
		String sep = "";
		final String sepNext = (oneLine ? ", " : "\n\t-");
		if (candidateClass!=null) {
			sb.append(sep).append("class=\"").append(candidateClass.getName()).append("\"");
			sep = sepNext;
		}
		if (parameters!=null && parameters.length() >0) {
			sb.append(sep).append("parameters=\"").append(parameters).append("\"");
			sep = sepNext;
		}
		if (variables!=null && variables.length() >0) {
			sb.append(sep).append("variables=\"").append(variables).append("\"");
			sep = sepNext;
		}
		if (filter!=null && filter.length() >0 && !filter.equals("true")) {
			sb.append(sep).append("filter=\"").append(filter).append("\"");
			sep = sepNext;
		}
		if (order!=null && order.size() >0) {
			sb.append(sep).append("order=\"");
			for (Iterator iter = order.iterator(); iter.hasNext();) {
                sb.append(iter.next());
            }
			sb.append("\"");
			sep = sepNext;
		}
		if (result != null) {
			sb.append(sep).append("result=\"").append(result).append("\"");
			sep = sepNext;
		}
		if (resultClass != null) {
			sb.append(sep).append("resultClass=\"").append(resultClass).append("\"");
			sep = sepNext;
		}
		if (grouping!=null && grouping.length() >0) {
		    sb.append(sep).append("grouping=\"").append(grouping).append("\"");
			sep = sepNext;
		}
		if (unique) {
		    sb.append(sep).append("unique=\"").append(unique).append("\"");
		    sep = sepNext;
		}
		if (ignoreCache) {
		    sb.append(sep).append("ignoreCache=\"").append(ignoreCache).append("\"");
		    sep = sepNext;
		}
		if (!withPrefetch) {
		    sb.append(sep).append("withPrefetch=\"").append(withPrefetch).append("\"");
		    sep = sepNext;
		}
		return sb.toString();
	}

	public int hashCode() {
		return candidateClass.hashCode();
	}

	public boolean equals(Object obj) {
		if (obj == null || !(obj instanceof QueryDefinition)) {
			return false;
		}
		QueryDefinition qd = (QueryDefinition) obj;
		return (candidateClass==null
				? qd.getCandidateClass()==null
				: candidateClass.equals(qd.getCandidateClass()))
				&& (parameters == null
				? qd.getParameters() == null
				: parameters.equals(qd.getParameters()))
				&& (variables == null
				? qd.getVariables() == null
				: variables.equals(qd.getVariables()))
				&& (filter == null
				? qd.getFilter() == null
				: filter.equals(qd.getFilter()))
				&& (order == null
				? qd.getOrder() == null
				: order.equals(qd.getOrder()))
				&& (result == null
						? qd.getResult() == null
						: result.equals(qd.getResult()))
				&& (resultClass == null
						? qd.getResultClass() == null
						: resultClass.equals(qd.getResultClass()))
				&& (grouping == null
						? qd.getGrouping() == null
						: grouping.equals(qd.getGrouping()))
				&& (qd.isCollection()
					? qd.getCollection().equals(candidateInstances) : true)
				&& (qd.getUnique() == unique)
				&& (qd.withPrefetch() == withPrefetch)
				&& sameRangeCondition(qd)
				;
	}
	
	private boolean sameRangeCondition(QueryDefinition qd) {
	    //index of first must be the same thing 0 or not 0
	    if ((indexFirst == 0 && qd.getIndexFirst() != 0)  
	            || (indexFirst != 0 && qd.getIndexFirst() == 0)) {
	            return false;
	    }
	    //index of the last must be the same thing Long.MAX_VALUE or not Long.MAX_VALUE
	    if ((indexLast == Long.MAX_VALUE && qd.getIndexLast() != Long.MAX_VALUE)  
	            || (indexLast != Long.MAX_VALUE && qd.getIndexLast() == Long.MAX_VALUE)) {
	            return false;
	    }
	    return true;
	}

}
