/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 10 f�vr. 2005
 *
 */
package org.objectweb.speedo.query.lib;

import java.lang.reflect.Constructor;

import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;

import org.objectweb.jorm.naming.api.PName;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Is a common classes used for compute the query result. This implementation
 * supports user class or persistent class as query result.
 *
 * @author S.Chassande-Barrioz
 */
public class QueryResultCommon {
    /**
	 * The inner medor TupleCollection
	 */
	protected TupleCollection tc = null;

	/**
	 * The underlying connection used to evalue the query. These connections
	 * will be closed when the result will be no more used.
	 */
	protected Object[] conns = null;

	/**
	 * The proxy manager representing the current execution context. The 
	 * persistenceManager is used to fetch the real object from the identifier.
	 */
	protected PersistenceManager pm = null;

	protected Logger logger;
	
	protected boolean debug;

    /**
     * is the Class encapsulating the result of the query. A null value means
     * the result of the query is primitive field or a persistent class.
     */
	private Class resultClass;

    /**
     * indicates if the result is directly the first selected field.
     * Of course if there are several selected fields, this fies must be false
     */
    private boolean firstIsResult;
    
    /**
     * is the table of field types which compose the query result.
     */
    private Class[] selectedFieldTypes;
    
    /**
     * is the constructor method to use for filling user class. This field is 
     * null when not result class has been defined.
     */
    private Constructor cons;
    
    /**
     * index of the first element in the tuple collection
     * This design supports data prefetching.
     * 0: indicates that the indexes must be computed each time
     * 1: is the first element 
     */
    private int firstElementIdx;

    /**
     * Builds a QueryResultList.
     * @param _tc the tuple collection representing the query result
     * @param _pm is the peristence manager linked to the query
     * @param _conns is the connection to the underlying support to close in 
     * same time than the query.
     * @param _resultClazz is the class encapsulated the result
     */
	public QueryResultCommon(TupleCollection _tc, 
	        PersistenceManager _pm,
            Object[] _conns, 
            Class _resultClazz, 
            Class[] _selectedFieldTypes, 
            boolean staticFirstElementIndex,
            Logger _logger)
            throws MedorException, SpeedoException {
        this.tc = _tc;
		this.pm = _pm;
        this.conns = _conns;
        this.resultClass = _resultClazz;
        this.selectedFieldTypes = _selectedFieldTypes;
		this.logger = _logger;
        this.firstElementIdx = 0;

        debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
		if (pm == null || pm.isClosed() || ((Transaction) pm.currentTransaction()).getStatus()
				== WorkingSet.CTX_CLOSED) {
			throw new JDOUserException(
					"Impossible to use a query result without opened persistence manager");
		}
        firstIsResult = selectedFieldTypes.length == 1 //one projected field
        	&& (resultClass == null // no result class 
        	    || resultClass == selectedFieldTypes[0] // result class is the projected field type 
        	);

        if (staticFirstElementIndex) {
            //compute once time the first element index
            this.firstElementIdx = getFirstElementIndex();
            if (debug) {
                logger.log(BasicLevel.DEBUG, "index of the pname: " + firstElementIdx);
            }
        } //else dynamic pname index computed each time.

        if (resultClass != null) {
	        try {
	            cons = resultClass.getConstructor(selectedFieldTypes);
	        } catch (Exception e) {
	            StringBuffer sb = new StringBuffer();
	            sb.append("No consutructor '");
	            sb.append(resultClass.getName());
	            sb.append("(");
	            String sep = "";
	            for (int i = 0; i < selectedFieldTypes.length; i++) {
	                Class c = selectedFieldTypes[i];
	                sb.append(sep);
	                sep = ", "; 
	                sb.append(c == null ? "null" : selectedFieldTypes[i].getName());
	            }
	            sb.append(")' found. This constructor is required in a query result. Be carefull to not use non static inner class.");
	            throw new JDOUserException(sb.toString(), e);
	        }
        }
    }

    /**
	 * Close the connection to the data store. The query result will be no
	 * more usable after.
	 */
	public void close() {
	    pm = null;
		if (tc != null) {
			try {
				tc.close();
			} catch (MedorException e) {
				logger.log(BasicLevel.WARN, "Error during the connection closing: ", e);
			}
			tc = null;
		}
		if (conns != null) {
			for(int i=0; i<conns.length; i++) {
			    closeConnection(conns[i]);
			}
		}
        conns = null;
    }

	public static void closeConnection(Object conn) {
		if (conn instanceof ConnectionHolder) {
			try {
				((ConnectionHolder) conn).releaseCHConnection();
			} catch (PersistenceException e) {
			}
		}
	}
	
	/**
	 * Retrieves the PName index from the TupleStructure.
	 * @return
	 * @throws MedorException
	 */
	private int getFirstElementIndex() throws MedorException {
        if (firstElementIdx == 0) {
            return tc.getMetaData().getSize() - selectedFieldTypes.length + 1;
        } else {
            return firstElementIdx;
        }
	}
	
    /**
     * Fetch the persistent instance corresponding to an identifier
     * @param pn is the persistent object identifier
     * @return a persistent instance if it has been found, otherwise a null 
     * value.
     */
	protected Object pname2Object(PName pn) {
        if (pn != null && !pn.isNull()) {
            try {
                return pm.getObjectById(pn, false);
            } catch (JDOUserException e) {
                if (debug) {
                    logger.log(BasicLevel.DEBUG,
					"Ignore a deleted object, oid=" + pn, e);
                }
			// Do nothing, the persistent object does not exist
			// ==> ignore it and go to the next
            } catch (Exception e) {
                logger.log(BasicLevel.WARN, "Unexpected exception", e);
            }
        }
        return null;
    }
    
    /**
     * When the array #selectedFieldTypes contains a single element, the result
     * of the query can or cannot be encapsulated into a user class. When this
     * array contains several 
     * element and the #resultClass field is not definied, the  
     * @param tuple contains the result
     * @return the query result
     */
    protected Object getValue(Tuple tuple) throws MedorException {
        if (firstIsResult) {
            //return directly the first element
            Object o = tuple.getObject(getFirstElementIndex());
            if (o instanceof PName) {
                //the element is a PName, then fetch the persistent object 
                // from the identifier
                return pname2Object((PName) o);
            } else {
                return o;
            }
        } else {
            //Result is encapsulated into a user class or there are several
            // selected fields
            
            //fetch all values into an array
            Object[] values = new Object[selectedFieldTypes.length];
            for (int i = 0; i < values.length; i++) {
                values[i] = tuple.getObject(getFirstElementIndex() + i);
                if (values[i] instanceof PName) {
                    //the element is a PName, then fetch the persistent object 
                    // from the identifier
                    values[i] = pname2Object((PName) values[i]);
                }
            }
            if (resultClass == null) {
                //no user class specified but there are several selected fields
                //then return an Object[] containing the values.
                return values;
            }
            try {
                //instanciate the user class with values
                return cons.newInstance(values);
            } catch (Exception e) {
                StringBuffer sb = new StringBuffer();
                sb.append("Class '");
                sb.append(resultClass.getName());
                sb.append("' used as query result is compatible with the following result: [");
                String sep = "";
                for (int i = 0; i < selectedFieldTypes.length; i++) {
                    sb.append(sep);
                    sep = ", "; 
                    sb.append(values[i]);
                }
                sb.append("]");
                throw new JDOUserException(sb.toString());
            }
        }
    }
}
