/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S. Chassande-Barrioz
 *
 */

package org.objectweb.speedo.query.lib;

import java.io.CharArrayReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

import javax.jdo.JDOException;
import javax.jdo.JDOUserException;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.eval.lib.MedorEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferFactory;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.ConditionalAnd;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.optim.jorm.Jorm2Rdb;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.medor.optim.lib.PushNotInExpressionRule;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.jorm.lib.QueryBuilder;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.medor.type.lib.QType;
import org.objectweb.perseus.cache.api.UnFixProtocolException;
import org.objectweb.perseus.cache.replacement.api.ReplaceableCacheEntry;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.StateFilter;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.mapper.api.JormFactory;
import org.objectweb.speedo.mapper.lib.DelegatePMapper;
import org.objectweb.speedo.mim.api.SpeedoHome;
import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.query.api.CompiledQuery;
import org.objectweb.speedo.query.api.QueryDefinition;
import org.objectweb.speedo.query.parser.ASTSpeedoQL;
import org.objectweb.speedo.query.parser.ParseException;
import org.objectweb.speedo.query.parser.SelectGroupByVisitor;
import org.objectweb.speedo.query.parser.SimpleNode;
import org.objectweb.speedo.query.parser.SpeedoQL;
import org.objectweb.speedo.query.parser.SpeedoQLQueryFilterVisitor;
import org.objectweb.speedo.query.parser.SpeedoQLVariableVisitor;
import org.objectweb.speedo.usercache.api.UserCache;
import org.objectweb.speedo.usercache.lib.UserCacheKey;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * SpeedoCompiledQuery object represents a query. This object is created
 * when a new query is created, and can be used several times. A list of
 * SpeedoCompiledQuery is managed with a SpeedoQueryManager object.
 * When a user creates a new JDO Query object (SpeedoQuery), a
 * SpeedoCompiledQuery object is associated to the SpeedoQuery object which is
 * used to delegate some methods.
 */
public class SpeedoCompiledQuery
        implements CompiledQuery, StateFilter, ReplaceableCacheEntry {

    private short status;
	private ClassLoader classLoader;

    /**
     * Fields from javax.jdo.Query
     */
	private QueryDefinitionImpl qd = null;

    /**
     * vparams, hparams and hvars are internal objects to manipulate the list
     * of parameters, and the list of variables.
     */
    private Map hparams = null;
    private Map paramName2paramClass = null;
    private Map hvars = null;
    private QueryEvalContext[] qecs = null;
    private Class[] selectedFieldTypes = null;

    /**
     * The query tree which is the result from the parsing of the given filter
     */
    //private QueryTree qt = null;

    private Collection pncParams = null;
    /**
     * Logger for monolog
     */
    private Logger logger = null;
	private Logger varParserlogger = null;
	private Logger filterParserLogger = null;

    private DelegatePMapper mapper = null;
    private JormFactory jf;
	private PrefetchBufferFactory prefetchBufferFactory;
	private UserCache userCache = null;
	private Operand[] userCacheIndexes = null;


	private long age = 0;
	private int fixCount = 0;


    public void init(Logger l,
					 Logger logParserVar,
					 Logger logParserFil,
					 PMapper m,
					 PrefetchBufferFactory pbf,
					 JormFactory _jf) {
        logger = l;
		varParserlogger = logParserVar;
		filterParserLogger = logParserFil;
        mapper = new DelegatePMapper(m);
		setJormFactory(_jf);
		this.prefetchBufferFactory = pbf;
    }

    public PMapper getMapper() {
        return mapper;
    }

    public void setMapper(PMapper m) {
        mapper = new DelegatePMapper(m);
        if (jf != null) {
            mapper.setJormFactory(jf);
        }
    }

    public void setJormFactory(JormFactory jf) {
        this.jf = jf;
        if (mapper != null) {
            mapper.setJormFactory(jf);
        }
    }

	public Logger getLogger() {
		return logger;
	}

	public QueryEvalContext[] getQueryEvalContext() {
		return qecs;
	}

	public PrefetchBufferFactory getPrefetchBufferFactory() {
		return prefetchBufferFactory;
	}

	public boolean isPrefetchResult() {
		return qd.withPrefetch;
	}

    public void defineQuery(QueryDefinition _qd) {
	    this.qd = new QueryDefinitionImpl(_qd);
        classLoader = _qd.getCandidateClass().getClassLoader();
        if (classLoader == null) {
            classLoader = getClass().getClassLoader();
			if (classLoader == null) {
				classLoader = ClassLoader.getSystemClassLoader();
				logger.log(BasicLevel.DEBUG, "The system classLoader of the " +
						"class is assigned to the query: " + classLoader);
			} else {
				logger.log(BasicLevel.DEBUG, "The classLoader of Speedo is" +
						" assigned the query: " + classLoader);
			}
		} else {
			logger.log(BasicLevel.DEBUG, "The classLoader of the class is " +
					"assigned to the query: " + classLoader);
		}
        mapper.setClassLoader(classLoader);
        status = DEFINED;
    }

    // IMPLEMENTATION OF ReplaceableCacheEntry INTERFACE //
    //---------------------------------------------------//
	public long getCeAge() {
		return age;
	}

	public void setCeAge(long _age) {
		this.age = _age;
	}

	public void fixCe() {
		fixCount++;
	}

	public void unfixCe() throws UnFixProtocolException {
		fixCount--;
	}

	public int getCeFixCount() {
		return fixCount;
	}

	public Object getCeObject() {
		return this;
	}

	public Object getCeIdentifier() {
		return qd;
	}

    // IMPLEMENTATION OF CompiledQuery INTERFACE //
    //-------------------------------------------//
	public synchronized QueryDefinition getDefinition() {
		switch(status) {
		case DEFINED:
		case COMPILED:
			return qd;
		case UNDEFINED:
		default:
			return null;
		}
	}

    /**
     * compile the current SpeedoCompiledQuery.
     * The query is prepared to be executed.
     * The PersistenceManager is set (even if there was a previous definition
     * of a PersistenceManager.
     */
    public synchronized void compile() throws SpeedoException, MedorException, ExpressionException {
        if (status == UNDEFINED)
            throw new SpeedoException("Impossible to compile an undefined query");
        if (status == COMPILED)
            return;
		long timeToCompile = System.currentTimeMillis();
        boolean debug = logger.isLoggable(BasicLevel.DEBUG);
        // create a speedoQL object with a filter string
	    String filter = qd.filter;
	    filter = '(' + filter + ')';
        // create representations of the parameters list and the variable
        // list
        toHashtableParams(qd.parameters, ";,");
        toHashtableVars(qd.variables, ";,");
        Manager miManager = mapper.getMetaInfoManager();
        if (miManager == null)
            throw new SpeedoException(
                    "A non null Meta information manager is needed");
        try {
        	jf.getPClassMapping(
        			qd.candidateClass.getName(),
					classLoader);
        } catch (Exception e) {
			throw new SpeedoException(e);
		}
        SimpleNode node = null;
        try {
            node = new SpeedoQL(new CharArrayReader(filter.toCharArray())).SpeedoQL();
        } catch (ParseException e) {
            throw new SpeedoException(
                    "Impossible to parse the filter and to create AST", e);
        }
        SpeedoQLVariableVisitor sqvv = new SpeedoQLVariableVisitor(
                node, miManager, varParserlogger, hparams, hvars, qd.order,
                qd.candidateClass.getName(), qd.includeSubClasses); 
        // start the variable visitor to catch all variables an build a
        // first tree of them without collection navigation
        Map fields = sqvv.getFields();
        QueryBuilder qb = sqvv.getQueryBuilder();
        QueryTree qt = sqvv.getQueryTree();

        SelectProject sp = new SelectProject("");
        if (!filter.equals("(true)") && !filter.equals("true")) {
            //Ther is a filter and potentialy collection navigation
            if (debug) {
                logger.log(BasicLevel.DEBUG, "filter = " + qd.filter);
            }
            // start the query filter visitor, to build and expression tree of
            // the filter expression
            SpeedoQLQueryFilterVisitor sqfv = new SpeedoQLQueryFilterVisitor(
                    fields, sp, (ASTSpeedoQL) node, 
                    filterParserLogger, hparams, hvars, 
                    qd.candidateClass,
                    qb, jf);
            sp.setQueryFilter(sqfv.getQueryFilter());
        }
        assignMapper(sp);
        assignMapper(qt);

		QueryEvalContext qec = new QueryEvalContext(sp, this);
		
		SelectGroupByVisitor sgv = new SelectGroupByVisitor(
		        sp, qt, mapper, sqvv, qd, qec, classLoader);
		
		sgv.visit(qd);
		selectedFieldTypes = sgv.getSelectFieldTypes();
        assignMapper(qec.query);
		
		//Specify the ordering
        if (qd.order != null && qd.order.size() > 0) {
			OrderField[] ofs = new OrderField[qd.order.size()];
			for(int i=0; i<ofs.length; i++) {
				String o = (String) qd.order.get(i);
				int idx = o.indexOf(' ');
				boolean desc = false;
				if (idx != -1) {
					desc = o.substring(idx + 1).trim().equals("descending");
					o = o.substring(0, idx);
				}
				o = "this." + o;
				ofs[i] = new BasicOrderField((QueryTreeField)
					qt.getTupleStructure().getField(o), desc);
			}
			sp.setOrderBy(ofs);
		}

		logger.log(BasicLevel.INFO, "QueryTree built");
        if (debug) {
            QueryTreePrinter.printQueryTree(qec.query, logger);
        }
		//check for the use of the userCache
        if (qd.result == null && qd.variables == null) {
            //no variable used and the result is the candidate class
            Map field2value = new HashMap();
            if (getFieldComparaison(sp.getQueryFilter(), field2value)) {
                SpeedoHome sh = null;
                try {
                    sh = (SpeedoHome) jf.getPClassMapping(
                            qd.candidateClass.getName(),
                            classLoader);
                } catch (PException e) {
                    //never happen
                }
                userCache = sh.getUserCache(field2value.keySet());
                if (userCache != null) {
                    userCacheIndexes = new Operand[field2value.size()];
                    String[] ifs = userCache.getIndexFieldNames();
                    for (int i = 0; i < ifs.length; i++) {
                        userCacheIndexes[i] = (Operand) field2value.get(ifs[i]);
                    }
                }
            }
            
        }
        
		// Optimize the queryTree
		optimize(qec, debug);
		
		// Creates an evaluator associated to the QueryTree
		qec.evaluator = new MedorEvaluator(qec.query, 0);
		
		qecs = new QueryEvalContext[] {qec};
		timeToCompile = System.currentTimeMillis() - timeToCompile;
        status = COMPILED;
        logger.log(BasicLevel.INFO, "Query compiled in " + timeToCompile + "ms");
    }

    private boolean getFieldComparaison(Expression e, Map field2value) {
        if (e instanceof And || e instanceof ConditionalAnd) {
            return getFieldComparaison(((Operator) e).getExpression(0), field2value)
            	&& getFieldComparaison(((Operator) e).getExpression(1), field2value);
        } else if (e instanceof Equal){
            Expression tmpe = ((Operator) e).getExpression(0);
            if (!(tmpe instanceof Operand)) {
                return false;
            }
            Operand op0 = (Operand) tmpe;
            tmpe = ((Operator) e).getExpression(1);
            if (!(tmpe instanceof Operand)) {
                return false;
            }
            Operand op1 = (Operand) tmpe;
            //Equal between two operand
            if (op0 instanceof FieldOperand) {
            } else if (op1 instanceof FieldOperand) {
                //revert if op1 is 
                Operand o = op0;
                op0 = op1;
                op1 = o;
            } else {
                return false;
            }
            //op0 is a field operand
            if (op1 instanceof FieldOperand) { 
                return false;
            }
            // parameter operand or simple operand (constant)
            String fieldName = ((FieldOperand) op0).getField().getName();
            if (fieldName.indexOf('.') != -1) { // short or long navigation path
                if (fieldName.startsWith("this.")) { //path starts with "this."
                    fieldName = fieldName.substring(5); //remove "this."
                }
                if (fieldName.indexOf('.') != -1) { //long navigation path
                    return false;
                }
            }
            field2value.put(fieldName, op1);
            return true;
        }
        return false;
    }
    
    
    /**
     * Assign mappers and project and project name on the JormExtent nodes 
     * include in a QueryTree.
     * @throws SpeedoException
     */
    private void assignMapper(QueryTree qt)throws SpeedoException {
        //
        Collection extents = JormQueryTreeHelper.getJormExtents(qt);
        boolean debug = logger.isLoggable(BasicLevel.DEBUG); 
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Extent nodes: " + extents.size());
        }
        for (Iterator it = extents.iterator(); it.hasNext();) {
            JormExtent je = (JormExtent) it.next();
            try {
                PClassMapping pcm = jf.getPClassMapping(
                        JormPathHelper.getOriginClass(je.getJormName()),
                        classLoader);
                if (debug) {
                    logger.log(BasicLevel.DEBUG,
                            "JormExtent: " + je + " / pcm=" + pcm);
                }
                je.setPMapper(pcm.getPMapper(), pcm.getProjectName());
            } catch (PException e) {
                throw new SpeedoException(
                        "Error while fetching PClassPMapping of the class " 
                        + je.getJormName(), e);
            }
        }
    }

	private void optimize(QueryEvalContext pqc, boolean debug)
		throws MedorException, ExpressionException {
		ArrayList rules = new ArrayList();
		rules.add(new PushNotInExpressionRule());
		rules.add(new FlattenQueryTreeRule());
		//TODO: remove the dependency to JormFlatten2Rdb
		rules.add(new Jorm2Rdb());
		QueryTransformer queryTransformer = new BasicQueryRewriter(rules);
		IndexesGenerator indexesGenerator = new IndexesGenerator();
		try {
			pqc.query = queryTransformer.transform(pqc.query);
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Query optimized");
				QueryTreePrinter.printQueryTree(pqc.query, logger);
			}
			pqc.query = indexesGenerator.transform(pqc.query);
		} catch (Exception e) {
			throw new MedorException("Impossible to optimize the query", e);
		}

		pncParams = JormQueryTreeHelper.getRequiredPNameManagers(pqc.query);
		for (Iterator it = pncParams.iterator(); it.hasNext();) {
			ParameterOperand po = (ParameterOperand) it.next();
			String name = po.getName();
			if (debug) {
				logger.log(BasicLevel.DEBUG, "ParameterOperand " + name + " is expected.");
			}
			po.setValue(JormPathHelper.getPNameCoder(name, mapper));
		}
	}


    /**
     * evaluate the query with a single parameter which is a array of object
     * parameters.
     * @param pm    the persistence manager object
     * @param a    the array parameter of the query
     * @return  a Collection of result objects
     * @throws org.objectweb.medor.api.EvaluationException
     * @throws org.objectweb.medor.api.MedorException
     */
    public Object execute(Object[] a, ProxyManager  pm, QueryDefinition userqd)
            throws SpeedoException, MedorException, ExpressionException {
        if (status != COMPILED)
            throw new EvaluationException(
                    "Impossible to execute a query if it has not been defined and compiled before");
        ParameterOperand[] pos = null;
        if (a != null) {
            pos = new ParameterOperand[a.length + pncParams.size()];
            for(Iterator it = hparams.values().iterator(); it.hasNext();) {
				Object[] o = (Object[]) it.next();
				int idx = ((Integer) o[0]).intValue();
                pos[idx] = new BasicParameterOperand(
                    (BasicParameterOperand) o[1]);
				treatParameter(pos[idx], a[idx]);
            }
        } else {
            pos = new ParameterOperand[pncParams.size()];
        }
		if (pncParams.size() >0) {
			int i = (a == null ? 0 : a.length);
			for (Iterator it = pncParams.iterator(); it.hasNext(); i++) {
				pos[i++] = (ParameterOperand) it.next();
			}
		}
        return executeQT(pm, pos, userqd);
    }


    /**
     * evaluate the query with a single parameter which is a Map of object parameters.
     * @param pm    the persistence manager object
     * @param m    the map parameter of the query
     * @return  a Collection of result objects
     * @throws org.objectweb.medor.api.EvaluationException
     * @throws org.objectweb.medor.api.MedorException
     */
    public Object execute(Map m, ProxyManager  pm, QueryDefinition userqd)
            throws SpeedoException, MedorException, ExpressionException {
        if (status != COMPILED)
            throw new EvaluationException(
                    "Impossible to execute a query if it has not been defined and compiled before");
        ParameterOperand[] pos =
                new BasicParameterOperand[m.size() + pncParams.size()];
		for(Iterator it = hparams.values().iterator(); it.hasNext();) {
			Object[] o = (Object[]) it.next();
			int idx = ((Integer) o[0]).intValue();
			pos[idx] = new BasicParameterOperand(
				(BasicParameterOperand) o[1]);
			treatParameter(pos[idx], m.get(pos[idx].getName()));
			}
		if (pncParams.size() >0) {
			int i = (m == null ? 0 : m.size());
			for (Iterator it = pncParams.iterator(); it.hasNext(); i++) {
				pos[i++] = (ParameterOperand) it.next();
			}
		}
        return executeQT(pm, pos, userqd);
    }

	private void treatParameter(ParameterOperand po, Object value)
		throws SpeedoException, ExpressionException {
		Object val = value;
		if (po.getType().getTypeCode() == QType.TYPECODE_PNAME) {
			//Assign the pname of the parameter
			if (val == null) {
				try {
					val =
						jf.getPClassMapping(
						(Class) paramName2paramClass.get(po.getName()))
						.getPBinder().getNull();
				} catch (Exception e) {
					throw new SpeedoException(
						"Impossible to find the null PName of the parameter (name= "
						+ po.getName()
						+ " / type= " + po.getType().getJormName(),
						e);
				}
			} else if (val instanceof SpeedoProxy) {
				val = ((SpeedoProxy) value).getPName();
			} else {
				throw new JDOUserException("The parameter '"
					+ po.getName()
					+ "' must be a persistent object: " + val);
			}
		}
		po.setValue(val);
	}

    public short getStatus() {
        return status;
    }

    // Private Methods //
    //-----------------//

    /**
     * Hash a String, and compute a Hashtable
     * example:
     *   ("String name, Float salary, Employee boss", ",")
     *   keys     | values
     *   ---------------------
     *   "name"   | "String"
     *   "salary" | "Float"
     *   "boss"   | "Employee"
     *
     * @param stringToHash	the String to hash
     * @param separator		the separator to tokenize
     */
    private void toHashtableParams(String stringToHash, String separator) {
        hparams = new HashMap();
	    paramName2paramClass = new HashMap();
        if (stringToHash != null) {
            StringTokenizer tok = new StringTokenizer(stringToHash, separator);
			int idx = 0;
            while (tok.hasMoreTokens()) {
                String tuple = tok.nextToken().trim();
                int i = tuple.indexOf(' ');
                String name = tuple.substring(i + 1, tuple.length());
	            String paramType = tuple.substring(0, i).trim();
	            PType type = getPType(paramType);
                hparams.put(name, new Object[]{
					new Integer(idx),
					new BasicParameterOperand(type, name)});
				idx ++;
	            if (type == PTypeSpaceMedor.PNAME) {
	                paramName2paramClass.put(name, getClass(paramType));
                }
            }
        }
    }

    private void toHashtableVars(String stringToHash, String separator) {
        hvars = new HashMap();
        if (stringToHash != null) {
            StringTokenizer tok = new StringTokenizer(stringToHash, separator);
            while (tok.hasMoreTokens()) {
                String tuple = tok.nextToken().trim();
                int i = tuple.indexOf(' ');
                String type = tuple.substring(0, i);
                hvars.put(tuple.substring(i + 1, tuple.length()), type);
            }
        }
    }

	private PType getPType(String name) {
		if ("String".equals(name)) {
			return PTypeSpace.STRING;
		}
		if ("Integer".equals(name)) {
			return PTypeSpace.OBJINT;
		}
		for (int i = 0; i < PTypeSpace.PREDEFINEDPTYPES.length; i++) {
			PType type = PTypeSpace.PREDEFINEDPTYPES[i];
			if (type.getJavaName().equals(name)
			    || type.getJormName().equals(name)) {
				return type;
			}
		}
		if ("Collection".equals(name) 
		        || PTypeSpaceMedor.COLLECTION.getJavaName().equals(name)) {
		    return PTypeSpaceMedor.COLLECTION;
		}
		return PTypeSpaceMedor.PNAME;
	}
    private Class getClass(String classname) {
        if (classLoader != null) {
            try {
                return classLoader.loadClass(classname);
            } catch (ClassNotFoundException e) {
                try {
                    return classLoader.loadClass(
                            qd.candidateClass.getPackage().getName() + "." + classname);
                } catch (ClassNotFoundException e2) {
                    return null;
                }
            }
        } else {
            return null;
        }
    }

    /**
     * executes a the current query, and returns a Collection of object.
     * @param userqd is the user query definition. It contains the range values
     * and some other values that can change from the original compiled query.
     * @return  a new Collection of objects.
     * @throws org.objectweb.medor.api.EvaluationException
     * @throws org.objectweb.medor.api.MedorException
     */
	private Object executeQT(ProxyManager pm, ParameterOperand[] pos, QueryDefinition userqd)
        throws EvaluationException, MedorException, SpeedoException {
		if (!qd.ignoreCache || !pm.getIgnoreCache()) {
			if (logger.isLoggable(BasicLevel.DEBUG)) {
			    logger.log(BasicLevel.DEBUG,
			    	"Flush dirty instances of the working set ...");
			}
			try {
				pm.getTransactionalPersistenceManager()
						.flush((Transaction) pm.currentTransaction(), this);
			} catch (PersistenceException e) {
				throw new JDOException("Impossible to use the cache into a " +
						"query: Error during the flushing of data on the support", e);
			}
		}
		if (userCache != null) { //the query matches to an unique index
		    Object key = null;
		    //building the key
		    if (userCacheIndexes.length == 1) {
		        //Index is single
		        key = getValueFromOperand(userCacheIndexes[0], pos);
		    } else {
		        //key is composite
		        key = new UserCacheKey(userCacheIndexes.length);
		        for (int i = 0; i < userCacheIndexes.length; i++) {
		            ((UserCacheKey) key).setPart(i, 
		                    getValueFromOperand(userCacheIndexes[0], pos));
                }
		    }
		    // search in the user cache
		    key = userCache.lookup(key);
		    if (key != null) {
		        try {
                    if (key instanceof PName) { //the cache contains PName
                        key = pm.getObjectById(key, false);
                    }
                    if (qd.unique) {
                        return key;
                    } else {
                        return Collections.singletonList(key);
                    }
                } catch (JDOException e) {
                    logger.log(BasicLevel.WARN, 
                            "User cache is not up to date:", e);
                }
		    }
		}
		
		Object connection = ((Transaction) pm.currentTransaction())
				.getConnectionHolder();
		TupleCollection queryResult = null;
		if (qecs.length == 1) {
			//optimization: avoid to use an intermediate class
			queryResult = qecs[0].eval(pm, pos, connection, userqd);
			if (queryResult == null || queryResult.isEmpty()) {
				logger.log(BasicLevel.INFO,
						"Be careful, the result is empty");
				if (queryResult != null) {
				    queryResult.close();
				}
				QueryResultCommon.closeConnection(connection);
				if (qd.unique) {
				    return null;
				} else {
				    return Collections.EMPTY_SET;
				}
			}
		} else {
			//Use a multiplex of query
			queryResult = new QueriesUnion(pos, pm, connection, qecs, userqd);
		}
		if (qd.unique) {
		    return new QueryResultUnique(
		            queryResult, 
		            pm, 
		            new Object[]{connection},
		            qd.resultClass,
		            selectedFieldTypes,
		            qecs.length == 1,
		            logger).getResult();
		} else {
		    return new QueryResultList(
		            queryResult, 
		            pm, 
		            new Object[]{connection},
		            qd.resultClass,
		            selectedFieldTypes,
		            qecs.length == 1,
		            logger);
		}
    }

	private Object getValueFromOperand(Operand op, ParameterOperand[] pos) {
	    if (op instanceof ParameterOperand) {
	        String param = ((ParameterOperand) op).getName();
	        for (int i = 0; i < pos.length; i++) {
                if (pos[i].getName().equals(param)) {
                    return pos[i].getObject();
                }
            }
		    return null;
	    } else {
	        return op.getObject();
	    }
	}
	
	// IMPLEMENTATION OF THE CacheEntryFilter INTERFACE //
	//--------------------------------------------------//

	public boolean accept(State ce) {
		if (logger.isLoggable(BasicLevel.DEBUG)) {
		    logger.log(BasicLevel.DEBUG,
		    	"Flush dirty instance of the working state, identifier: " 
		            + ce.getCacheEntry().getCeIdentifier());
		}
		return true;
	}


}