/**
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 26 mars 2005
 *
 */
package org.objectweb.speedo.tools;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import org.objectweb.asm.Attribute;
import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Constants;
import org.objectweb.asm.attrs.Attributes;

/**
 * This ASM visitor assignes a value to a static field. If the field does not
 * exist, it is added. 
 *
 * @author S.Chassande-Barrioz
 */
public class ClassFieldModifier extends ClassAdapter {

    private String className;
    
    /**
     * The name of the field to modify
     */
    private String fieldName;
    
    /**
     * The value to set to the field
     */
    private String fieldValue;
    
    /**
     * indicates if the field has been found in the class
     */
    private boolean fieldFound;
    
    private boolean classModified = false;
    
    /**
     * @param fieldName is the name of the field to modify
     * @param fieldValue is the value to set to the field
     */
    public ClassFieldModifier(final ClassVisitor cv, 
            final String fieldName, 
            final String fieldValue) {
        super(cv);
        this.fieldName = fieldName; 
        this.fieldValue = fieldValue; 
        fieldFound = false;
    }

    public void visit(final int version, final int access,
            final String name,
            final String superName,
            final String[] interfaces,
            final String sourceFile) {
        className = name;
        super.visit(version, access, name, superName, interfaces, sourceFile);
    }
    public void visitField(final int access,
            final String name,
            final String desc,
            final Object value,
            final Attribute attrs) {
        Object val = value;
        if (name.equals(fieldName)) {
            fieldFound = true;
            if (value == null || !value.equals(fieldValue)) {
                val = fieldValue;
                classModified = true;
                System.out.println("Set the field '" + fieldName 
                        + "' of the class '" + className 
                        + "' to '" + val 
                        + "' (old value: '" + value + "').");
            }
        }
        super.visitField(access, name, desc, val, attrs);
    }

    public void visitEnd() {
        if (!fieldFound) {
            classModified = true;
            super.visitField(Constants.ACC_PUBLIC 
                    | Constants.ACC_STATIC 
                    | Constants.ACC_FINAL, 
                fieldName, 
                "Ljava/lang/String;", 
                fieldValue, 
                null);
            System.out.println("Add the field '" + fieldName 
                    + "' to the class '" + className 
                    + "' with the value '" + fieldValue + "'.");
        }
        super.visitEnd();
    }

    public boolean isClassModified() {
        return classModified;
    }
    
    public static void main(String[] args) throws IOException {
        if (args == null || args.length < 3) {
            usage();
            return;
        }
        String classFileName = args[0];
        String fieldName = args[1];
        String fieldValue = args[2];
        ClassReader cr;
        File f = new File(classFileName);
        FileInputStream fis = new FileInputStream(f);
        cr = new ClassReader(fis);
        fis.close();
        ClassWriter cw = new ClassWriter(true);
        ClassFieldModifier cfm = new ClassFieldModifier(cw, fieldName, fieldValue);
        cr.accept(cfm, Attributes.getDefaultAttributes(), false);
        if (cfm.isClassModified()) {
	        FileOutputStream fos = new FileOutputStream(f);
	        fos.write(cw.toByteArray());
	        fos.close();
        }
    }
        
    public final static void usage() {
        System.out.println("Usage: <class file> <field name> <field value>");
    }
}
