/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.usercache.lib;

import org.objectweb.speedo.usercache.api.UserCache;

import java.util.HashMap;
import java.util.Map;

/**
 * Is a basic UserCacheView implementation using WeakReference in order to 
 * not disturb the real cache of persistent object.
 * 
 * @author S.Chassande-Barrioz
 */
public class UserCacheImpl implements UserCache {

    public final static int DEFAULT_SIZE = 50;
    
	/**
	 * Contains the association user key TO object identifier
	 */
	protected Map key2oid;

	/**
	 * Contains the association object identifier TO user key
	 */
	protected Map oid2key;

    /**
     * The name of the user cache
     */
	protected String name;
	
	/**
	 * The identifier of the user cache (light weight pattern)
	 */
	protected int id;
    
    /**
     * Indicates if the user cache is active
     */
	protected boolean active;
    
    /**
     * is the list of field name used for the cache index
     */
	protected String[] fields;

	/**
	 * Allocates an user cache with the default size.
	 * @see #DEFAULT_SIZE
     * @see #UserCacheImpl(int)
	 */
	public UserCacheImpl() {
	    this(DEFAULT_SIZE);
	}

	/**
	 * Allocates an user cache with a particular size. The size is not a maximal
	 * size, but serves only for the underlying underlying hash structure.
	 *   
	 * @param size is the size of the cache.
	 */
	public UserCacheImpl(int size) {
	    key2oid = new HashMap(size);
	    oid2key = new HashMap(size);
	}

	public synchronized Object bind(Object key, Object oid) {
	    Object oldKey = oid2key.put(oid, key);
	    if (oldKey != null && !oldKey.equals(key)) { 
	        //rebind with another key
	        key2oid.remove(oldKey);
	    } else {
		    key2oid.put(key, oid);
	    }
	    return oldKey;
	}

	public synchronized Object unbindFromKey(Object key) {
	    Object oldOID = key2oid.remove(key);
	    if (oldOID != null) {
	        oid2key.remove(oldOID);
	    }
	    return oldOID;
	}
	
	public synchronized Object unbindFromOID(Object oid) {
	    Object oldKey = oid2key.remove(oid);
	    if (oldKey != null) {
	        key2oid.remove(oldKey);
	    }
	    return oldKey;
	}
	
	public synchronized Object lookup(Object key) {
	    return key2oid.get(key);
	}
	
    public int getId() {
        return id;
    }
    
    public void setId(int _id) {
        this.id = _id;
    }
    public String[] getIndexFieldNames() {
        return fields;
    }
    public void setIndexFieldNames(String[] fns) {
        this.fields = fns;
    }
    public String getName() {
        return name;
    }
    public void setName(String n) {
        this.name = n;
    }
    public boolean isActive() {
        return active;
    }
    public void setActive(boolean a) {
        this.active = a;
    }
}
