/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.workingset.lib;

import java.util.ArrayList;
import java.util.Iterator;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.naming.InitialContext;
import javax.transaction.Status;
import javax.transaction.Synchronization;
import javax.transaction.TransactionManager;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.api.VirtualState;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSetLifeCycle;
import org.objectweb.perseus.persistence.lib.BasicWorkingSet;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.TransactionListener;
import org.objectweb.speedo.mim.api.SpeedoAccessor;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Is a working set assocaited to a ProxyManager. The working set can be
 * transactional or not. It contains the list of reached instances.
 *
 * @see javax.jdo.Transaction
 * @see org.objectweb.speedo.workingset.api.Transaction
 * @see org.objectweb.perseus.persistence.api.WorkingSet
 * @see org.objectweb.perseus.persistence.lib.BasicWorkingSet
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoTransaction
        extends BasicWorkingSet
        implements Transaction, LifeCycleController {

    public final static String PROXY_MANAGER_BINDING = "proxy-manager";
    public final static String MAPPER_BINDING = "mapper";
    public final static String TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING
            = "transactional-persistence-manager";
    public final static String COMPONENT_BINDING = "component";
	public static TransactionListener txListener = null;

    /**
	 * Is the linked Proxy manager.
	 */
	private ProxyManager pm = null;

	/**
	 * is the mapper permitting to reach the prefetch cache and to invalidate
	 * prefetched buffer at working set closing time.
	 */
	private PMapper mapper = null;

	/**
	 * Is used to delegates working set/transaction demercation
	 */
	private TransactionalPersistenceManager tpm = null;

	/**
	 * the reference to this component in Speedo
	 */
	private Transaction thisT;

    private boolean nontransactionalRead;
    private boolean nontransactionalWrite;

	/**
	 * Indicates if the transaction is optimistic.
	 */
	private boolean optimistic;

	/**
	 * The JDO user synchronization registered (can be null if none has been
	 * registered).
	 */
	private Synchronization synchronization = null;

	/**
	 * indicates if the transaction is managed by a J2EE environnement.
	 */
	private boolean managedEnv = false;
	
	/**
	 * Indicates if the jdo transaction must be rolledback.
	 */
	private boolean rollbackOnly = false;

    /**
     * It activates the working set. This is used to delimit the begining of
     * the working set.
     */
    public void active() throws PersistenceException {
        tpm.createWS(thisT);
        try {
            status = WorkingSetLifeCycle.getNextStatus(
                    status, WorkingSetLifeCycle.ACTIVE_ACTION);
        } catch (PersistenceException e) {
            logger.log(BasicLevel.WARN, "Bad initial state of the working set:", e);
            status = WorkingSet.CTX_ACTIVE;
        }
    }

	/**
	 * Attaches an entry to the transaction.
	 * Plus version update.
	 * @param state the state which must be attached to the transaction
	 * @param mode the action that stared the binding: either read or write intention
	 */
	public State bind(State state, Object oid, byte mode) {
		State old = super.bind(state, oid, mode);
		if(! (state instanceof VirtualState)){
			SpeedoAccessor sa = (SpeedoAccessor) state;
			if (mode == BasicWorkingSet.WRITE_INTENTION) { 
				sa.changeVersion();
			}
		}
		return old;
	}

	/**
	 * Invalidates the prefetch buffer associated to this working set.
	 *
	 * @throws PersistenceException
	 */
	public void beforeWSPrepare() throws PersistenceException {
		logger.log(BasicLevel.DEBUG, "Starting beforeWSPrepare");
		Iterator it = oid2state.values().iterator();
		ArrayList exceptions = null;
		while(it.hasNext()) {
			State state = (State) it.next();
			if (state == VirtualState.instance) {
				continue;
			}
			try {
				((SpeedoAccessor) state).prepareWrite();
			} catch (Exception e) {
				if (exceptions == null) {
					exceptions = new ArrayList();
				}
				exceptions.add(e);
				int level = e instanceof JDOException ? BasicLevel.DEBUG : BasicLevel.ERROR ; 
			    if (logger.isLoggable(level)) {
			        logger.log(level, 
				        "Error on SpeedoAccessor preparation for flushing: "
				        + "\n\tstate.ce.identifier=" + state.getCacheEntry().getCeIdentifier()
				        + "\n\tstate=" + state
				        + "\n\texception: ", e);
				}
			}
		}
		// close the prefetch buffers associated to the context
		PrefetchCache pc = mapper.getPrefetchCache();
		if (pc != null) {
			pc.invalidatePrefetchBuffer(thisT);
		}
		logger.log(BasicLevel.DEBUG, "Ending beforeWSPrepare");
		if (exceptions != null) {
			Exception inner; 
			if (exceptions.size() == 1) {
			    throw new PersistenceException((Exception) exceptions.get(0));
			} else {
			    throw new PersistenceException(new JDOUserException(
					"Impossible to prepare instances before flushing",
					(Exception[]) exceptions.toArray(new Exception[exceptions.size()])));
			}
		}
	}

	/**
	 * Signal to the persistent instances reached in the working set that the
	 * current working set is closed. Some actions on persistent instances at
	 * this time can be done, such as reference unswizlling
	 */
	public void onWSEnd() {
		logger.log(BasicLevel.DEBUG, "Starting onWSEnd");
		ArrayList exceptions = null;
		if (!oid2state.isEmpty()) {
			Iterator it = oid2state.values().iterator();
			while(it.hasNext()) {
				State state = (State) it.next();
				if (state == VirtualState.instance) {
					continue;
				}
				try {
					((SpeedoAccessor) state).workingSetClosed();
				} catch (Exception e) {
					if (exceptions == null) {
						exceptions = new ArrayList();
					}
					exceptions.add(e);
					if (!(e instanceof JDOException)) {
					    logger.log(BasicLevel.ERROR, 
					        "Error on workingSetClosed for the SpeedoAccessor: "
					        + "\n\tstate.ce.identifier=" + state.getCacheEntry().getCeIdentifier()
					        + "\n\tstate=" + state
					        + "\n\texception: ", e);
					}
				}
			}
		}
		logger.log(BasicLevel.DEBUG, "Ending onWSEnd");
		if (exceptions != null) {
			Exception inner; 
			if (exceptions.size() == 1) {
				throw new JDOException(
				        "Error when signal the close of working set on state(s):", 
				        (Exception) exceptions.get(0));
			} else {
				throw new JDOUserException(
					"Error when signal the close of working set on states:",
					(Exception[]) exceptions.toArray(new Exception[exceptions.size()]));
			}
		}
	}

	// IMPLEMENTATION OF THE Transaction INTERFACE //
	//-----------------------------------------------------//

	public void setConnectionHolder(ConnectionHolder ch) {
		connectionHolder = ch;
		connectionHolder.bindWorkingSet(thisT);
	}

	public JDOFatalException rollBackOnInternalError(Exception _e) {
		Exception ie = ExceptionHelper.getNested(_e);
		if (isActive()) {
			if (managedEnv) {
				logger.log(BasicLevel.INFO, ".");
				String tmName = pm.getPersistenceManagerFactory().getProperties()
						.getProperty(SpeedoProperties.TM_NAME);
				if (tmName == null) {
					return new JDOFatalException("No transaction manager jndi name found in initialisation properties");
				}
				try {
					Object o = new InitialContext().lookup(tmName);
					if (o == null) {
						String msg = "The transaction must be marked as rollbackOnly: JNDI retrieves a null transaction manager for the name '" + tmName + "'.";
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					if (!(o instanceof TransactionManager)) {
						String msg = "The transaction must be marked as rollbackOnly: JNDI retrieves an object which is not a javax.transaction.TransactionManager (JNDI name: " + tmName + "): " + o;
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					javax.transaction.Transaction t = ((TransactionManager) o).getTransaction();
					if (t == null) {
						String msg = "Impossible to rollback an unexisting transaction (TM.getTransaction()=null)";
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					t.setRollbackOnly();
				} catch (Exception e) {
					String msg = "The transaction must be marked as rollbackOnly: Error when lookup the transaction manager in JNDI with the name '"
							+ tmName + "'";
					logger.log(BasicLevel.ERROR, msg, e);
					return new JDOFatalException(msg);
				}
			} else {
				rollback();
			}
			return new JDOFatalException("The current transaction has been rolledback, please retry", ie);
		} else {
			return new JDOFatalException("The current working set occrurs an error, please retry", ie);
		}
	}

    public boolean isManagedEnv() {
        return managedEnv;
    }

    // IMPLEMENTATION OF THE LifeCycleController INTERFACE //
    //-----------------------------------------------------//

    public String getFcState() {
        return null;
    }

    public void startFc() {
        managedEnv = pm.getPersistenceManagerFactory().getProperties()
            .getProperty(SpeedoProperties.MANAGED, "")
            .equals("true");
    }

    public void stopFc() {
    }

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public String[] listFc() {
        String[] names = super.listFc();
        String[] itfs = new String[names.length + 3];
        itfs[0] = PROXY_MANAGER_BINDING;
        itfs[1] = TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING;
        itfs[2] = MAPPER_BINDING;
        System.arraycopy(names, 0, itfs, 3, names.length);
        return itfs;
    }

    public Object lookupFc(String c) {
        if (PROXY_MANAGER_BINDING.equals(c))
            return pm;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            return tpm;
        else if (MAPPER_BINDING.equals(c))
            return mapper;
        else
            return super.lookupFc(c);
    }

    public void bindFc(String c, Object s) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = (ProxyManager) s;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = (TransactionalPersistenceManager) s;
        else if (MAPPER_BINDING.equals(c))
            mapper = (PMapper) s;
        else if (COMPONENT_BINDING.equals(c)) {
            try {
                thisT = (Transaction) ((Component) s).getFcInterface("transaction");
            } catch (Exception e) {
                throw new JDOFatalInternalException(
                        "Impossible to get self transaction",
                        ExceptionHelper.getNested(e));
            }
        } else
            super.bindFc(c, s);
    }

    public void unbindFc(String c) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = null;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = null;
        else if (MAPPER_BINDING.equals(c))
            mapper = null;
        else
            super.unbindFc(c);
    }

	public void setStatus(byte status) throws PersistenceException {
		try {
		    switch(status) {
		    case TransactionalWorkingSet.CTX_PREPARED:
				beforeWSPrepare();
		    	break;
			case TransactionalWorkingSet.CTX_COMMITTED:
			case TransactionalWorkingSet.CTX_ABORTED:
				onWSEnd();
				break;
			case TransactionalWorkingSet.CTX_CLOSED:
				if (oid2state.isEmpty()) {
					// close the prefetch buffers associated to the context
					PrefetchCache pc = mapper.getPrefetchCache();
					if (pc != null) {
						pc.invalidatePrefetchBuffer(thisT);
					}
				}
				onWSEnd();
				break;
			}
		} finally {
			super.setStatus(status);
		}
	}

    // IMPLEMENTATION OF THE javax.jdo.Transactionn INTERFACE //
    //--------------------------------------------------------//

    public void begin() {
		logger.log(BasicLevel.INFO, "Begin the transaction");
        try {
            rollbackOnly = false; //initialize the flag
            tpm.begin(thisT);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR,
                    "Error during the begin of the transaction:", ie);
            throw new JDOException("", ie);
        }
		if (txListener != null) {
			txListener.transactionBegun(this);
		}
    }

    public void commit() {
		int size = oid2state.size();
        logger.log(BasicLevel.INFO, "Commit the transaction, working set size: " + size);
        if (synchronization != null) {
            synchronization.beforeCompletion();
        }
		//register working set size for statistics
        boolean validated = rollbackOnly;
        try {
            if (rollbackOnly) {
                tpm.rollback(thisT);
            } else {
                tpm.commit(thisT);
    			validated = true;
            }
        } catch (PersistenceException e) {
			Exception ie = ExceptionHelper.getNested(e);
			if (ie instanceof JDOException) {
				throw (JDOException) ie;
			} else {
			    ie = new JDOFatalException(
					"Transaction rolledback due to an exception at commit time: ", ie);
			    logger.log(BasicLevel.INFO, ie.getMessage(), ie);
			    throw (JDOFatalException) ie;
			}
        } finally {
			if (synchronization != null)
				synchronization.afterCompletion((validated
						? Status.STATUS_COMMITTED : Status.STATUS_ROLLEDBACK));
			if (txListener != null) {
				if (validated) {
					txListener.transactionCommitted(this, size);
				} else {
					txListener.transactionAborted(this, size);
				}
			}
        }
    }

    public void rollback() {
        logger.log(BasicLevel.INFO, "Roll back a transaction: ");
		int size = oid2state.size();
        try {
            tpm.rollback(thisT);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR,
                    "Error during the rollback of the transaction:", ie);
            throw new JDOException("", ie);
        } finally {
			if (txListener != null) {
				txListener.transactionAborted(this, size);
			}
			if (synchronization != null) {
				synchronization.afterCompletion(Status.STATUS_ROLLEDBACK);
			}
		}
    }

    public boolean isActive() {
        switch(status){
        case TransactionalWorkingSet.CTX_ACTIVE_TRANSACTIONAL:
        case TransactionalWorkingSet.CTX_PREPARED:
        case TransactionalWorkingSet.CTX_PREPARED_OK:
        case TransactionalWorkingSet.CTX_PREPARED_FAIL:
            return true;
        case TransactionalWorkingSet.CTX_ACTIVE:
        case TransactionalWorkingSet.CTX_COMMITTED:
        case TransactionalWorkingSet.CTX_ABORTED:
        case TransactionalWorkingSet.CTX_CLOSED:
        default:
            return false;
        }
    }

    public void setNontransactionalRead(boolean b) {
        nontransactionalRead = b;
    }

    public boolean getNontransactionalRead() {
        return nontransactionalRead;
    }

    public void setNontransactionalWrite(boolean b) {
        nontransactionalWrite = b;
    }

    public boolean getNontransactionalWrite() {
        return nontransactionalWrite;
    }

    public void setRetainValues(boolean b) {
        setWSRetainValues(b);
    }

    public boolean getRetainValues() {
        return getWSRetainValues();
    }

    public void setRestoreValues(boolean b) {
        setWSRestoreValues(b);
    }

    public boolean getRestoreValues() {
        return getWSRestoreValues();
    }

    public void setOptimistic(boolean b) {
        optimistic = b;
    }

    public boolean getOptimistic() {
        return optimistic;
    }

    public void setSynchronization(Synchronization s) {
        synchronization = s;
    }

    public Synchronization getSynchronization() {
        return synchronization;
    }

    public PersistenceManager getPersistenceManager() {
        return pm;
    }
    
    public boolean getRollbackOnly() {
        return rollbackOnly;
    }

    public void setRollbackOnly() {
        rollbackOnly = true;
    }
}
