/**
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 17 fvr. 2005
 *
 */
package org.objectweb.speedo.runtime.collection;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;

import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.pobjects.collection.Group;
import org.objectweb.speedo.pobjects.collection.User;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * It tests the thin locking on a collection.
 *
 * @author S.Chassande-Barrioz
 */
public class TestThinLock extends SpeedoTestHelper {

    /**
     * @param s
     */
    public TestThinLock(String s) {
        super(s);
    }

    /**
     * @see org.objectweb.speedo.SpeedoTestHelper#getLoggerName()
     */
    protected String getLoggerName() {
		return LOG_NAME + ".rt.collection.TestThinLock";
    }

    /**
     * Specify the properties activating the thin lock for the Group.users
     * collection field.
     * @see org.objectweb.speedo.SpeedoTestHelper#getPMFProperties()
     */
	public Properties getPMFProperties() {
	    Properties p = super.getPMFProperties();
	    p.put(SpeedoProperties.TRANSACTION_LOCKING_LEVEL_ENABLETHIN, "true");
	    p.put(SpeedoProperties.TRANSACTION_LOCKING_LEVEL + "(" 
	            + Group.class.getName() + "#users)", 
	            SpeedoProperties.TRANSACTION_LOCKING_LEVEL_FIELD);
	    return p;
	}
	
	public void testGroupUser2WithBarrier() {
	    testGroupUser(2,true);
	}
	public void testGroupUser20WithBarrier() {
	    testGroupUser(20,true);
	}
	public void testGroupUser20() {
	    testGroupUser(20,false);
	}
	public void testGroupUser50() {
	    testGroupUser(50,true);
	}
	public void testGroupUser100() {
	    testGroupUser(100,false);
	}
	
	/**
	 * Create A group with several users
	 *
	 */
	public void testGroupUser(final int NB, final boolean withBarrier) {
	    logger.log(BasicLevel.INFO, "testGroupUser" + NB + 
	            (withBarrier ? "WithBarrier" : ""));
	    final Waiter waiter = new Waiter(0, withBarrier); 
	    final Group group = new Group("TestThinLock.testGroupUser.group1");
	    final User[] users = new User[NB];
	    for(int i=0; i<NB; i++) {
	        users[i] = new User("TestThinLock.testGroupUser.user" + i); 
		    group.getUsers().add(users[i]);
	    }
	    PersistenceManager pm = pmf.getPersistenceManager();
	    pm.currentTransaction().begin();
	    pm.makePersistent(group);
	    pm.currentTransaction().commit();
	    pm.close();
	    Thread[] ths = new Thread[NB];
	    for(int i=0; i<NB; i++) {
	        ths[i] =  new Thread(new TestGroupUser(waiter, group, users[i],
	                    new User("TestThinLock.testGroupUser.user" + (NB + i)),
	                    pmf)); 
    	}
	    for(int i=0; i<NB; i++) {
	        ths[i].start();
	    }
	    Collection threads = Arrays.asList(ths);
	    List res = waiter.nextAction(threads, 1000);
	    assertTrue("Thread blocked on first point: " + res, res.isEmpty());
	    res = waiter.nextAction(threads, 1000);
	    if (!res.isEmpty()) {
		    for(int i=0; i<ths.length; i++) {
		        if (ths[i].isAlive()) {
		            ths[i].interrupt();
		        }
		    }
	    }
	    assertTrue("Thread blocked on second point: " + res, res.isEmpty());
	    res = waiter.nextAction(threads, 1000);
	    for(int i=0; i<ths.length; i++) {
	        if (ths[i].isAlive()) {
	            try {
                    ths[i].join(1000);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
	        }
	    }
	    pm = pmf.getPersistenceManager();
	    int size = group.getUsers().size();
	    pm.currentTransaction().begin();
	    pm.deletePersistentAll(group.getUsers());
	    pm.deletePersistent(group);
	    pm.currentTransaction().commit();
	    pm.close();
	    assertEquals("Bad number of users", NB, size);
	}
    
	
	public static class Waiter {
	    int actionId;
	    ArrayList waiters = new ArrayList();
	    boolean withBarrier;
	    public Waiter(int actionId, boolean withBarrier) {
	        this.actionId = actionId;
	        this.withBarrier = withBarrier;
	    }
	    public synchronized void waitAction(int aid) {
	        if (!withBarrier) {
	            return;
	        }
	        waiters.add(Thread.currentThread());
	        notifyAll();
	        while(actionId < aid) {
	            try {
                    wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
	        }
	    }
	    public synchronized List nextAction(Collection threads, long time) {
	        if (!withBarrier) {
	            return Collections.EMPTY_LIST;
	        }
		    while(!waiters.containsAll(threads)) {
		        try {
	                wait(time);
	            } catch (InterruptedException e) {
	                e.printStackTrace();
	                break;
	            }
		    }
		    if (waiters.containsAll(threads)) {
		    	actionId++;
		    	waiters.clear();
		    	notifyAll();
		    	return Collections.EMPTY_LIST;
		    } else {
		        return Collections.unmodifiableList(
		                new ArrayList(waiters));
		    }
	    }
	}
    
    public static class TestGroupUser implements Runnable {
        public Waiter waiter;
        public Group group;
        public User userToRemove;
        public User userToAdd;
        public PersistenceManagerFactory pmf;
        
        public TestGroupUser(Waiter waiter, 
                Group group, 
                User userToRemove,
                User userToAdd, 
                PersistenceManagerFactory pmf) {
            super();
            this.waiter = waiter;
            this.group = group;
            this.userToRemove = userToRemove;
            this.userToAdd = userToAdd;
            this.pmf = pmf;
        }
        
        public void run() {
            PersistenceManager pm = pmf.getPersistenceManager();
            pm.currentTransaction().begin();
            Collection users = group.getUsers();
            waiter.waitAction(1);
            
            users.add(userToAdd);
            users.remove(userToRemove);
            pm.deletePersistent(userToRemove);
            
            waiter.waitAction(2);
            pm.currentTransaction().commit();
            pm.close();

            waiter.waitAction(3);
        }
    }
}
