/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo.runtime.concurrency;

import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.pobjects.basic.BasicA;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;

import junit.framework.Assert;

import java.util.Properties;

public class TestPessimistic extends SpeedoTestHelper {

	public TestPessimistic(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return LOG_NAME + ".rt.concurrency.TestPessimistic";
	}

	public Properties getPMFProperties() {
		Properties p = super.getPMFProperties();
		p.setProperty("javax.jdo.option.Optimistic", "false");
		return p;
	}
	
    public void testConcurrentReads() {
        logger.log(BasicLevel.INFO, "test concurrent reads");
        // setup
        BasicA ba = new BasicA();
        ba.writeF1("1");
        ba.writeF2(2);
        PersistenceManager pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        pm.makePersistent(ba);
        final Object id = pm.getObjectId(ba);
        Assert.assertNotNull("Null object identifier", id);
        pm.currentTransaction().commit();
        pm.close();

        ba = null;

        // begin transaction 1
        PersistenceManager pm1 = pmf.getPersistenceManager();
        logger.log(BasicLevel.INFO, "Thread1: begin");
        pm1.currentTransaction().begin();
        logger.log(BasicLevel.INFO, "Thread1: getObjectById");
        ba = (BasicA) pm1.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        logger.log(BasicLevel.INFO, "Thread1: read");
        ba.readF2();
        logger.log(BasicLevel.INFO, "Thread1: read done");

        // begin transaction 2
        Thread t = new Thread((new Runnable () {
            public void run () {
                PersistenceManager pm2 = pmf.getPersistenceManager();
                logger.log(BasicLevel.INFO, "Thread2: begin");
                pm2.currentTransaction().begin();
                logger.log(BasicLevel.INFO, "Thread2: getObjectById");
                BasicA ba = (BasicA) pm2.getObjectById(id, true);
                Assert.assertNotNull("Object not found", ba);
                logger.log(BasicLevel.INFO, "Thread2: read");
                ba.readF2();
                logger.log(BasicLevel.INFO, "Thread2: read done");
                logger.log(BasicLevel.INFO, "Thread2: commit");
                pm2.currentTransaction().commit();
                pm2.close();
            }
        }));
        t.start();

        // end transaction 1
        logger.log(BasicLevel.INFO, "Thread1: commit");
        pm1.currentTransaction().commit();
        pm1.close();

        assertUnlock(t);

        // clean up
        pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        ba = (BasicA) pm.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        Assert.assertEquals("Bad f1 value", "1", ba.readF1());
        Assert.assertEquals("Bad f2 value", 2, ba.readF2());
        pm.deletePersistent(ba);
        pm.currentTransaction().commit();
        pm.close();
    }

    public void testReadConcurrentWrite() {
        logger.log(BasicLevel.INFO, "test read concurrent write");
        // setup
        BasicA ba = new BasicA();
        ba.writeF1("1");
        ba.writeF2(2);
        PersistenceManager pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        pm.makePersistent(ba);
        final Object id = pm.getObjectId(ba);
        Assert.assertNotNull("Null object identifier", id);
        pm.currentTransaction().commit();
        pm.close();

        ba = null;

        // begin transaction 1
        PersistenceManager pm1 = pmf.getPersistenceManager();
        logger.log(BasicLevel.INFO, "Thread1: begin");
        pm1.currentTransaction().begin();
        logger.log(BasicLevel.INFO, "Thread1: getObjectById");
        ba = (BasicA) pm1.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        logger.log(BasicLevel.INFO, "Thread1: read");
        ba.readF2();
        logger.log(BasicLevel.INFO, "Thread1: read done");

        // begin transaction 2
        Thread t = assertLock(new Runnable () {
            public void run () {
                PersistenceManager pm2 = pmf.getPersistenceManager();
                logger.log(BasicLevel.INFO, "Thread2: begin");
                pm2.currentTransaction().begin();
                logger.log(BasicLevel.INFO, "Thread2: getObjectById");
                BasicA ba = (BasicA) pm2.getObjectById(id, true);
                Assert.assertNotNull("Object not found", ba);
                logger.log(BasicLevel.INFO, "Thread2: write");
                ba.writeF2(ba.readF2() + 1);
                logger.log(BasicLevel.INFO, "Thread2: write done");
                logger.log(BasicLevel.INFO, "Thread2: commit");
                pm2.currentTransaction().commit();
                pm2.close();
            }
        });

        // end transaction 1
        logger.log(BasicLevel.INFO, "Thread1: commit");
        pm1.currentTransaction().commit();
        pm1.close();

        assertUnlock(t);

        // clean up
        pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        ba = (BasicA) pm.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        Assert.assertEquals("Bad f1 value", "1", ba.readF1());
        Assert.assertEquals("Bad f2 value", 3, ba.readF2());
        pm.deletePersistent(ba);
        pm.currentTransaction().commit();
        pm.close();
    }

    public void testWriteConcurrentRead() {
        logger.log(BasicLevel.INFO, "test write concurrent read");
        // setup
        BasicA ba = new BasicA();
        ba.writeF1("1");
        ba.writeF2(2);
        PersistenceManager pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        pm.makePersistent(ba);
        final Object id = pm.getObjectId(ba);
        Assert.assertNotNull("Null object identifier", id);
        pm.currentTransaction().commit();
        pm.close();

        ba = null;

        // begin transaction 1
        PersistenceManager pm1 = pmf.getPersistenceManager();
        logger.log(BasicLevel.INFO, "Thread1: begin");
        pm1.currentTransaction().begin();
        logger.log(BasicLevel.INFO, "Thread1: getObjectById");
        ba = (BasicA) pm1.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        logger.log(BasicLevel.INFO, "Thread1: write");
        ba.writeF2(ba.readF2() + 1);
        logger.log(BasicLevel.INFO, "Thread1: write done");

        // begin transaction 2
        Thread t = assertLock(new Runnable () {
            public void run () {
                PersistenceManager pm2 = pmf.getPersistenceManager();
                logger.log(BasicLevel.INFO, "Thread2: begin");
                pm2.currentTransaction().begin();
                logger.log(BasicLevel.INFO, "Thread2: getObjectById");
                BasicA ba = (BasicA) pm2.getObjectById(id, true);
                Assert.assertNotNull("Object not found", ba);
                logger.log(BasicLevel.INFO, "Thread2: read");
                ba.readF2();
                logger.log(BasicLevel.INFO, "Thread2: read done");
                logger.log(BasicLevel.INFO, "Thread2: commit");
                pm2.currentTransaction().commit();
                pm2.close();
            }
        });

        // end transaction 1
        logger.log(BasicLevel.INFO, "Thread1: commit");
        pm1.currentTransaction().commit();
        pm1.close();

        assertUnlock(t);

        // clean up
        pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        ba = (BasicA) pm.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        Assert.assertEquals("Bad f1 value", "1", ba.readF1());
        Assert.assertEquals("Bad f2 value", 3, ba.readF2());
        pm.deletePersistent(ba);
        pm.currentTransaction().commit();
        pm.close();
    }

    public void testWriteConcurrentRead2() {
        logger.log(BasicLevel.INFO, "test write concurrent read 2");
        // setup
        BasicA ba = new BasicA();
        ba.writeF1("1");
        ba.writeF2(2);
        PersistenceManager pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        pm.makePersistent(ba);
        final Object id = pm.getObjectId(ba);
        Assert.assertNotNull("Null object identifier", id);
        pm.currentTransaction().commit();
        pm.close();

        ba = null;

        // begin transaction 1
        PersistenceManager pm1 = pmf.getPersistenceManager();
        logger.log(BasicLevel.INFO, "Thread1: begin");
        pm1.currentTransaction().begin();
        logger.log(BasicLevel.INFO, "Thread1: getObjectById");
        BasicA ba1 = (BasicA) pm1.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba1);

        // begin transaction 2
        Thread t = new Thread((new Runnable () {
            public void run () {
                PersistenceManager pm2 = pmf.getPersistenceManager();
                logger.log(BasicLevel.INFO, "Thread2: begin");
                pm2.currentTransaction().begin();
                logger.log(BasicLevel.INFO, "Thread2: getObjectById");
                BasicA ba2 = (BasicA) pm2.getObjectById(id, true);
                Assert.assertNotNull("Object not found", ba2);
                logger.log(BasicLevel.INFO, "Thread2: read");
                ba2.readF2();
                logger.log(BasicLevel.INFO, "Thread2: read done");
                logger.log(BasicLevel.INFO, "Thread2: commit");
                pm2.currentTransaction().commit();
                pm2.close();
            }
        }));
        t.start();

        // end transaction 1
        logger.log(BasicLevel.INFO, "Thread1: write");
        ba1.writeF2(ba1.readF2() + 1);
        logger.log(BasicLevel.INFO, "Thread1: write done");
        logger.log(BasicLevel.INFO, "Thread1: commit");
        pm1.currentTransaction().commit();
        pm1.close();

        assertUnlock(t);

        // clean up
        pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        ba = (BasicA) pm.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        Assert.assertEquals("Bad f1 value", "1", ba.readF1());
        Assert.assertEquals("Bad f2 value", 3, ba.readF2());
        pm.deletePersistent(ba);
        pm.currentTransaction().commit();
        pm.close();
    }

	public void testConcurrentWrites() {
        logger.log(BasicLevel.INFO, "test concurrent writes");
        // setup
		BasicA ba = new BasicA();
		ba.writeF1("1");
		ba.writeF2(2);
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		pm.makePersistent(ba);
		final Object id = pm.getObjectId(ba);
		Assert.assertNotNull("Null object identifier", id);
		pm.currentTransaction().commit();
		pm.close();

		ba = null;

        // begin transaction 1
        PersistenceManager pm1 = pmf.getPersistenceManager();
        logger.log(BasicLevel.INFO, "Thread1: begin");
        pm1.currentTransaction().begin();
        logger.log(BasicLevel.INFO, "Thread1: getObjectById");
        ba = (BasicA) pm1.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        logger.log(BasicLevel.INFO, "Thread1: write");
        ba.writeF2(ba.readF2() + 1);
        logger.log(BasicLevel.INFO, "Thread1: write done");

        // begin transaction 2
        Thread t = assertLock(new Runnable () {
            public void run () {
                PersistenceManager pm2 = pmf.getPersistenceManager();
                try {
                    logger.log(BasicLevel.INFO, "Thread2: begin");
                    pm2.currentTransaction().begin();
                    logger.log(BasicLevel.INFO, "Thread2: getObjectById");
                    BasicA ba = (BasicA) pm2.getObjectById(id, true);
                    Assert.assertNotNull("Object not found", ba);
                    logger.log(BasicLevel.INFO, "Thread2: write");
                    ba.writeF2(ba.readF2() + 1);
                    logger.log(BasicLevel.INFO, "Thread2: write done");

                    logger.log(BasicLevel.INFO, "Thread2: commit");
                    pm2.currentTransaction().commit();
                } catch (JDOFatalException e) {

                } finally {
                    pm2.close();
                }
            }
        });

        // end transaction 1
        logger.log(BasicLevel.INFO, "Thread1: commit");
        pm1.currentTransaction().commit();
        pm1.close();

        // end transaction 2
        assertUnlock(t);

        // clean up
		pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		ba = (BasicA) pm.getObjectById(id, true);
		Assert.assertNotNull("Object not found", ba);
		Assert.assertEquals("Bad f1 value", "1", ba.readF1());
		Assert.assertEquals("Bad f2 value", 4, ba.readF2());
		pm.deletePersistent(ba);
		pm.currentTransaction().commit();
		pm.close();
	}

    public void testConcurrentWrites2() {
        logger.log(BasicLevel.INFO, "test concurrent writes 2");
        // setup
        BasicA ba = new BasicA();
        ba.writeF1("1");
        ba.writeF2(2);
        PersistenceManager pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        pm.makePersistent(ba);
        final Object id = pm.getObjectId(ba);
        Assert.assertNotNull("Null object identifier", id);
        pm.currentTransaction().commit();
        pm.close();

        ba = null;

        // begin transaction 1
        PersistenceManager pm1 = pmf.getPersistenceManager();
        logger.log(BasicLevel.INFO, "Thread1: begin");
        pm1.currentTransaction().begin();
        logger.log(BasicLevel.INFO, "Thread1: getObjectById");
        ba = (BasicA) pm1.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);

        // begin transaction 2
        Thread t = assertLock(new Runnable () {
            public void run () {
                PersistenceManager pm2 = pmf.getPersistenceManager();
                try {
                    logger.log(BasicLevel.INFO, "Thread2: begin");
                    pm2.currentTransaction().begin();
                    logger.log(BasicLevel.INFO, "Thread2: getObjectById");
                    BasicA ba = (BasicA) pm2.getObjectById(id, true);
                    Assert.assertNotNull("Object not found", ba);
                    logger.log(BasicLevel.INFO, "Thread2: write");
                    ba.writeF2(ba.readF2() + 1);
                    logger.log(BasicLevel.INFO, "Thread2: write done");

                    logger.log(BasicLevel.INFO, "Thread2: commit");
                    pm2.currentTransaction().commit();
                } catch (JDOFatalException e) {
                } finally {
                    pm2.close();
                }
            }
        });

        // end transaction 1
        logger.log(BasicLevel.INFO, "Thread1: write");
        boolean ok = true;
        try {
            ba.writeF2(ba.readF2() + 1);
            ok = true;
            logger.log(BasicLevel.INFO, "Thread1: write done");
            logger.log(BasicLevel.INFO, "Thread1: commit");
            pm1.currentTransaction().commit();
        } catch (JDOException e) {
            logger.log(BasicLevel.INFO, "Thread1: write aborted"
                    + ExceptionHelper.getNested(e).getMessage());
            ok = false;
        } finally {
            pm1.close();
        }
        assertUnlock(t);

        // clean up
        pm = pmf.getPersistenceManager();
        pm.currentTransaction().begin();
        ba = (BasicA) pm.getObjectById(id, true);
        Assert.assertNotNull("Object not found", ba);
        Assert.assertEquals("Bad f1 value", "1", ba.readF1());
        Assert.assertEquals("Bad f2 value", ( ok ? 4 : 3), ba.readF2());
        pm.deletePersistent(ba);
        pm.currentTransaction().commit();
        pm.close();
    }

    private Thread assertLock (Runnable r) {
        Thread t = new Thread(r);
        t.start();
        try {
            t.join(5000);
        } catch (InterruptedException _) {
            // ignored
        }
        assertTrue(t.isAlive());
        logger.log(BasicLevel.INFO, "Thread2: locked");
        return t;
    }

    private void assertUnlock (Thread t) {
        try {
            t.join(5000);
        } catch (InterruptedException _) {
            // ignored
        }
        assertTrue(!t.isAlive());
    }
}
