/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Col.
 *
 */

package org.objectweb.speedo.runtime.relations;

import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;

import junit.framework.Assert;

import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.pobjects.relations.A;
import org.objectweb.speedo.pobjects.relations.B;

public class TestCascadeDeleteRelations extends SpeedoTestHelper {

	public TestCascadeDeleteRelations(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return LOG_NAME + "rt.relations.CascadeDeleteRelations";
	}

	/**
	 * This method tests the cascade delete in a one-one relationship.
	 *
	 */
	public void testOneOne() {
        A a1 = new A("testOneOne_a1");
        A a2 = new A("testOneOne_a2");
        B b1 = new B("testOneOne_b1");
        B b2 = new B("testOneOne_b2");

        // a1 references b1 (so b1 references a1)
        a1.setB(b1);
        assertTrue(b1.getA() == a1);

        // a2 references b2 (so b2 references a2)
        a2.setB(b2);
        assertTrue(b2.getA() == a2);

        // We make  the As persistent
        PersistenceManager pm = pmf.getPersistenceManager();
        pm.makePersistent(a1);
        pm.makePersistent(a2);
        
        // We get the ids
        Object a1Id = pm.getObjectId(a1);
        Object a2Id = pm.getObjectId(a2);
        Object b1Id = pm.getObjectId(b1);
        Object b2Id = pm.getObjectId(b2);
        Assert.assertNotNull("null object identifier for a1", a1Id);
        Assert.assertNotNull("null object identifier for a2", a2Id);
        Assert.assertNotNull("null object identifier for b1", b1Id);
        Assert.assertNotNull("null object identifier for b2", b2Id);
        
        pm.close();
        
        // Now, we are going to get a1 and delete it. Then we'll check that b1 has been deleted too
        pm = pmf.getPersistenceManager();
        a1 = (A)pm.getObjectById(a1Id, true);
        b2 = (B)pm.getObjectById(b2Id, true);
        pm.currentTransaction().begin();
        pm.deletePersistent(a1);
        pm.deletePersistent(b2);
        pm.currentTransaction().commit();
       
        try {
        	b1 = (B)pm.getObjectById(b1Id, true);
        	Assert.fail("Should have thrown a JDOUser Exception : b1 has not been deleted");
        }
        catch (JDOUserException jdoEx) {
        	// The test has passed
        	Assert.assertTrue(true);
        }

        try {
        	a2 = (A)pm.getObjectById(a2Id, true);
        	Assert.fail("Should have thrown a JDOUser Exception : a2 has not been deleted");
        }
        catch (JDOUserException jdoEx) {
        	// The test has passed
        	Assert.assertTrue(true);
        }

        pm.close();
	}

	
    public void testManyManyDelete() {
        A a1 = new A("testManyManyDelete_a1");
        A a2 = new A("testManyManyDelete_a2");
        A a3 = new A("testManyManyDelete_a3");

        B b1 = new B("testManyManyDelete_b1");
        B b2 = new B("testManyManyDelete_b2");
        B b3 = new B("testManyManyDelete_b3");

        PersistenceManager pm = pmf.getPersistenceManager();
        pm.makePersistent(a1);
        pm.makePersistent(a2);
        pm.makePersistent(a3);
        pm.makePersistent(b1);
        pm.makePersistent(b2);
        pm.makePersistent(b3);
        pm.close();

        try {
            pm = pmf.getPersistenceManager();

            // We add b1 and b2 to a1's Bs
            a1.getBs().add(b1);
            assertTrue(b1.getAs().contains(a1));
            a1.getBs().add(b2);
            assertTrue(b2.getAs().contains(a1));

            // We add b1, b2 and b3 to a2's Bs
            a2.getBs().add(b1);
            assertTrue(b1.getAs().contains(a2));
            a2.getBs().add(b2);
            assertTrue(b2.getAs().contains(a2));
            a2.getBs().add(b3);
            assertTrue(b3.getAs().contains(a2));

            // We add b2 and b3 to a3's Bs
            a3.getBs().add(b2);
            assertTrue(b2.getAs().contains(a3));
            a3.getBs().add(b3);
            assertTrue(b3.getAs().contains(a3));

            // We get some ids to try to get deleted objects
            Object a3Id = pm.getObjectId(a3);
            Object b2Id = pm.getObjectId(b2);
            Object b3Id = pm.getObjectId(b3);
            
            pm.close();

            // We delete a3 : as bs relationship in A has a cascade delete extension with the true value,
            // b2 and b3 should be delete.
            pm = pmf.getPersistenceManager();
            pm.currentTransaction().begin();
            pm.deletePersistent(a3);
            pm.currentTransaction().commit();
            pm.close();

            pm = pmf.getPersistenceManager();

            // a1 should only have b1
            assertTrue("a1.getBs() has " + a1.getBs().size() + " elements, expected 1", a1.getBs().size() == 1);
            assertTrue("a1 should contain b1", a1.getBs().contains(b1));
            assertTrue("a1 should not contain b2", !a1.getBs().contains(b2));

            // a2 should only have b1
            assertTrue("a2.getBs() has " + a2.getBs().size() + " elements, expected 1", a2.getBs().size() == 1);
            assertTrue("a2 should contain b1", a2.getBs().contains(b1));
            assertTrue("a2 should not contain b3", !a2.getBs().contains(b3));

            // a3 should not exist anymore
            try {
	   	        a3 = (A) pm.getObjectById(a3Id, true);
	   	        Assert.fail("Should have thrown a JDOUserException : a3 has not been deleted");
            }
            catch (JDOUserException jdoEx) {
            	// The test has passed
            }

            // b2 should not exist anymore
            try {
	   	        b2 = (B) pm.getObjectById(b2Id, true);
	   	        Assert.fail("Should have thrown a JDOUserException : b2 has not been deleted");
            }
            catch (JDOUserException jdoEx) {
            	// The test has passed
            }

            // b3 should not exist anymore
            try {
	   	        b3 = (B) pm.getObjectById(b3Id, true);
	   	        Assert.fail("Should have thrown a JDOUserException : b3 has not been deleted");
            }
            catch (JDOUserException jdoEx) {
            	// The test has passed
            }

            // b1 should still contain a1 and a2, but no more a3
            assertTrue("b1 should contain a1", b1.getAs().contains(a1));
            assertTrue("b1 should contain a2", b1.getAs().contains(a2));
            assertTrue("b1 should not contain a3", !b1.getAs().contains(a3));

            pm.close();
            
        } finally {
            pm = pmf.getPersistenceManager();
            pm.currentTransaction().begin();
            pm.deletePersistent(a1);
            pm.deletePersistent(a2);
            pm.deletePersistent(b1);
            pm.currentTransaction().commit();
            pm.close();
        }
    }
}
