/*
 * Speedo: an implementation of JDO compliant personality on top of JORM
 * generic I/O sub-system. Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Release: 1.0
 *
 * Created on 1 mars 2004 @author fmillevi@yahoo.com
 *
 */
package org.objectweb.speedo.j2eedo.bo;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import org.objectweb.speedo.Alea;
import org.objectweb.speedo.j2eedo.common.PMHolder;
import org.objectweb.speedo.j2eedo.database.Address;
import org.objectweb.speedo.j2eedo.database.DatabaseObjectInterface;
import org.objectweb.speedo.j2eedo.database.Department;
import org.objectweb.speedo.j2eedo.database.Employee;
import org.objectweb.speedo.j2eedo.database.Project;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

/**
 * This class handle each request on the <b>J2EEDO</b> application.
 * The doAction() method performs the current requested action.
 * <p>
 * It starts and commits a transaction if needed
 * </p>
 * <p>
 * Call the asked action. When the action in the
 * {@link DatabaseImpl#actionArray action list}, the 3 local static lists of
 * id (department, employee and project) are redefined.
 * </p>
 *
 * @author fmillevi@yahoo.com
 */
public class DatabaseImpl {
	/**
	 * Defines the action <code>PARAMETER_PING</code><b> its value is ("ping")</b>
	 * <p>
	 * This action is use just to check if the application can be reached
	 * </p>
	 */
	final public static String PARAMETER_PING = "ping";
	/**
	 * Defines the action <code>PARAMETER_NEW_DEPARTMENT</code><b> its value is
	 * ("newDept")</b>
	 * <p>
	 * This action Creates a new department and some new employees (random
	 * number defined between 10 and 70).
	 * </p>
	 *
	 * @see DepartmentFactory#MIN_EMPLOYEE_PER_DEPARTMENT
	 * @see DepartmentFactory#MAX_EMPLOYEE_PER_DEPARTMENT
	 * @see DepartmentFactory#newDepartmentWithEmployees
	 */
	final public static String PARAMETER_NEW_DEPARTMENT = "newDept";
	/**
	 * Defines the action <code>PARAMETER_NEW_PROJECT</code><b> its value is
	 * ("newProj")</b>
	 * <p>
	 * This action Creates a new project and affect few employees (random
	 * number defined between 5 and 20).
	 * </p>
	 *
	 * @see ProjectFactory#MIN_MEMBER_PER_PROJECT
	 * @see ProjectFactory#MAX_MEMBER_PER_PROJECT
	 * @see ProjectFactory#newProjectWithEmployees
	 */
	final public static String PARAMETER_NEW_PROJECT = "newProj";
	/**
	 * Defines the action <code>PARAMETER_NEW_EMPLOYEE</code><b> its value is
	 * ("newEmp")</b>
	 * <p>
	 * Selects one of the known departments and creates a new employee.
	 * </p>
	 * @see EmployeeFactory#newEmployee
	 */
	final public static String PARAMETER_NEW_EMPLOYEE = "newEmp";
	/**
	 * Defines the action <code>PARAMETER_SET_BOSS</code><b> its value is
	 * ("setDeptBoss")</b>
	 * @see DepartmentFactory#setManagerForADepartment()
	 */
	final public static String PARAMETER_SET_BOSS = "setDeptBoss";
	/**
	 * Defines the action <code>PARAMETER_REM_EMPLOYEE</code><b> its value is
	 * ("delEmployee")</b>
	 * @see EmployeeFactory#deleteEmployee
	 */
	final public static String PARAMETER_REM_EMPLOYEE = "delEmployee";
	/**
	 * Defines the action <code>PARAMETER_REM_PROJECT</code><b> its value is
	 * ("delProject")</b>
	 * @see ProjectFactory#deleteProject
	 */
	final public static String PARAMETER_REM_PROJECT = "delProject";
	/**
	 * Defines the action <code>PARAMETER_SPLIT_PROJECT</code><b> its value is
	 * ("splitProject")</b>
	 * @see ProjectFactory#splitProject
	 */
	final public static String PARAMETER_SPLIT_PROJECT = "splitProject";
	/**
	 * Defines the action <code>PARAMETER_SPLIT_DEPARTMENT</code><b> its value is
	 * ("splitDepartment")</b>
	 * @see DepartmentFactory#splitDepartment
	 */
	final public static String PARAMETER_SPLIT_DEPARTMENT = "splitDepartment";
	/**
	 * Defines the action <code>PARAMETER_MERGE_DEPARTMENT</code><b> its value is
	 * ("mergeDept")</b>
	 * @see DepartmentFactory#mergeDepartment
	 */
	final public static String PARAMETER_MERGE_DEPARTMENT = "mergeDept";
	/**
	 * Defines the action <code>PARAMETER_INCREASE_SALARY</code><b> its value is
	 * ("incSalary")</b> for good folsk only...
	 * @see EmployeeFactory#increaseSalary()
	 */
	final public static String PARAMETER_INCREASE_SALARY = "incSalary";
	/**
	 * Defines the action <code>PARAMETER_GET_DEPARTMENT</code><b> its value is
	 * ("getDepartment")</b>
	 * @see DepartmentFactory#getDepartment()
	 */
	final public static String PARAMETER_GET_DEPARTMENT = "getDepartment";
	/**
	 * Defines the action <code>PARAMETER_GET_PROJECT</code><b> its value is
	 * ("getProject")</b>
	 * @see ProjectFactory#getProject()
	 */
	final public static String PARAMETER_GET_PROJECT = "getProject";
	/**
	 * Defines the action <code>PARAMETER_GET_EMPLOYEE</code><b> its value is
	 * ("getEmployee")</b>
	 * @see EmployeeFactory#getEmployees()
	 */
	final public static String PARAMETER_GET_EMPLOYEE = "getEmployee";
	/**
	 * Defines the action <code>PARAMETER_QUERY_PROJECTS</code><b> its value is
	 * ("queryProjects")</b>
	 * <p>
	 * Performs one of the four queries:
	 * <ul>
	 * <li>Get employee by its id,</li>
	 * <li>Get employees between min and max,</li>
	 * <li>Get employees having the same manager,</li>
	 * <li>Get employees member of a project.</li>
	 * </ul>
	 * </p>
	 * @see EmployeeFactory#getEmployees()
	 */
	final public static String PARAMETER_QUERY_PROJECTS = "queryProjects";
	/**
	 * Defines the action <code>PARAMETER_QUERY_EMPLOYEES</code><b> its value is
	 * ("queryEmployees")</b>
	 * <p>
	 * Performs one of the two queries:
	 * <ul>
	 * <li>Get project by its id,</li>
	 * <li>Get projects by member.</li>
	 * </ul>
	 * </p>
	 * @see ProjectFactory#getProjects()
	 */
	final public static String PARAMETER_QUERY_EMPLOYEES = "queryEmployees";

	final public static String PARAMETER_EVICTALL = "evictall";
	/**
	 * The String array <code>actionArray</code> gives the liste of actions
	 * avalaibles
	 */
	final public static String actionArray[] =
		{
			DatabaseImpl.PARAMETER_NEW_DEPARTMENT,
			DatabaseImpl.PARAMETER_NEW_PROJECT,
			DatabaseImpl.PARAMETER_NEW_EMPLOYEE,
			DatabaseImpl.PARAMETER_SET_BOSS,
			DatabaseImpl.PARAMETER_REM_EMPLOYEE,
			DatabaseImpl.PARAMETER_REM_PROJECT,
			DatabaseImpl.PARAMETER_SPLIT_PROJECT,
			DatabaseImpl.PARAMETER_SPLIT_DEPARTMENT,
			DatabaseImpl.PARAMETER_MERGE_DEPARTMENT,
			DatabaseImpl.PARAMETER_INCREASE_SALARY,
			DatabaseImpl.PARAMETER_GET_DEPARTMENT,
			DatabaseImpl.PARAMETER_GET_PROJECT,
			DatabaseImpl.PARAMETER_GET_EMPLOYEE,
			DatabaseImpl.PARAMETER_QUERY_PROJECTS,
			DatabaseImpl.PARAMETER_QUERY_EMPLOYEES,
			DatabaseImpl.PARAMETER_PING };

	final public static int READ = 1;
	final public static int WRITE = 0;

	/**
	 * The double dimension array <code>actionWeightArray</code> gives for
	 * each action the relative read and write weight
	 *
	 * @see #actionArray
	 */
	final public static int actionWeightArray[][] = { //{write,read}
		{ 4, 0 }, //DatabaseImpl.PARAMETER_NEW_DEPARTMENT,
		{
			5, 3 }, //DatabaseImpl.PARAMETER_NEW_PROJECT,
		{
			1, 1 }, //DatabaseImpl.PARAMETER_NEW_EMPLOYEE,
		{
			4, 2 }, //DatabaseImpl.PARAMETER_SET_BOSS,
		{
			1, 1 }, //DatabaseImpl.PARAMETER_REM_EMPLOYEE,
		{
			4, 2 }, //DatabaseImpl.PARAMETER_REM_PROJECT,
		{
			4, 3 }, //DatabaseImpl.PARAMETER_SPLIT_PROJECT,
		{
			4, 4 }, //DatabaseImpl.PARAMETER_SPLIT_DEPARTMENT
		{
			5, 2 }, //DatabaseImpl.PARAMETER_MERGE_DEPARTMENT,
		{
			3, 2 }, //DatabaseImpl.PARAMETER_INCREASE_SALARY,
		{
			0, 5 }, //DatabaseImpl.PARAMETER_GET_DEPARTMENT,
		{
			0, 3 }, //DatabaseImpl.PARAMETER_GET_PROJECT,
		{
			0, 1 }, //DatabaseImpl.PARAMETER_GET_EMPLOYEE,
		{
			0, 3 }, //DatabaseImpl.PARAMETER_QUERY_PROJECT,
		{
			0, 2 }, //DatabaseImpl.PARAMETER_QUERY_EMPLOYEE
		{
			0, 0 } //DatabaseImpl.PARAMETER_PING
	};

	private String action;
	private boolean performCommit;
	private Hashtable needTransactionArray;
	protected StringBuffer outStr;
	private PMHolder persistenceManagerHolder;

	/**
	 * Current Persistence Manager <code>pm</code>
	 */
	public PersistenceManager pm;
	private PollsSynchronizations pollsSync;

	private DepartmentFactory departmentFactory = null;
	private EmployeeFactory employeeFactory = null;
	private ProjectFactory projectFactory = null;

	/**
	 * The Vector <code>poolOfDepartmentId</code> is a static list of known
	 * departmentId used to keep in mind the list of department without doing
	 * any JDO request.
	 * <p>
	 * This is a local cache
	 * </p>
	 */
	public final static Vector poolOfDepartmentId = new Vector();
	/**
	 * The Vector <code>poolOfProjectId</code> is a static list of known
	 * projectId used to keep in mind the list of project without doing any JDO
	 * request.
	 * <p>
	 * This is a local cache
	 * </p>
	 */
	public final static Vector poolOfProjectId = new Vector();
	/**
	 * The Vector <code>poolOfEmployeeId</code> is a static list of known
	 * employeeId used to keep in mind the list of employee without doing any
	 * JDO request.
	 * <p>
	 * This is a local cache
	 * </p>
	 */
	public final static Vector poolOfEmployeeId = new Vector();

	private static boolean resetPools = true;

	static Logger logger = null;
	static {
		LoggerFactory lf = Monolog.initialize();
		logger = lf.getLogger(DatabaseImpl.class.getName());
	}

	private final static int PROJECT_INIT_SIZE = 100;
	private final static int DEPARTMENT_INIT_SIZE = 30;

	/**
	 * Empty Constructor initialize the current data model implementation
	 * object :
	 *
	 * @param pmHolder
	 *            is the Persistence Manager Holder used to get the PM.
	 */
	public DatabaseImpl(PMHolder pmHolder) {
		super();
		this.persistenceManagerHolder = pmHolder;
		this.action = "";
		this.performCommit = false;
		this.outStr = new StringBuffer();
		this.departmentFactory = new DepartmentFactory();
		this.employeeFactory = new EmployeeFactory();
		this.projectFactory = new ProjectFactory();
		this.needTransactionArray = new Hashtable();
		// define the action needing transaction.
		int writeWeight;
		for (int i = 0; i < DatabaseImpl.actionWeightArray.length; i++) {
			String _action = DatabaseImpl.actionArray[i];
			writeWeight = DatabaseImpl.actionWeightArray[i][DatabaseImpl.WRITE];
			if (0 == writeWeight)
				needTransactionArray.put(_action, new Boolean(false));
			else
				needTransactionArray.put(_action, new Boolean(true));
		}
	}

	/**
	 * This method initialize the action to be performed, disables transaction
	 * management and calls the private doAction() method.
	 *
	 * @param parameter
	 *            is the action to be performed
	 * @return threatment result as String
	 * @throws JDOException
	 * @throws Exception
	 */
	public String doAction(String parameter) throws JDOException, Exception {
		this.action = parameter;
		this.performCommit = false;
		return doAction();
	}

	/**
	 * This method initialize the action to be performed, enables or disables
	 * transaction management according the parameter withTransaction and calls
	 * the private doAction() method.
	 *
	 * @param parameter
	 *            is the action to be performed
	 * @param withTransaction
	 *            is a boolean use to enable or disable the use of transaction
	 * @return threatment result as String
	 * @throws JDOException
	 * @throws Exception
	 */
	public String doAction(String parameter, boolean withTransaction)
		throws JDOException, Exception {
		this.action = parameter;
		this.performCommit = withTransaction;
		return doAction();
	}

	/**
	 * This method performs the current requested action.
	 * <p>
	 * It starts and commits a transaction if needed
	 * </p>
	 * <p>
	 * Call the asked action. When the action in the
	 * {@see #actionArray actuion list}defined above, the 3 local static list
	 * of id (department, employee and project) are redefined.
	 * </p>
	 *
	 * @return threatment result as String
	 * @throws JDOException
	 * @throws Exception
	 */
	private String doAction() throws JDOException, Exception {
		this.outStr = new StringBuffer();
		// check if a transaction is required
		boolean needTrans =
			(this.performCommit)
				? ((Boolean) this.needTransactionArray.get(this.action))
					.booleanValue()
				: false;
		try {
			if (PARAMETER_PING.equalsIgnoreCase(action)) {
				return "Alive...";
			}
			this.pm = this.persistenceManagerHolder.getPersistenceManager();
			if (PARAMETER_EVICTALL.equalsIgnoreCase(action)) {
			    logger.log(BasicLevel.WARN, "Flushing cache ...");
			    this.pm.evictAll();
				resetPolls(this.pm, this.outStr);
			    return "Cache flushed !";
			}
			DatabaseImpl.initPolls(this.pm);

			// start a transaction if required
			if (!"".equalsIgnoreCase(this.action)
				&& ((Boolean) this.needTransactionArray.get(this.action))
					.booleanValue()) {
				this.pollsSync = new PollsSynchronizations();
				this.pm.setUserObject(pollsSync);
				this.pm.currentTransaction().setSynchronization(pollsSync);
			}

			if (needTrans) {
				logger.log(
					BasicLevel.DEBUG,
					"Start a new transaction "
						+ this.pm.currentTransaction().toString());
				this.pm.currentTransaction().begin();
			}
			logger.log(BasicLevel.DEBUG, "do action " + this.action);
			if (PARAMETER_NEW_DEPARTMENT.equalsIgnoreCase(action)) {
				this.outStr.append("\nCreate new deparment with its employees");
				this.departmentFactory.setOutStr(this.outStr);
				this.departmentFactory.setPm(this.pm);
				this.departmentFactory.newDepartmentWithEmployees(
					this.pollsSync);
			} else if (PARAMETER_NEW_PROJECT.equalsIgnoreCase(action)) {
				this.outStr.append("\nCreate new project with its members");
				this.projectFactory.setOutStr(this.outStr);
				this.projectFactory.setPm(this.pm);
				this.projectFactory.newProjectWithEmployees(this.pollsSync);
			} else if (PARAMETER_SET_BOSS.equalsIgnoreCase(action)) {
				this.outStr.append(
					"\nUpdate the boss for a department's employees");
				this.departmentFactory.setOutStr(this.outStr);
				this.departmentFactory.setPm(this.pm);
				this.departmentFactory.setManagerForADepartment();
			} else if (PARAMETER_NEW_EMPLOYEE.equalsIgnoreCase(action)) {
				this.outStr.append("\nCreate an employee");
				this.employeeFactory.setOutStr(this.outStr);
				this.employeeFactory.setPm(this.pm);
				this.employeeFactory.newEmployee(this.pollsSync);
			} else if (PARAMETER_REM_EMPLOYEE.equalsIgnoreCase(action)) {
				this.outStr.append("\nDelete an employee");
				this.employeeFactory.setOutStr(this.outStr);
				this.employeeFactory.setPm(this.pm);
				this.employeeFactory.deleteEmployee(this.pollsSync);
			} else if (PARAMETER_REM_PROJECT.equalsIgnoreCase(action)) {
				this.outStr.append("\nDelete a project");
				this.projectFactory.setOutStr(this.outStr);
				this.projectFactory.setPm(this.pm);
				this.projectFactory.deleteProject(this.pollsSync);
			} else if (PARAMETER_SPLIT_PROJECT.equalsIgnoreCase(action)) {
				this.outStr.append("\nSplit a project");
				this.projectFactory.setOutStr(this.outStr);
				this.projectFactory.setPm(this.pm);
				this.projectFactory.splitProject(this.pollsSync);
			} else if (PARAMETER_MERGE_DEPARTMENT.equalsIgnoreCase(action)) {
				this.outStr.append("\nMerge two departments");
				this.departmentFactory.setOutStr(this.outStr);
				this.departmentFactory.setPm(this.pm);
				this.departmentFactory.mergeDepartment(this.pollsSync);
			} else if (PARAMETER_SPLIT_DEPARTMENT.equalsIgnoreCase(action)) {
				this.outStr.append("\nSplit a department");
				this.departmentFactory.setOutStr(this.outStr);
				this.departmentFactory.setPm(this.pm);
				this.departmentFactory.splitDepartment(this.pollsSync);
			} else if (PARAMETER_INCREASE_SALARY.equalsIgnoreCase(action)) {
				this.outStr.append("\nIncrease salary");
				this.employeeFactory.setOutStr(this.outStr);
				this.employeeFactory.setPm(this.pm);
				this.employeeFactory.increaseSalary();
			} else if (PARAMETER_GET_DEPARTMENT.equalsIgnoreCase(action)) {
				this.outStr.append("\nGet a department");
				this.departmentFactory.setOutStr(this.outStr);
				this.departmentFactory.setPm(this.pm);
				this.departmentFactory.getDepartment();
			} else if (PARAMETER_GET_PROJECT.equalsIgnoreCase(action)) {
				this.outStr.append("\nGet a project");
				this.projectFactory.setOutStr(this.outStr);
				this.projectFactory.setPm(this.pm);
				this.projectFactory.getProject();
			} else if (PARAMETER_GET_EMPLOYEE.equalsIgnoreCase(action)) {
				this.outStr.append("\nGet an employee");
				this.employeeFactory.setOutStr(this.outStr);
				this.employeeFactory.setPm(this.pm);
				this.employeeFactory.getEmployee();
			} else if (PARAMETER_QUERY_PROJECTS.equalsIgnoreCase(action)) {
				this.outStr.append("\nGet some projects");
				this.projectFactory.setOutStr(this.outStr);
				this.projectFactory.setPm(this.pm);
				this.projectFactory.getProjects();
			} else if (PARAMETER_QUERY_EMPLOYEES.equalsIgnoreCase(action)) {
				this.outStr.append("\nGet some persons");
				this.employeeFactory.setOutStr(this.outStr);
				this.employeeFactory.setPm(this.pm);
				this.employeeFactory.getEmployees();
			} else {
				logger.log(BasicLevel.DEBUG, "Resets and shows pools");
				resetPolls(this.pm, new StringBuffer());
			}
			logger.log(BasicLevel.DEBUG, "End of action : " + action);
			// commit the current transaction if required
			if (needTrans) {
				logger.log(
					BasicLevel.DEBUG,
					"Valid the current transaction "
						+ this.pm.currentTransaction().toString());
				this.pm.currentTransaction().setSynchronization(this.pollsSync);
				this.pm.currentTransaction().commit();
			}
		} catch (JDOFatalException e) {
			if (logger.isLoggable(BasicLevel.DEBUG)) {
			    logger.log(BasicLevel.DEBUG, "The action : '" + action 
			        + "' fails : Deadlock detected. Action canceled: ", e);
			} else {
			    logger.log(BasicLevel.WARN, "The action : '" + action 
			        + "' fails : Deadlock detected. Action canceled." + e.getMessage());
			}
			if (needTrans)
				this.pm.currentTransaction().rollback();

			//			if (!DatabaseImpl.resetPools) {
			//				DatabaseImpl.resetPools = true;
			//				this.doAction("", false);
			//			}
		} catch (JDOException e) {
			logger.log(
				BasicLevel.INFO,
				"The action : '"
					+ action
					+ "' fails : it throws a JDO exception.",
				e);
			throw e;
		} catch (Exception e) {
			logger.log(
				BasicLevel.WARN,
				"The action : '" + action + "' fails : it throws an exception.",
				e);
			throw e;
		} finally {
			this.persistenceManagerHolder.closePersistenceManager();
		}
		return this.outStr.toString();
	}

	//private static void removeFromPoll(Collection poll, long id) {
	//	poll.remove(new Long(id));
	//}

	private synchronized static void resetPolls(PersistenceManager pm, StringBuffer outStr) 
	throws JDOException, Exception {
		DatabaseImpl.poolOfDepartmentId.clear();
		DatabaseImpl.poolOfEmployeeId.clear();
		DatabaseImpl.poolOfProjectId.clear();
		DatabaseImpl.resetPools = true;
		DatabaseImpl.initPolls(pm);
		if (outStr != null) {
			outStr.append("\nDo nothing and dump static poll contents");
			outStr.append("\nDepartments:");
			outStr.append(DatabaseImpl.poolOfDepartmentId.toString());
			outStr.append("\nEmployees:");
			outStr.append(DatabaseImpl.poolOfEmployeeId.toString());
			outStr.append("\nProjects:");
			outStr.append(DatabaseImpl.poolOfProjectId.toString());
		}
	}
	
	private static synchronized void initPolls(PersistenceManager pm)
		throws JDOException, Exception {
		if (!DatabaseImpl.resetPools)
			return;
		try {
			if (DatabaseImpl.poolOfDepartmentId.size() == 0) {
				DatabaseImpl.initIdPoll(
					DatabaseImpl.poolOfDepartmentId,
					Department.class,
					pm);
				logger.log(
					BasicLevel.DEBUG,
					"Initialize the static pool of Departments Id : "
						+ DatabaseImpl.poolOfDepartmentId.toString());
			}
			if (DatabaseImpl.poolOfProjectId.size() == 0) {
				DatabaseImpl.initIdPoll(
					DatabaseImpl.poolOfProjectId,
					Project.class,
					pm);
				logger.log(
					BasicLevel.DEBUG,
					"Initialize the static pool of Projects Id : "
						+ DatabaseImpl.poolOfProjectId.toString());
			}
			if (DatabaseImpl.poolOfEmployeeId.size() == 0) {
				DatabaseImpl.initIdPoll(
					DatabaseImpl.poolOfEmployeeId,
					Employee.class,
					pm);
				logger.log(
					BasicLevel.DEBUG,
					"Initialize the static pool of Employees Id : "
						+ DatabaseImpl.poolOfEmployeeId.toString());
			}
		} finally {
			DatabaseImpl.resetPools = false;
		}
	}

	private static void initIdPoll(
		Collection poll,
		Class c,
		PersistenceManager pm)
		throws JDOException, Exception {
		Query query = pm.newQuery(c);
		logger.log(
			BasicLevel.DEBUG,
			"(Re)Initialize static pool id for classe : " + c.getName());
		Collection col = (Collection) query.execute();
		Iterator it = col.iterator();
		while (it.hasNext()) {
			Object o = it.next();
			try {
				poll.add(new Long(((DatabaseObjectInterface) o).getId()));
			} catch (Exception e) {
				throw e;
			}
		}
		query.close(col);
	}

	final private static long getIdFromPool(Vector poll) {
		// Get an id from the pool.
		while (DatabaseImpl.resetPools) {
			try {
				logger.log(
					BasicLevel.DEBUG,
					"sleep until the end off static pool id reset...");
				Thread.sleep(10);
			} catch (InterruptedException e) {
			}
		}
		int alea = Alea.rand(1, poll.size());
		Iterator it = poll.iterator();
		long current = -1;
		while (it.hasNext() && alea > 0) {
			current = ((Long) it.next()).longValue();
			alea--;
			if (!it.hasNext())
				it = poll.iterator();
		}
		return current;
	}

	public static void initTestData(PMHolder pmHolder)
		throws JDOException, Exception {
		String str;
		PersistenceManager pm = pmHolder.getPersistenceManager();
		DatabaseImpl departmentImpl = null;
		departmentImpl = new DatabaseImpl(pmHolder);
		pm.evictAll();
		Iterator objIter = null;
		DatabaseImpl.poolOfDepartmentId.clear();
		DatabaseImpl.poolOfEmployeeId.clear();
		DatabaseImpl.poolOfProjectId.clear();
		// remove all employee
		Class[] classes = new Class[] {
		        Employee.class,
		        Address.class,
		        Project.class,
		        Department.class
		};
		pm.currentTransaction().begin();
		for (int i = 0; i < classes.length; i++) {
			logger.log(BasicLevel.DEBUG, "Removing " + classes[i].getName() + " ...");
			pm.deletePersistentAll((Collection) pm.newQuery(classes[i]).execute());
			logger.log(BasicLevel.INFO, "All " + classes[i].getName() + " have been removed.");
        }
		pm.currentTransaction().commit();

		pm.evictAll();

		pm.currentTransaction().begin();
		logger.log(BasicLevel.INFO, "Init departments and employees data.");
		for (int i = 0; i < DEPARTMENT_INIT_SIZE; i++) {
			str = departmentImpl.doAction(DatabaseImpl.PARAMETER_NEW_DEPARTMENT);
			logger.log(BasicLevel.DEBUG, str);
		}
		DatabaseImpl.resetPools = true;
		logger.log(BasicLevel.INFO, "Init projects data.");
		for (int i = 0; i < PROJECT_INIT_SIZE; i++) {
			str = departmentImpl.doAction(DatabaseImpl.PARAMETER_NEW_PROJECT);
			logger.log(BasicLevel.DEBUG, str);
		}
		DatabaseImpl.resetPools = true;
		pm.currentTransaction().commit();
		logger.log(BasicLevel.INFO, "Initial data set.");
		resetPolls(pm, null);
		pm.evictAll();
		logger.log(BasicLevel.DEBUG, "Remove all cache entries.");
		pmHolder.closePersistenceManager();
		logger.log(BasicLevel.DEBUG, "Close the persistenceManager.");
	}

	/**
	 * Returns one of the existing department id
	 *
	 * @return a department id
	 */
	public static long getDepartmentIdFromPool() {
		//synchronized (DatabaseImpl.poolOfDepartmentId) {
		return DatabaseImpl.getIdFromPool(DatabaseImpl.poolOfDepartmentId);
		//}
	}

	/**
	 * Returns one of the existing employee id
	 *
	 * @return a employee id
	 */
	public static long getEmployeeIdFromPool() {
		//synchronized (DatabaseImpl.poolOfEmployeeId) {
		return DatabaseImpl.getIdFromPool(DatabaseImpl.poolOfEmployeeId);
		//}
	}

	/**
	 * Returns one of the existing project id
	 *
	 * @return a project id
	 */
	public static long getProjectIdFromPool() {
		//synchronized (DatabaseImpl.poolOfProjectId) {
		return DatabaseImpl.getIdFromPool(DatabaseImpl.poolOfProjectId);
		//}
	}
}