/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.stress;

import org.objectweb.speedo.pobjects.relations.C;
import org.objectweb.speedo.pobjects.relations.D;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.Alea;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;

/**
 *
 * @author M. Guillemin
 */
public abstract class CDRelationHelper extends StressHelper {

	/**
	 * is the lists of object identifier prepared before the transaction
	 * execution.
	 * if (oids == null) {
	 *   db is not initialised
	 * } else if (oids != null && oids.length==0) {
	 *   db initialised and keepOid = false
	 * } else {
	 *   db initialised and keepOid == true
	 * }
	 */
	
	protected static Object[] oidsC = null;
	protected static Object[] oidsD = null;
	protected String DBSIZE = getLoggerName() + ".dbsize";
	protected String NO_DB_INIT = getLoggerName() + ".nodbinit";


	public CDRelationHelper(String s) {
		super(s);
	}

	protected String[] getClassNamesToInit() {
		return new String[]{
				org.objectweb.speedo.pobjects.relations.C.class.getName(),
				org.objectweb.speedo.pobjects.relations.D.class.getName()
		};		
	}

	protected boolean keepOid() {
		return true;
	}

	protected abstract int getNbSucc();
	
	/**
	 * IMPORTANT: dot not removed data on support in order to avoid next
	 * creations
	 */
	public void setUp() throws Exception {
		logger.log(BasicLevel.DEBUG, "setUp.");
		cleanup();
		initDataStructure(false);
		debug = logger.isLoggable(BasicLevel.DEBUG);
	}

	protected void prepareTest(TaskManager tm, Object ctx) {
		super.prepareTest(tm, ctx);
		Task task = tm.tasks[0];
		CDCtx cdctx = (CDCtx) ctx;
 
		if (task.txToExecute.length * getNbSucc() > cdctx.dbSize) {
			fail("Adjust parameter: tx <= "+cdctx.dbSize/getNbSucc() + " OR dbSize >= "+task.txToExecute.length * getNbSucc());
		}
	}

	/**
	 * Creates the persistent object if it is not already done.
	 * @param task the task to prepare
	 * @param _ctx the context of the test.
	 */
	protected void prepareTask(Task task, Object _ctx) {
		super.prepareTask(task, _ctx);
		CDCtx ctx = (CDCtx) _ctx;
		if (oidsC == null) {
			synchronized (getClass()) {
				if (oidsC == null && !Boolean.getBoolean(NO_DB_INIT)) {
					//Initialisation the database
					logger.log(BasicLevel.INFO, "\tPreparing test...");
					new PrepareTestCDRelation(this)
					        .prepare(ctx, keepOid());
					if (keepOid()) {
						//keep oids in the static variable
						oidsC = ctx.oidsC;
						oidsD = ctx.oidsD;
					} else {
						//db initialized without oids
						oidsC = new Object[0];
						oidsD = new Object[0];
					}
					logger.log(BasicLevel.INFO, "\tTest Prepared.");
				}
			}
		}
		ctx.oidsC = oidsC;
		ctx.oidsD = oidsD;
	}

	protected int getOid(CDCtx cdctx, boolean useTwiceTheSameOid, int nbSucc) {
		int oid = Alea.rand(0, cdctx.dbSize - nbSucc);
		if (!useTwiceTheSameOid) {
			synchronized (cdctx.usedOids) {
				while (oid % nbSucc != 0) {
					oid++;
				}
				while (cdctx.usedOids[oid]) {
					oid = Alea.rand(0, cdctx.dbSize - nbSucc);
					while (oid % nbSucc != 0) {
						oid++;
					}
				}
				for (int i=0; i<nbSucc; i++) {
					cdctx.usedOids[oid+i] = true;
				}
			}			
		}
		return oid;
	}
	
	/**
	 * The context to use for the CDRelation object
	 */
	public class CDCtx {
		/**
		 * The identifier of the created object (see keepOid method)
		 */
		public Object oidsC[];
		public Object oidsD[];
		public boolean usedOids[];
		/**
		 * The number of persistent object
		 */
		public int dbSize;
		/**
		 * indicates if the oid must be kept during the preparation of the task
		 */
		public boolean keepOid;
		
		
		public CDCtx(int dbSize) {
			this.dbSize = dbSize;
			oidsC = new Object[dbSize];
			oidsD = new Object[dbSize];
			usedOids = new boolean[dbSize];
		}

		public void initOnPrepare(int nbTx, boolean fetchOid) {
			this.keepOid = fetchOid;	
			for (int i=0; i<dbSize; i++)
				usedOids[i] = false;
		}

		public String toString() {
			return "dbSize = " + dbSize;
		}		
	}
}

class PrepareTestCDRelation extends StressHelper {

	private final static int NB_CREATION = 100;
	private final static int NB_THREAD = 4;

	public PrepareTestCDRelation(CDRelationHelper helper) {
		super(helper.getName());
	}

	protected String[] getClassNamesToInit() {
		return new String[]{
				org.objectweb.speedo.pobjects.relations.C.class.getName(),
				org.objectweb.speedo.pobjects.relations.D.class.getName()
		};				
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".CDRelationHelper";
	}

	protected String getLogPrefix() {
		return super.getLogPrefix() + "\t";
	}

	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       PerformResult res) {
		CDRelationHelper.CDCtx pctx = (CDRelationHelper.CDCtx) ctx;
		int plus = pctx.dbSize % NB_CREATION;
		int nbTx = task.txToExecute.length;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			if (plus > 0) {
				//The first transaction creates the additional object
				for (int oid = 0; txId == 0 && oid < plus; oid++) {
					C c = new C("c" + oid);
					D d = new D("d" + oid);
					pm.makePersistent(c);
					pm.makePersistent(d);
					if (pctx.keepOid) {
						pctx.oidsC[oid] = pm.getObjectId(c);
						pctx.oidsD[oid] = pm.getObjectId(d);
					}
				}
			} 
			// The other transactions create 'nbCreation' objects
			for (int no = 0; no < NB_CREATION; no++) {
				int oid = (txId * NB_CREATION) + no + plus;
				if (oid < pctx.dbSize) {
					C c = new C("c" + oid);
					D d = new D("d" + oid);
					pm.makePersistent(c);
					pm.makePersistent(d);
					if (pctx.keepOid) {
						pctx.oidsC[oid] = pm.getObjectId(c);
						pctx.oidsD[oid] = pm.getObjectId(d);
					}
				}
			}
			
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}

	}

	public void prepare(CDRelationHelper.CDCtx ctx, boolean fetchOid) {
		int plus = ctx.dbSize % NB_CREATION;
		int nbTx = (ctx.dbSize / NB_CREATION) + (plus > 0 ? 1 : 0);
		ctx.initOnPrepare(nbTx, fetchOid);
		perform(NB_THREAD, nbTx, Integer.getInteger(TIMEOUT, 200000).intValue(), ctx);
	}
}

