/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.stress;

import org.objectweb.speedo.pobjects.userid.IntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

/**
 *
 * @author S.Chassande-Barrioz
 */
public abstract class IntUserIdHelper extends StressHelper {

	/**
	 * is the lists of object identifier prepared before the transaction
	 * execution.
	 * if (oids == null) {
	 *   db is not initialised
	 * } else if (oids != null && oids.length==0) {
	 *   db initialised and keepOid = false
	 * } else {
	 *   db initialised and keepOid == true
	 * }
	 */
	protected static Object[] oids = null;
	protected String DBSIZE = getLoggerName() + ".dbsize";
	protected String NO_DB_INIT = getLoggerName() + ".nodbinit";

	public class IUICtx {
		public Object oids[];
		public int dbSize;
		public boolean keepOid;

		public IUICtx(int dbSize) {
			this.dbSize = dbSize;
			keepOid = false;
		}

		public String toString() {
			return "dbSize = " + dbSize;
		}
	}

	public IntUserIdHelper(String s) {
		super(s);
	}

	/**
	 * Default implementation return true.
	 * @return true if the oid of the created object must be kept into the
	 * context
	 */
	protected boolean keepOid() {
		return true;
	}

	protected String[] getClassNamesToInit() {
		return new String[]{IntUserId.class.getName()};
	}

	/**
	 * IMPORTANT: dot not removed data on support in order to avoid next
	 * creations
	 */
	public void setUp() throws Exception {
		logger.log(BasicLevel.DEBUG, "setUp.");
		cleanup();
		initDataStructure(false);
	}

	protected void prepareTask(Task task, Object _ctx) {
		debug = logger.isLoggable(BasicLevel.DEBUG);
		super.prepareTask(task, _ctx);
		IUICtx ctx = (IUICtx) _ctx;
		if (oids == null) {
			synchronized (getClass()) {
				if (oids == null && !Boolean.getBoolean(NO_DB_INIT)) {
					ctx.oids = new Object[ctx.dbSize];
					logger.log(BasicLevel.INFO, "\tPreparing test...");
					//Initialisation the database
					PrepareTest prep = new PrepareTest(this);
					try {
						prep.setUp();
					} catch (Exception e) {
						fail(e.getMessage());
					}
					prep.prepare(ctx, keepOid());
					prep.tearDown();
					logger.log(BasicLevel.INFO, "\ttest Prepared");
					if (keepOid()) {
						//keep oids in the static variable
						oids = ctx.oids;
					} else {
						//db initialized without oids
						oids = new Object[0];
					}
				}
			}
		}
		ctx.oids = oids;
	}
}

class PrepareTest extends StressHelper {

	private final static int NB_CREATION = 100;
	private final static int NB_THREAD = 4;

	public PrepareTest(IntUserIdHelper helper) {
		super(helper.getName());
	}

	protected String[] getClassNamesToInit() {
		return new String[]{IntUserId.class.getName()};
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".IntUserIdHelper";
	}

	protected String getLogPrefix() {
		return super.getLogPrefix() + "\t";
	}

	/**
	 * IMPORTANT: dot not removed data on support in order to avoid next
	 * creations
	 */
	public void setUp() throws Exception {
		logger.log(BasicLevel.DEBUG, "setUp.");
		cleanup();
		initDataStructure(false);
	}

	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       PerformResult res) {
		IntUserIdHelper.IUICtx gctx = (IntUserIdHelper.IUICtx) ctx;
		final int plus = gctx.dbSize % NB_CREATION;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			if (plus > 0) {
				//The first transaction creates the additional object
				for (int oid = 0; txId == 0 && oid < plus; oid++) {
					IntUserId iui = new IntUserId(oid, "Obj No " + oid);
					pm.makePersistent(iui);
					if (gctx.keepOid) {
						gctx.oids[oid] = pm.getObjectId(iui);
					}
				}
			} 
			// The other transactions create 'nbCreation' objects
			for (int no = 0; no < NB_CREATION; no++) {
				int oid = (txId * NB_CREATION) + no + plus;
				if (oid < gctx.dbSize) {
					IntUserId iui = new IntUserId(oid, "Obj No " + oid);
					pm.makePersistent(iui);
					if (gctx.keepOid) {
						gctx.oids[oid] = pm.getObjectId(iui);
					}
				}
			}

			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	/**
	 * Tests the concurrency of several writer on a same persistent object.
	 * (500 users)
	 */
	public void prepare(IntUserIdHelper.IUICtx ctx, boolean keepOid) {
		int plus = ctx.dbSize % NB_CREATION;
		int nbTx = (ctx.dbSize / NB_CREATION) + (plus > 0 ? 1 : 0);
		ctx.keepOid = keepOid;
		perform(NB_THREAD, nbTx, Integer.getInteger(TIMEOUT, 200000).intValue(), ctx);
	}
}

