/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.pobjects.userid.IntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;

import junit.framework.TestSuite;
import junit.textui.TestRunner;

/**
 * Stresses the creation function of Speedo.
 * @author P. Dechamboux
 */
public class TestCreDel extends StressHelper {

	protected String CREATION = getLoggerName() + ".creation";
	protected String THRESHOLD = getLoggerName() + ".threshold";

	public TestCreDel(String s) {
		super(s);
	}
	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".stress.TestCreDel";
	}

	protected String[] getClassNamesToInit() {
		return new String[]{IntUserId.class.getName()};
	}

	public static void main(String[] args) {
		TestRunner.run(new TestSuite(TestCreDel.class));
	}

	class CreDelContext {
		public int nbCreDel;
		public int startId;
		public int threshold;
		public CreDelContext(int nbCreDel, int startId, int threshold) {
			this.nbCreDel = nbCreDel;
			this.startId = startId;
			this.threshold = threshold;
		}

		public String toString() {
			return "nbCreDel = " + nbCreDel + " / startid = " + startId
			        + " / threshold = " + threshold;
		}
	}

	class CreDelThreadManager extends TaskManager {
		public boolean[] txCreateStarted;
		public Threshold threshold;

		public CreDelThreadManager(int[] nbThread,
		                           int[] nbTx,
		                           int timeout,
		                           Object ctx) {
			super(nbThread, nbTx, timeout, ctx);
			threshold = new Threshold(((CreDelContext) ctx).threshold);
			txCreateStarted = new boolean[nbTx[0]];
		}

		public Task newTask(int taskId,
		                    int nbTx,
		                    int nbThread,
		                    TaskManager taskManager,
		                    Object ctx) {
			return new CreDelTask(taskId, nbTx, nbThread,taskManager);
		}
	}

	class CreDelTask extends Task {
		public CreDelTask(int taskId,
		                  int nbTx,
		                  int nbThread,
						  TaskManager taskManager) {
			super(taskId, nbTx, nbThread, taskManager);
		}

		public int getTxId(int threadId) {
			if (taskId == 0) {
				return getCreateTxId(threadId);
			} else if (taskId == 1) {
				return getDeleteTxId(threadId);
			} else {
				return -1;
			}
		}
		public int getCreateTxId(int threadId) {
			synchronized (txToExecute) {
				boolean[] txCreateStarted = ((CreDelThreadManager) taskManager).txCreateStarted;
				for (int txId = 0; txId < txToExecute.length; txId++) {
					if (txToExecute[txId]) {
						if (txCreateStarted[txId]) {
							continue;
						}
						txCreateStarted[txId] = true;
						return txId;
					}
				}
			}
			return -1;
		}
		public int getDeleteTxId(int threadId) {
			boolean[] txCreateToExecute = ((CreDelThreadManager) taskManager).tasks[0].txToExecute;
			synchronized (txCreateToExecute) {
				int txId;
				int nbDel = 0;
				for (txId = 0; txId < txToExecute.length; txId++) {
					if (!txToExecute[txId]) {
						nbDel++;
						continue;
					}
					if (!txCreateToExecute[txId]) {
						break;
					}
				}
				if (txId == txToExecute.length) {
					if (nbDel == txToExecute.length) {
						return -1;
					}
					try {
						logger.log(BasicLevel.DEBUG,
						        "Deletion thread " + threadId + " go to sleep");
						txCreateToExecute.wait();
						logger.log(BasicLevel.DEBUG,
						        "Deletion thread " + threadId + " wakes up!!!");
					} catch (InterruptedException e) {
					}
					return -2;
				} else {
					if (!((CreDelThreadManager) taskManager).threshold.canDelete()) {
						try {
							logger.log(BasicLevel.DEBUG, "Deletion thread "
							        + threadId + " go to sleep (threshold/level="
							        + ((CreDelThreadManager) taskManager).threshold.getLevel() + ")...");
							txCreateToExecute.wait();
							logger.log(BasicLevel.DEBUG, "Deletion thread "
							        + threadId + " wakes up (threshold)!!!");
						} catch (InterruptedException e) {
						}
						return -2;
					}
				}
				txToExecute[txId] = false;
				((CreDelThreadManager) taskManager).threshold.deleteDone();
				return txId;
			}
		}
	}

	protected void perform(Task task, int threadId, int txId, Object ctx, PerformResult res) {
		if (task.taskId == 0) {
			performCreate(task, threadId, txId, ctx, res);
		} else if (task.taskId == 1) {
			performRemove(task, threadId, txId, ctx, res);
		} else {
			throw new RuntimeException("Task Id not managed: " + task.taskId);
		}
	}

	protected void performCreate(Task task, int threadId, int txId, Object _ctx, PerformResult res) {
		CreDelContext ctx = (CreDelContext) _ctx;
		PersistenceManager pm = getPM(task, threadId,txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			for (int no = 0; no < ctx.nbCreDel; no++) {
				int oid = (txId * ctx.nbCreDel) + no + ctx.startId;
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Creating object "  + oid);
				}
				IntUserId iui = new IntUserId(oid, "Obj No " + oid);
				pm.makePersistent(iui);
			}
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackExceptions.add(e);
			res.ok = false;
			logger.log(BasicLevel.INFO, "Creation thread " + threadId + " has been rolledback: try again !", e);
			if (pm.currentTransaction().isActive()) {
				pm.currentTransaction().rollback();
				errors.add(e);
			}
		} catch (Exception e) {
			Exception ie = ExceptionHelper.getNested(e);
			errors.add(ie);
			res.ok = false;
			logger.log(BasicLevel.ERROR, "Creation thread " + threadId + " has a problem", ie);
			pm.currentTransaction().rollback();
		} catch (Throwable e) {
			logger.log(BasicLevel.ERROR, "Creation thread " + threadId + " has a problem", e);
			res.ok = false;
			pm.currentTransaction().rollback();
		} finally {
			try {
				task.txToExecute[txId] = false;
				pm.close();
			} catch (JDOException e) {
				logger.log(BasicLevel.ERROR, "Creation thread " + threadId
				        + " has been "
				        + (res.ok ? "committed" : "rolledback")
				        + " and the close occurs an error", e);
				throw e;
			}
		}

		synchronized (task.txToExecute) {
			task.txToExecute[txId] = false;
			Threshold threshold = ((CreDelThreadManager) task.taskManager).threshold;
			threshold.createDone();
			if (txId == (task.txToExecute.length - 1)) {
				for (int i = 0; i < threshold.getThreshold(); i++) {
					threshold.createDone();
				}
			}
			task.txToExecute.notifyAll();
		}
	}

	protected void performRemove(Task task, int threadId, int txId, Object _ctx, PerformResult res) {
		CreDelContext ctx = (CreDelContext) _ctx;
		PersistenceManager pm = getPM(task, threadId,txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			for (int no = 0; no < ctx.nbCreDel; no++) {
				int id = (txId * ctx.nbCreDel) + no + ctx.startId;
				Object oid = pm.newObjectIdInstance(IntUserId.class, String.valueOf(id));
				pm.deletePersistent(pm.getObjectById(oid, false));
			}
			pm.currentTransaction().commit();
			res.endTest();
			logger.log(BasicLevel.DEBUG, "Deletion Tx " + txId + " finished");
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	private void perform(int nbth, int nbtx, int nbc, int to, int startid) {
		perform(nbth, nbtx, to, new CreDelContext(nbc, startid, 0));
	}



	/**
	 * Tests the creation of a lot of persistent objects, with interactive
	 * setting of test parameteres (see file userconf/project.properties).
	 */
	public void testInteractive() {
		if (interactive) {
			perform(Integer.getInteger(THREAD, 10).intValue(),
			        Integer.getInteger(TX, 100).intValue(),
			        Integer.getInteger(TIMEOUT, 1000000).intValue(),
			        new CreDelContext(
			                Integer.getInteger(CREATION, 1000).intValue(), 0,
			                Integer.getInteger(THRESHOLD, 0).intValue()));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1 object using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr1() {
		if (!interactive) {
			perform(1, 10, 1, 1000000, getStartId(ID_Th1Tx10Cr1));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1 object using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr1() {
		if (!interactive) {
			perform(1, 100, 1, 1000000, getStartId(ID_Th1Tx100Cr1));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1 object using 1 thread.
	 */
	public void testCreDelTh1Tx1000Cr1() {
		if (!interactive) {
			perform(1, 1000, 1, 1000000, getStartId(ID_Th1Tx1000Cr1));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 1 object using 1 thread.
	 */
	public void testCreDelTh1Tx10000Cr1() {
		if (!interactive) {
			perform(1, 10000, 1, 1000000, getStartId(ID_Th1Tx10000Cr1));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 1 object using 1 thread.
	 */
	public void testCreDelTh1Tx100000Cr1() {
		if (!interactive) {
			perform(1, 100000, 1, 10000000, getStartId(ID_Th1Tx100000Cr1));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr10() {
		if (!interactive) {
			perform(1, 10, 10, 1000000, getStartId(ID_Th1Tx10Cr10));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr10() {
		if (!interactive) {
			perform(1, 100, 10, 1000000, getStartId(ID_Th1Tx100Cr10));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 10 objects using 1 thread.
	 */
	public void testCreDelTh1Tx1000Cr10() {
		if (!interactive) {
			perform(1, 1000, 10, 1000000, getStartId(ID_Th1Tx1000Cr10));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 10 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10000Cr10() {
		if (!interactive) {
			perform(1, 10000, 10, 10000000, getStartId(ID_Th1Tx10000Cr10));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 10 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100000Cr10() {
		if (!interactive) {
			perform(1, 100000, 10, 100000000, getStartId(ID_Th1Tx100000Cr10));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 100 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr100() {
		if (!interactive) {
			perform(1, 10, 100, 1000000, getStartId(ID_Th1Tx10Cr100));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 100 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr100() {
		if (!interactive) {
			perform(1, 100, 100, 1000000, getStartId(ID_Th1Tx100Cr100));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 100 objects using 1 thread.
	 */
	public void testCreDelTh1Tx1000Cr100() {
		if (!interactive) {
			perform(1, 1000, 100, 10000000, getStartId(ID_Th1Tx1000Cr100));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 100 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10000Cr100() {
		if (!interactive) {
			perform(1, 10000, 100, 100000000, getStartId(ID_Th1Tx10000Cr100));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 500 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr500() {
		if (!interactive) {
			perform(1, 10, 500, 1000000, getStartId(ID_Th1Tx10Cr500));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 500 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr500() {
		if (!interactive) {
			perform(1, 100, 500, 10000000, getStartId(ID_Th1Tx100Cr500));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 500 objects using 1 thread.
	 */
	public void testCreDelTh1Tx1000Cr500() {
		if (!interactive) {
			perform(1, 1000, 500, 100000000, getStartId(ID_Th1Tx1000Cr500));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr1000() {
		if (!interactive) {
			perform(1, 10, 1000, 1000000, getStartId(ID_Th1Tx10Cr1000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr1000() {
		if (!interactive) {
			perform(1, 100, 1000, 10000000, getStartId(ID_Th1Tx100Cr1000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx1000Cr1000() {
		if (!interactive) {
			perform(1, 1000, 1000, 100000000, getStartId(ID_Th1Tx1000Cr1000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 2.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr2000() {
		if (!interactive) {
			perform(1, 10, 2000, 10000000, getStartId(ID_Th1Tx10Cr2000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 2.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr2000() {
		if (!interactive) {
			perform(1, 100, 2000, 100000000, getStartId(ID_Th1Tx100Cr2000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 2.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx1000Cr2000() {
		if (!interactive) {
			perform(1, 1000, 2000, 1000000000, getStartId(ID_Th1Tx1000Cr2000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx10Cr10000() {
		if (!interactive) {
			perform(1, 10, 10000, 100000000, getStartId(ID_Th1Tx10Cr10000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10.000 objects using 1 thread.
	 */
	public void testCreDelTh1Tx100Cr10000() {
		if (!interactive) {
			perform(1, 100, 10000, 1000000000, getStartId(ID_Th1Tx100Cr10000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1 object using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr1() {
		if (!interactive) {
			perform(2, 10, 1, 1000000, getStartId(ID_Th2Tx10Cr1));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1 object using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr1() {


		if (!interactive) {
			perform(2, 100, 1, 1000000, getStartId(ID_Th2Tx100Cr1));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1 object using 2 threads.
	 */
	public void testCreDelTh2Tx1000Cr1() {


		if (!interactive) {
			perform(2, 1000, 1, 1000000, getStartId(ID_Th2Tx1000Cr1));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 1 object using 2 threads.
	 */
	public void testCreDelTh2Tx10000Cr1() {


		if (!interactive) {
			perform(2, 10000, 1, 1000000, getStartId(ID_Th2Tx10000Cr1));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 1 object using 2 threads.
	 */
	public void testCreDelTh2Tx100000Cr1() {


		if (!interactive) {
			perform(2, 100000, 1, 1000000, getStartId(ID_Th2Tx100000Cr1));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr10() {


		if (!interactive) {
			perform(2, 10, 10, 1000000, getStartId(ID_Th2Tx10Cr10));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr10() {


		if (!interactive) {
			perform(2, 100, 10, 1000000, getStartId(ID_Th2Tx100Cr10));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 10 objects using 2 threads.
	 */
	public void testCreDelTh2Tx1000Cr10() {


		if (!interactive) {
			perform(2, 1000, 10, 1000000, getStartId(ID_Th2Tx1000Cr10));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 10 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10000Cr10() {


		if (!interactive) {
			perform(2, 10000, 10, 1000000, getStartId(ID_Th2Tx10000Cr10));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 10 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100000Cr10() {


		if (!interactive) {
			perform(2, 100000, 10, 1000000, getStartId(ID_Th2Tx100000Cr10));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 100 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr100() {


		if (!interactive) {
			perform(2, 10, 100, 1000000, getStartId(ID_Th2Tx10Cr100));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 100 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr100() {


		if (!interactive) {
			perform(2, 100, 100, 1000000, getStartId(ID_Th2Tx100Cr100));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 100 objects using 2 threads.
	 */
	public void testCreDelTh2Tx1000Cr100() {


		if (!interactive) {
			perform(2, 1000, 100, 1000000, getStartId(ID_Th2Tx1000Cr100));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 100 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10000Cr100() {


		if (!interactive) {
			perform(2, 10000, 100, 1000000, getStartId(ID_Th2Tx10000Cr100));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 500 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr500() {


		if (!interactive) {
			perform(2, 10, 500, 1000000, getStartId(ID_Th2Tx10Cr500));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 500 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr500() {


		if (!interactive) {
			perform(2, 100, 500, 1000000, getStartId(ID_Th2Tx100Cr500));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 500 objects using 2 threads.
	 */
	public void testCreDelTh2Tx1000Cr500() {


		if (!interactive) {
			perform(2, 1000, 500, 1000000, getStartId(ID_Th2Tx1000Cr500));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr1000() {


		if (!interactive) {
			perform(2, 10, 1000, 1000000, getStartId(ID_Th2Tx10Cr1000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr1000() {


		if (!interactive) {
			perform(2, 100, 1000, 1000000, getStartId(ID_Th2Tx100Cr1000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx1000Cr1000() {


		if (!interactive) {
			perform(2, 1000, 1000, 1000000, getStartId(ID_Th2Tx1000Cr1000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 2.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr2000() {


		if (!interactive) {
			perform(2, 10, 2000, 10000000, getStartId(ID_Th2Tx10Cr2000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 2.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr2000() {


		if (!interactive) {
			perform(2, 100, 2000, 10000000, getStartId(ID_Th2Tx100Cr2000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 2.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx1000Cr2000() {


		if (!interactive) {
			perform(2, 1000, 2000, 10000000, getStartId(ID_Th2Tx1000Cr2000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx10Cr10000() {


		if (!interactive) {
			perform(2, 10, 10000, 10000000, getStartId(ID_Th2Tx10Cr10000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10.000 objects using 2 threads.
	 */
	public void testCreDelTh2Tx100Cr10000() {


		if (!interactive) {
			perform(2, 100, 10000, 10000000, getStartId(ID_Th2Tx100Cr10000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1 object using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr1() {


		if (!interactive) {
			perform(3, 10, 1, 1000000, getStartId(ID_Th3Tx10Cr1));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1 object using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr1() {


		if (!interactive) {
			perform(3, 100, 1, 1000000, getStartId(ID_Th3Tx100Cr1));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1 object using 3 threads.
	 */
	public void testCreDelTh3Tx1000Cr1() {


		if (!interactive) {
			perform(3, 1000, 1, 1000000, getStartId(ID_Th3Tx1000Cr1));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 1 object using 3 threads.
	 */
	public void testCreDelTh3Tx10000Cr1() {


		if (!interactive) {
			perform(3, 10000, 1, 1000000, getStartId(ID_Th3Tx10000Cr1));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 1 object using 3 threads.
	 */
	public void testCreDelTh3Tx100000Cr1() {


		if (!interactive) {
			perform(3, 100000, 1, 1000000, getStartId(ID_Th3Tx100000Cr1));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr10() {


		if (!interactive) {
			perform(3, 10, 10, 1000000, getStartId(ID_Th3Tx10Cr10));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr10() {


		if (!interactive) {
			perform(3, 100, 10, 1000000, getStartId(ID_Th3Tx100Cr10));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 10 objects using 3 threads.
	 */
	public void testCreDelTh3Tx1000Cr10() {


		if (!interactive) {
			perform(3, 1000, 10, 1000000, getStartId(ID_Th3Tx1000Cr10));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 10 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10000Cr10() {


		if (!interactive) {
			perform(3, 10000, 10, 1000000, getStartId(ID_Th3Tx10000Cr10));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 10 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100000Cr10() {


		if (!interactive) {
			perform(3, 100000, 10, 1000000, getStartId(ID_Th3Tx100000Cr10));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 100 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr100() {


		if (!interactive) {
			perform(3, 10, 100, 1000000, getStartId(ID_Th3Tx10Cr100));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 100 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr100() {


		if (!interactive) {
			perform(3, 100, 100, 1000000, getStartId(ID_Th3Tx100Cr100));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 100 objects using 3 threads.
	 */
	public void testCreDelTh3Tx1000Cr100() {


		if (!interactive) {
			perform(3, 1000, 100, 1000000, getStartId(ID_Th3Tx1000Cr100));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 100 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10000Cr100() {


		if (!interactive) {
			perform(3, 10000, 100, 1000000, getStartId(ID_Th3Tx10000Cr100));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 500 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr500() {


		if (!interactive) {
			perform(3, 10, 500, 1000000, getStartId(ID_Th3Tx10Cr500));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 500 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr500() {


		if (!interactive) {
			perform(3, 100, 500, 1000000, getStartId(ID_Th3Tx100Cr500));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 500 objects using 3 threads.
	 */
	public void testCreDelTh3Tx1000Cr500() {


		if (!interactive) {
			perform(3, 1000, 500, 1000000, getStartId(ID_Th3Tx1000Cr500));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr1000() {


		if (!interactive) {
			perform(3, 10, 1000, 1000000, getStartId(ID_Th3Tx10Cr1000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr1000() {


		if (!interactive) {
			perform(3, 100, 1000, 1000000, getStartId(ID_Th3Tx100Cr1000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx1000Cr1000() {


		if (!interactive) {
			perform(3, 1000, 1000, 1000000, getStartId(ID_Th3Tx1000Cr1000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 2.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr2000() {


		if (!interactive) {
			perform(3, 10, 2000, 10000000, getStartId(ID_Th3Tx10Cr2000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 2.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr2000() {


		if (!interactive) {
			perform(3, 100, 2000, 10000000, getStartId(ID_Th3Tx100Cr2000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 2.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx1000Cr2000() {


		if (!interactive) {
			perform(3, 1000, 2000, 10000000, getStartId(ID_Th3Tx1000Cr2000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx10Cr10000() {


		if (!interactive) {
			perform(3, 10, 10000, 10000000, getStartId(ID_Th3Tx10Cr10000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10.000 objects using 3 threads.
	 */
	public void testCreDelTh3Tx100Cr10000() {


		if (!interactive) {
			perform(3, 100, 10000, 10000000, getStartId(ID_Th3Tx100Cr10000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1 object using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr1() {


		if (!interactive) {
			perform(4, 10, 1, 1000000, getStartId(ID_Th4Tx10Cr1));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1 object using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr1() {


		if (!interactive) {
			perform(4, 100, 1, 1000000, getStartId(ID_Th4Tx100Cr1));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1 object using 4 threads.
	 */
	public void testCreDelTh4Tx1000Cr1() {


		if (!interactive) {
			perform(4, 1000, 1, 1000000, getStartId(ID_Th4Tx1000Cr1));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 1 object using 4 threads.
	 */
	public void testCreDelTh4Tx10000Cr1() {


		if (!interactive) {
			perform(4, 10000, 1, 1000000, getStartId(ID_Th4Tx10000Cr1));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 1 object using 4 threads.
	 */
	public void testCreDelTh4Tx100000Cr1() {


		if (!interactive) {
			perform(4, 100000, 1, 1000000, getStartId(ID_Th4Tx100000Cr1));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr10() {


		if (!interactive) {
			perform(4, 10, 10, 1000000, getStartId(ID_Th4Tx10Cr10));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr10() {


		if (!interactive) {
			perform(4, 100, 10, 1000000, getStartId(ID_Th4Tx100Cr10));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 10 objects using 4 threads.
	 */
	public void testCreDelTh4Tx1000Cr10() {


		if (!interactive) {
			perform(4, 1000, 10, 1000000, getStartId(ID_Th4Tx1000Cr10));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 10 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10000Cr10() {


		if (!interactive) {
			perform(4, 10000, 10, 1000000, getStartId(ID_Th4Tx10000Cr10));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 10 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100000Cr10() {


		if (!interactive) {
			perform(4, 100000, 10, 1000000, getStartId(ID_Th4Tx100000Cr10));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 100 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr100() {


		if (!interactive) {
			perform(4, 10, 100, 1000000, getStartId(ID_Th4Tx10Cr100));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 100 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr100() {


		if (!interactive) {
			perform(4, 100, 100, 1000000, getStartId(ID_Th4Tx100Cr100));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 100 objects using 4 threads.
	 */
	public void testCreDelTh4Tx1000Cr100() {


		if (!interactive) {
			perform(4, 1000, 100, 1000000, getStartId(ID_Th4Tx1000Cr100));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 100 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10000Cr100() {


		if (!interactive) {
			perform(4, 10000, 100, 1000000, getStartId(ID_Th4Tx10000Cr100));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 500 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr500() {


		if (!interactive) {
			perform(4, 10, 500, 1000000, getStartId(ID_Th4Tx10Cr500));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 500 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr500() {


		if (!interactive) {
			perform(4, 100, 500, 1000000, getStartId(ID_Th4Tx100Cr500));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 500 objects using 4 threads.
	 */
	public void testCreDelTh4Tx1000Cr500() {


		if (!interactive) {
			perform(4, 1000, 500, 1000000, getStartId(ID_Th4Tx1000Cr500));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr1000() {


		if (!interactive) {
			perform(4, 10, 1000, 1000000, getStartId(ID_Th4Tx10Cr1000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr1000() {


		if (!interactive) {
			perform(4, 100, 1000, 1000000, getStartId(ID_Th4Tx100Cr1000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx1000Cr1000() {


		if (!interactive) {
			perform(4, 1000, 1000, 1000000, getStartId(ID_Th4Tx1000Cr1000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 2.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr2000() {


		if (!interactive) {
			perform(4, 10, 2000, 10000000, getStartId(ID_Th4Tx10Cr2000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 2.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr2000() {


		if (!interactive) {
			perform(4, 100, 2000, 10000000, getStartId(ID_Th4Tx100Cr2000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 2.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx1000Cr2000() {


		if (!interactive) {
			perform(4, 1000, 2000, 10000000, getStartId(ID_Th4Tx1000Cr2000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx10Cr10000() {


		if (!interactive) {
			perform(4, 10, 10000, 10000000, getStartId(ID_Th4Tx10Cr10000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10.000 objects using 4 threads.
	 */
	public void testCreDelTh4Tx100Cr10000() {


		if (!interactive) {
			perform(4, 100, 10000, 10000000, getStartId(ID_Th4Tx100Cr10000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1 object using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr1() {


		if (!interactive) {
			perform(10, 10, 1, 1000000, getStartId(ID_Th10Tx10Cr1));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1 object using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr1() {


		if (!interactive) {
			perform(10, 100, 1, 1000000, getStartId(ID_Th10Tx100Cr1));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1 object using 10 threads.
	 */
	public void testCreDelTh10Tx1000Cr1() {


		if (!interactive) {
			perform(10, 1000, 1, 1000000, getStartId(ID_Th10Tx1000Cr1));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 1 object using 10 threads.
	 */
	public void testCreDelTh10Tx10000Cr1() {


		if (!interactive) {
			perform(10, 10000, 1, 1000000, getStartId(ID_Th10Tx10000Cr1));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 1 object using 10 threads.
	 */
	public void testCreDelTh10Tx100000Cr1() {


		if (!interactive) {
			perform(10, 100000, 1, 1000000, getStartId(ID_Th10Tx100000Cr1));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr10() {


		if (!interactive) {
			perform(10, 10, 10, 1000000, getStartId(ID_Th10Tx10Cr10));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr10() {


		if (!interactive) {
			perform(10, 100, 10, 1000000, getStartId(ID_Th10Tx100Cr10));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 10 objects using 10 threads.
	 */
	public void testCreDelTh10Tx1000Cr10() {


		if (!interactive) {
			perform(10, 1000, 10, 1000000, getStartId(ID_Th10Tx1000Cr10));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 10 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10000Cr10() {


		if (!interactive) {
			perform(10, 10000, 10, 1000000, getStartId(ID_Th10Tx10000Cr10));
		}
	}

	/**
	 * Tests 100.000 transactions where each of them creates 10 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100000Cr10() {


		if (!interactive) {
			perform(10, 100000, 10, 1000000, getStartId(ID_Th10Tx100000Cr10));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 100 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr100() {


		if (!interactive) {
			perform(10, 10, 100, 1000000, getStartId(ID_Th10Tx10Cr100));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 100 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr100() {


		if (!interactive) {
			perform(10, 100, 100, 1000000, getStartId(ID_Th10Tx100Cr100));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 100 objects using 10 threads.
	 */
	public void testCreDelTh10Tx1000Cr100() {


		if (!interactive) {
			perform(10, 1000, 100, 1000000, getStartId(ID_Th10Tx1000Cr100));
		}
	}

	/**
	 * Tests 10.000 transactions where each of them creates 100 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10000Cr100() {


		if (!interactive) {
			perform(10, 10000, 100, 1000000, getStartId(ID_Th10Tx10000Cr100));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 500 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr500() {
		if (!interactive) {
			perform(10, 10, 500, 1000000, getStartId(ID_Th10Tx10Cr500));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 500 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr500() {
		if (!interactive) {
			perform(10, 100, 500, 1000000, getStartId(ID_Th10Tx100Cr500));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 500 objects using 10 threads.
	 */
	public void testCreDelTh10Tx1000Cr500() {


		if (!interactive) {
			perform(10, 1000, 500, 1000000, getStartId(ID_Th10Tx1000Cr500));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr1000() {
		if (!interactive) {
			perform(10, 10, 1000, 1000000, getStartId(ID_Th10Tx10Cr1000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 1.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr1000() {
		if (!interactive) {
			perform(10, 100, 1000, 1000000, getStartId(ID_Th10Tx100Cr1000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 1.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx1000Cr1000() {
		if (!interactive) {
			perform(10, 1000, 1000, 1000000, getStartId(ID_Th10Tx1000Cr1000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 2.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr2000() {
		if (!interactive) {
			perform(10, 10, 2000, 10000000, getStartId(ID_Th10Tx10Cr2000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 2.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr2000() {
		if (!interactive) {
			perform(10, 100, 2000, 10000000, getStartId(ID_Th10Tx100Cr2000));
		}
	}

	/**
	 * Tests 1.000 transactions where each of them creates 2.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx1000Cr2000() {
		if (!interactive) {
			perform(10, 1000, 2000, 10000000, getStartId(ID_Th10Tx1000Cr2000));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 10.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx10Cr10000() {
		if (!interactive) {
			perform(10, 10, 10000, 10000000, getStartId(ID_Th10Tx10Cr10000));
		}
	}

	/**
	 * Tests 100 transactions where each of them creates 10.000 objects using 10 threads.
	 */
	public void testCreDelTh10Tx100Cr10000() {
		if (!interactive) {
			perform(10, 100, 10000, 10000000, getStartId(ID_Th10Tx100Cr10000));
		}
	}
}

class Threshold {
	private final int threshold;
	private int level;

	Threshold(int th) {
		threshold = th;
		level = 0;
	}

	void createDone() {
		level++;
	}

	void deleteDone() {
		level--;
	}

	boolean canDelete() {
		return level >= threshold;
	}

	int getLevel() {
		return level;
	}

	int getThreshold() {
		return threshold;
	}
}
