/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import org.objectweb.speedo.Alea;
import org.objectweb.speedo.pobjects.userid.LinkedIntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;

/**
 * Stresses the read access function of Speedo.
 *
 * @author M. Guillemin/S.Rodiere
 */
public class TestGetObject extends LinkedIntUserIdHelper {

	private String NBREAD = getLoggerName() + ".nbread";
	private String NBREADPERSISTENT = getLoggerName() + ".nbreadpersistent";

	public class GetCtx extends LIUICtx {
		public int nbRead;
		public int nbReadPersistent;
		public ObjectStatistic stat = new ObjectStatistic();

		public GetCtx(int dbSize, int nbRead, int nbReadPersistent) {
			super(dbSize);
			this.nbRead = nbRead;
			this.nbReadPersistent = nbReadPersistent;
		}

		public String toString() {
			return super.toString() + " / nbRead = " + nbRead
			        + " / nbReadPersistent = " + nbReadPersistent;
		}
	}

	public TestGetObject(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestGetObject";
	}

	protected void perform(Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       PerformResult res) {
		GetCtx gctx = (GetCtx) ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			// Execute one transaction: do nbRead access to a persistent
			LinkedIntUserId oUserId;
			for (int no = 0; no < gctx.nbRead; no++) {
				int oid = Alea.rand(0, gctx.dbSize - 1);
				gctx.stat.touch(new Integer(oid));
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Reading object " + oid);
				}
				oUserId = (LinkedIntUserId)
				        pm.getObjectById(gctx.oids[oid], false);
				assertNotNull("Null object (oid = " + oid + ")", oUserId);
				oUserId.readName(gctx.nbReadPersistent);
			}
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	protected void logStatistic(TaskManager tm, long exectime, Object _ctx) {
		GetCtx ctx = (GetCtx) _ctx;
		logger.log(BasicLevel.INFO,  getLogPrefix() + "\tGetObject: objects read:"
		        + ctx.stat.getAccessedOid().size()
		        + ", nb access: " + ctx.stat.getNbAccess());
		super.logStatistic(tm, exectime, ctx);
	}

	/**
	 * Tests the read of a lot of persistent objects, with interactive setting
	 * of test parameteres (see file userconf/project.properties).
	 */
	public void testInteractive() {
		if (interactive) {
			logger.log(BasicLevel.INFO, "testInteractive");
			perform(Integer.getInteger(THREAD, 10).intValue(),
			        Integer.getInteger(TX, 100).intValue(),
			        Integer.getInteger(TIMEOUT, 200000).intValue(),
			        new GetCtx(
			                Integer.getInteger(DBSIZE, 1000).intValue(),
			                Integer.getInteger(NBREAD, 10).intValue(),
			                Integer.getInteger(NBREADPERSISTENT, 10).intValue()));
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testInteractive");
		}
	}

	private void perform(
	        int nbThread,
	        int dbSize,
	        int nbTx,
	        int nbRead,
	        int nbReadPersistent,
	        int threadTimeout) {
		perform(nbThread, nbTx, threadTimeout,
		        new GetCtx(dbSize, nbRead, nbReadPersistent));
	}

	/**
	 * Tests 100 transactions where each of them read 1 object with 1 read
	 * using 1 thread.
	 */
	public void testGetObjectTh1Tx100Re1Rp1() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh1Tx100Re1Rp1");
			perform(1, 10000, 100, 1, 1, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh1Tx100Re1Rp1");
		}
	}

	/**
	 * Tests 100 transactions where each of them read 1 object with 1 read
	 * using 2 thread.
	 */
	public void testGetObjectTh2Tx100Re1Rp1() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh2Tx100Re1Rp1");
			perform(2, 10000, 100, 1, 1, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh2Tx100Re1Rp1");
		}
	}


	/**
	 * Tests 100 transactions where each of them read 1 object with 10 read
	 * using 1 thread.
	 */
	public void testGetObjectTh1Tx100Re1Rp10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh1Tx100Re1Rp10");
			perform(1, 10000, 100, 1, 10, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh1Tx100Re1Rp10");
		}
	}

	/**
	 * Tests 100 transactions where each of them read 1 object with 10 read
	 * using 2 thread.
	 */
	public void testGetObjectTh2Tx100Re1Rp10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh2Tx100Re1Rp10");
			perform(2, 10000, 100, 1, 10, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh2Tx100Re1Rp10");
		}
	}

	/**
	 * Tests 100 transactions where each of them read 1 object with 10 read
	 * using 10 thread.
	 */
	public void testGetObjectTh10Tx100Re1Rp10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh10Tx100Re1Rp10");
			perform(10, 10000, 100, 1, 10, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh10Tx100Re1Rp10");
		}
	}

	/**
	 * Tests 1000 transactions where each of them read 100 object with 100
	 read
	 * using 1 thread.
	 */
	public void testGetObjectTh1Tx1000Re10Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh1Tx1000Re10Rp100");
			perform(1, 10000, 1000, 10, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh1Tx1000Re10Rp100");
		}
	}

	/**
	 * Tests 1000 transactions where each of them read 100 object with 100
	 read
	 * using 2 thread.
	 */
	public void testGetObjectTh2Tx1000Re10Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh2Tx1000Re10Rp100");
			perform(2, 10000, 1000, 10, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh2Tx1000Re10Rp100");
		}
	}

	/**
	 * Tests 1000 transactions where each of them read 100 object with 100
	 read
	 * using 10 thread.
	 */
	public void testGetObjectTh10Tx1000Re10Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh10Tx1000Re10Rp100");
			perform(10, 10000, 1000, 10, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh10Tx1000Re10Rp100");
		}
	}

	/**
	 * Tests 10000 transactions where each of them read 100 object with 100
	 read
	 * using 1 thread.
	 */
	public void testGetObjectTh1Tx100Re100Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh1Tx100Re100Rp100");
			perform(1, 10000, 100, 100, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh1Tx100Re100Rp100");
		}
	}

	/**
	 * Tests 10000 transactions where each of them read 100 object with 100
	 read
	 * using 2 thread.
	 */
	public void testGetObjectTh2Tx100Re100Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh2Tx100Re100Rp100");
			perform(2, 10000, 100, 100, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh2Tx100Re100Rp100");
		}
	}

	/**
	 * Tests 10000 transactions where each of them read 100 object with
	 100
	 * read using 3 thread.
	 */
	public void testGetObjectTh3Tx100Re100Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh3Tx100Re100Rp100");
			perform(3, 10000, 100, 100, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh3Tx100Re100Rp100");
		}
	}

	/**
	 * Tests 10000 transactions where each of them read 100 object with 100
	 read
	 * using 10 thread.
	 */
	public void testGetObjectTh10Tx100Re100Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh10Tx100Re100Rp100");
			perform(10, 10000, 100, 100, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh10Tx100Re100Rp100");
		}
	}


	/**
	 * Tests 10000 transactions where each of them read 100 object with
	 10000
	 * read using 1 thread.
	 */
	public void testGetObjectTh1Tx100Re100Rp10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh1Tx100Re100Rp10000");
			perform(1, 10000, 100, 100, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh1Tx100Re100Rp10000");
		}
	}

	/**
	 * Tests 10000 transactions where each of them read 100 object with
	 10000
	 * read using 3 thread.
	 */
	public void testGetObjectTh3Tx100Re100Rp10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh3Tx100Re100Rp10000");
			perform(3, 10000, 100, 100, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh3Tx100Re100Rp10000");
		}
	}

	/**
	 * Tests 10000 transactions where each of them read 100 object with
	 100
	 * read using 100 thread.
	 */
	public void testGetObjectTh100Tx100Re100Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh100Tx100Re100Rp100");
			perform(100, 10000, 100, 100, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh100Tx100Re100Rp100");
		}
	}

	/**
	 * Tests 1000 transactions where each of them read 100 object with
	 100
	 * read using 1 thread.
	 */
	public void testGetObjectTh1000Tx100Re100Rp100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testGetObjectTh1000Tx100Re100Rp100");
			perform(100, 10000, 1000, 100, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testGetObjectTh1000Tx100Re100Rp100");
		}
	}
}
